#include "TPhSensorMapDB.h"

#include "TPhSensorMapDBSensorItem.h"
#include "TPhSensorMapDBFECItem.h"

#include "TClass.h"

#include <string.h>
#include <stdio.h>
#include <iostream.h>

const Int_t kMaxNumFEC = 50;

ClassImp(TPhSensorMapDB)
  ////////////////////////////////////////////////////////////////
  //
  // This class gives access to the geometrical sensor map database
  // 
  // $Id: TPhSensorMapDB.cxx,v 1.4 1998/10/22 15:31:24 decowski Exp $
  //
  ////////////////////////////////////////////////////////////////

TPhSensorMapDB::TPhSensorMapDB()
{
  // Default constructor
  
  fSensorTable = NULL;
  fFECTable = NULL;
  fVerbose = 0;
}

TPhSensorMapDB::~TPhSensorMapDB()
{
  // Default destructor

  if(fSensorTable) {
    fSensorTable->Delete();
    delete fSensorTable;
  }

  if(fFECTable) {
    fFECTable->Delete();
    delete fFECTable;
  }
}

Bool_t TPhSensorMapDB::ReadASCIIData(Char_t *FileName)
{
  // Initialize this object. This memberfunction is only called to create
  // a TPhSensorMapDB file, that then gets put in the SensorDB.root file. 
  // Once the file has been created from the Oracle database output, 
  // you DO NOT have to call this file. You just use the regular Streamer 
  // function to retrieve it. 
  //
  // Return: kFALSE if an error occured while reading the file in 
  // and kTRUE on success.

  TPhSensorMapDBSensorItem *ItemSensor;
  TPhSensorMapDBFECItem *ItemFEC;

  Char_t SensName[10];
  Int_t Type, FEC, String, ChPos, Row, Col, BondPat;
  FILE *handle;

  // The list of unused variables:
  Int_t Pad, SensorPos, PositionFlag, WurstID, WurstReadPos;
  Char_t HybridPos[10], ModTyp[20];

  handle = fopen(FileName, "r");

  if(handle == NULL)
    return kFALSE;

  fSensorTable = new THashTable();
  fFECTable = new TObjArray(kMaxNumFEC);

  // Read the ASCII File. Note that once past the BondPat, we do not do
  // anything with the read in information!
  while (fscanf(handle, "%s %d %d %d %d %d %d %d %d %d %d %d %d %s %s", 
		SensName, &Type, &FEC, &String, &ChPos, 
		&Row, &Col, &BondPat, &Pad,
		&SensorPos, &PositionFlag, &WurstID, &WurstReadPos,
		HybridPos, ModTyp) != EOF) {

    // Store the sensor information
    ItemSensor = 
      (TPhSensorMapDBSensorItem*) fSensorTable->FindObject(SensName);

    if(ItemSensor == NULL) {
      ItemSensor = new TPhSensorMapDBSensorItem(SensName, Type);
      fSensorTable->Add(ItemSensor);
    }
    ItemSensor->Store(Row, Col, FEC, String, ChPos, BondPat);

    // Store the FEC information
    ItemFEC = (TPhSensorMapDBFECItem*) fFECTable->At(FEC);
    if(ItemFEC == NULL) {
      ItemFEC = new TPhSensorMapDBFECItem(FEC);
      fFECTable->AddAt(ItemFEC, FEC);
    }
    ItemFEC->Store(String, ChPos, SensName, Row, Col, BondPat);
  }
  fclose(handle);
  
  // Set the title of the object
  SetTitle("(FEC, String, ChPos) <-> (Sensor, Row, Col) lookup Db");

  return kTRUE;
}

TPhSensorMapDBSensorItem* TPhSensorMapDB::GetSensor(Char_t *Name)
{
  // Get an Item describing the sensor of name Name. 
  //
  // Return: Pointer to the SensorItem, or NULL if not found.

  return (TPhSensorMapDBSensorItem*) fSensorTable->FindObject(Name);
}

TPhSensorMapDBFECItem* TPhSensorMapDB::GetFEC(Int_t Num)
{
  // Get an Item describing the FEC number Num
  //
  // Return: Pointer to the FECItem, or NULL if not found.

  return (TPhSensorMapDBFECItem*) fFECTable->At(Num);
}


Bool_t TPhSensorMapDB::Lookup(Char_t *Name, Int_t Row, Int_t Col, 
			      Int_t *FEC, Int_t *String, Int_t *ChPos, 
			      Int_t *BondPat)
{
  // Lookup what (Sensor, Row, Col) is in (FEC, String, ChPos) and BondPat
  // This is the users access point to the database. 
  // 
  // Return: kTRUE if successful, kFALSE if unsuccessful 

  TPhSensorMapDBSensorItem *Item;

  Item = (TPhSensorMapDBSensorItem*) fSensorTable->FindObject(Name);
  if(Item) {
    if(Row < Item->GetRows() && Col < Item->GetColumns()) 
      Item->Lookup(Row, Col, FEC, String, ChPos, BondPat);
    else {
      if(fVerbose) {
	Warning("Lookup", "Row or Col too large! Row = %d, Col = %d\n",
		Row, Col);
	return kFALSE;
      }
    }
  }
  else {
    if(fVerbose)
      Warning("Lookup", 
	      "Failed to find the entry in the SensorMap database!\n");
    *FEC = -1; *String = -1; *ChPos = -1; *BondPat = -1;
    return kFALSE;
  }
  return kTRUE;
}

Bool_t TPhSensorMapDB::Lookup(Char_t *Name, Int_t Row, Int_t Col, 
			      Int_t *FEC, Int_t *String, Int_t *ChPos)
{
  // Lookup what (Sensor, Row, Col) is in (FEC, String, ChPos)
  // This is the users access point to the database. 
  // 
  // Return: kTRUE if successful, kFALSE if unsuccessful 

  Int_t dummyBondPat;
  return Lookup(Name, Row, Col, FEC, String, ChPos, &dummyBondPat);

}

Bool_t TPhSensorMapDB::Lookup(Int_t FEC, Int_t String, Int_t ChPos, 
			      Char_t *Sensor, Int_t *Row, Int_t *Col, 
			      Int_t *BondPat, Int_t *Pad)
{
  // Lookup what (FEC, String, ChPos) is in (Sensor, Row, Col) and 
  // (BondPat, Pad)
  // This is the users access point to the database. 
  // 
  // Return: kTRUE if successful, kFALSE if unsuccessful 

  TPhSensorMapDBFECItem *FECItem;
  TPhSensorMapDBSensorItem *SensorItem;
  Bool_t OkFlag;

  FECItem = (TPhSensorMapDBFECItem*) fFECTable->At(FEC);
  if(FECItem)
    OkFlag = FECItem->Lookup(String, ChPos, Sensor, Row, Col, BondPat);
  else
    OkFlag = kFALSE;
  
  if(!OkFlag) {
    if(fVerbose)
      Warning("Lookup", 
	      "Failed to find FEC=%d String=%d ChPos=%d in SensorMapDB!\n",
	      FEC, String, ChPos);
    *Row = -1; *Col = -1; *BondPat = -1; *Pad = -1;
    return kFALSE;
  }

  // Calculate Pad
  SensorItem = (TPhSensorMapDBSensorItem*) fSensorTable->FindObject(Sensor);
  if(SensorItem)
    *Pad = ChPos % SensorItem->GetBondPatSize();
  else
    Error("Lookup", 
	  "Inconsistency! Sensor %s, FEC = %d String = %d ChPos = %d", 
	  Sensor, FEC, String, ChPos);

  return kTRUE;
}

Bool_t TPhSensorMapDB::Lookup(Int_t FEC, Int_t String, Int_t ChPos, 
			      Char_t *Sensor, Int_t *Row, Int_t *Col)
{
  // Lookup what (FEC, String, ChPos) is in (Sensor, Row, Col)
  // This is the users access point to the database. 
  // This routine is coded separately to avoid additional lookup when
  // not interested in (BondPattern, Pad)
  // 
  // Return: kTRUE if successful, kFALSE if unsuccessful 

  Int_t dummyBondPat;
  Bool_t OkFlag;

  TPhSensorMapDBFECItem *FECItem;

  FECItem = (TPhSensorMapDBFECItem*) fFECTable->At(FEC);
  if(FECItem)
    OkFlag = FECItem->Lookup(String, ChPos, Sensor, Row, Col, &dummyBondPat);
  else 
    OkFlag = kFALSE;

  if(!OkFlag) {
    if(fVerbose)
      Warning("Lookup", 
	      "Failed to find FEC=%d String=%d ChPos=%d in SensorMapDB!\n",
	      FEC, String, ChPos);
    *Row = -1; *Col = -1;
    return kFALSE;
  }

  return kTRUE;
}

void TPhSensorMapDB::ls(Option_t *option)
{
  // List all the FECItems and SensorItems that are currently 
  // available in the Mapping DB.
  
  IndentLevel();
  cout << IsA()->GetName() << " Name: " << GetName() 
       << " contains " << fFECTable->GetEntries() << " FEC Items and " 
       << fSensorTable->GetSize() << " Sensor Items" << endl;

  // Loop over all Sensor Items
  cout << "List of Sensor Items:" << endl;
  TObject::IncreaseDirLevel();
  fSensorTable->ls(option);
  TObject::DecreaseDirLevel();

  // Loop over all FEC Items
  cout << "List of FEC Items:" << endl;
  TObject::IncreaseDirLevel();
  fFECTable->ls(option);
  TObject::DecreaseDirLevel();
}
    
