#include "TPhEvent.h"
#include "TPhDataObject.h"
#include "TPhVertex.h"
#include "TPhDetector.h"
#include "TPhVTXSensor.h"
#include "TPhHitArray.h"

#include "TCanvas.h"
#include "TStyle.h"
#include "TROOT.h"

#include <iostream.h>

ClassImp(TPhEvent)
//////////////////////////////////////////////////////////////////////////
//
// TPhEvent is a PHOBOS data class providing storage and 
// access function for event data information. The data
// is stored in TPhDataObject objects that are kept inside
// a THashTable.
// This allows modification of the TPhEvent content by the 
// user, with the TPhDataObject providing a standard interface
// for information retrieval, bookkeeping and I/O selection.
// 
// The user adds the objects by calling the AddObject() datamember 
// and retrieves the stored objects through GetObject() (remember that 
// you probably will have to make an explicit cast when retrieving the 
// object back from an event). 
//
// $Id: TPhEvent.cxx,v 1.9 1999/02/10 04:31:17 decowski Exp $
//
///////////////////////////////////////////////////////////////////////////

TPhEvent::TPhEvent()
{
// Constructor. Set counter and list data members to zero.
// Don't use this constructor unless you have to and know
// what you are doing
// Use TPhEvent(Int_t run, Int_t event) instead

   fEventHeader = NULL;
   fObjectList  = NULL;
   fFineVertex  = NULL;
   fCoarseVertex = NULL;
   fVerbose = 0;
}

TPhEvent::TPhEvent(Int_t run, Int_t event)
{
// Constructor. Create the event header and the hash table
// for storing the data objects for this event. Set the 
// event name.

   fEventHeader = new TPhEventHeader(run,event);
   fObjectList  = new THashTable();
   fFineVertex  = NULL;
   fCoarseVertex = NULL;
   fVerbose = 0;

   Char_t txt[512];
   sprintf(txt,"event_%d_%d",run,event);

   SetName(txt);
   sprintf(txt,"run %d, event %d",run,event);
   SetTitle(txt);
}

TPhEvent::~TPhEvent()
{
// Destructor. Delete TPhEvent and all the data objects
// currently owned by TPhEvent

  if(fEventHeader)
    delete fEventHeader;

  if(fObjectList) {
    fObjectList->Delete();
    delete fObjectList;
  }

  // Since we deleted all the hitarrays that the sensors could be pointing
  // to (after being filled with TPhDetector::FillWithEventHitArrays), we
  // must clear the sensors to prevent any dangling pointers!
  TPhDetector *det = TPhDetector::GetInstance();
  if(det) {
    det->ClearHitArrays();
    det->ClearHits();
  }

  // Delete the histograms used to calculate the vertex
  SafeDelete(fFineVertex);
  SafeDelete(fCoarseVertex);
}

Int_t TPhEvent::AddObject(TPhDataObject *Object)
{
// Adds an object to the event. 
//
// Return: number of entries already added to the event

  fObjectList->Add(Object);
  if(fVerbose) {
    fprintf(stderr,"Add data Object %s, total # of Objects %d\n",
	    Object->GetName(),
	    fObjectList->GetSize());
  }
  return fObjectList->GetSize();
}

void TPhEvent::DisplayVertex()
{
  // Display the vertex histograms if they have been calculated

  TCanvas *c1;

  TF1 *fcn = fFineVertex->GetFunction("VertexFitFcn");
  if(!fFineVertex || !fCoarseVertex || !fcn)
    return;

  // Set some colors to the histos and funcs 
  TStyle OldStyle;
  gStyle->Copy(OldStyle);

  gStyle->SetHistFillColor(4);
  gStyle->SetStatH(0.40);
  gStyle->SetStatW(0.25);

  // Use an existing canvas if it was already declared earlier
  if((c1 = (TCanvas*) gROOT->FindObject("VertexFitCanvas")) == NULL)
    c1 = new TCanvas("VertexFitCanvas", "Vertex", 500, 700);
  c1->Clear();
  c1->Divide(1, 2);

  c1->cd(1);
  fCoarseVertex->UseCurrentStyle();
  fCoarseVertex->Draw();
    
  fcn->SetLineColor(6);
  fcn->SetLineWidth(4);

  if(fcn) 
    fcn->Draw("LSame");

  c1->cd(2);
  gStyle->SetOptFit(1);
  fFineVertex->UseCurrentStyle();
  fFineVertex->Draw();
  if(fcn)
    fcn->Draw("LSame");

  c1->Update();
  //  gPad->Update();

  OldStyle.Copy(*gStyle);
}

TPhVertex* TPhEvent::FindVertex(Bool_t Force)
{
  // Find the main vertex in the event. 
  // The Z-vertex is found by extrapolating all the combinations of 
  // the different hist into a histogram and then fitting.
  //
  // The Vertex is appended to the event by the name "MainVertex"
  // 
  // Return: the found TPhVertex 
  //         NULL is only returned when an error occured that has 
  //         nothing to do with the vertex

  TPhDetector *det;
  TPhVertex *Vtx;
  TPhHitArray *hi, *ho;

  Char_t Sensor[10];
  Int_t outV, outCol, inCol;
  Int_t inVert, outVert;
  Int_t inRow, outRow;
  Double_t Zv, Xo, Yo, Zo, Xi, Yi, Zi;
  Float_t LargestChi2;

  // If we already have a main vertex in the event and we don't force
  // the recalculation then simply return the vertex already found
  if((Vtx = (TPhVertex*)GetObject("MainVertex")) && Force == kFALSE)
    return Vtx;

  if((det = TPhDetector::GetInstance()) == NULL) {
    Error("FindVertex", "No Detector connected!");
    return NULL;
  }

  // Book the histograms that will be used for keeping track of all the
  // combinations after first deleting any previously used ones.

  fFineVertex = new TH1F("FineVertex", "200-micron bins", 2400, -24., 24.);
  fCoarseVertex = new TH1F("CoarseVertex", "400-micron bins", 1200, -24., 24.);
  fFineVertex->SetXTitle("Z (cm)");
  fCoarseVertex->SetXTitle("Z (cm)");

  // Loop over all 4 inner vertex Sensors
  for (inVert = 0; inVert < 4; inVert++) {
    sprintf (Sensor, "VTA%d", inVert);
    TPhVTXSensor* inVtx = (TPhVTXSensor*) det->GetSensor(Sensor);
    hi = inVtx->GetHitArray();

    if(!hi) {
      if(fVerbose)
	Warning("FindVertex", "%s does not have a HitArray!", Sensor);
      continue;
    }

    // Loop over 4 rows of inner vertex Sensor
    for (inRow = 0; inRow < hi->GetRows(); inRow++){
      
      // Loop over 4 outer vertex Sensors
      for (outVert = 0; outVert < 4; outVert++) {
	// We only make the combinations that could be in a straight line
	// trajectory from the inner to outer vertex
	if(inRow < 2 ) {
	  outV = outVert;
	  outRow = inRow;
	}
	else {
	  outV = outVert + 4;
	  outRow = inRow - 2;
	}
       	sprintf (Sensor, "VTB%d", outV);

	TPhVTXSensor* outVtx = (TPhVTXSensor*)det->GetSensor(Sensor);
	ho = outVtx->GetHitArray();
	if(!ho) {
	  if(fVerbose)
	    Warning("FindVertex", "%s does not have a HitArray!", Sensor);
	  continue;
	}
	//  Loop over all hit combinations
	for (outCol = 0; outCol < ho->GetColumns(); outCol++) {
	  if(ho->GetADC(outRow, outCol) > 0) {
	    outVtx->GetCoords(outRow, outCol, &Xo, &Yo, &Zo);
	    for(inCol = 0; inCol < hi->GetColumns(); inCol++) {
	      if(hi->GetADC(inRow, inCol) > 0) {
		inVtx->GetCoords(inRow, inCol, &Xi, &Yi, &Zi);
		Zv = Zo - Yo*(Zo - Zi)/(Yo - Yi);
		fFineVertex->Fill(Zv);		
		fCoarseVertex->Fill(Zv);		
	      } //if(hi-> ...
	    } //for(inCol = ....
	  }  //if(ho-> ...
	} //for(outCol ...
      } //for(outVert ...
    } //for(inRow ...
  } //for(invert ...

  Float_t MaximumBin, maximum, value;
  Int_t binx;

  // Create the new vertex
  Vtx = new TPhVertex("MainVertex");
  AddObject(Vtx);

  // if we have less than two entries, then we don't have a valid vertex
  if(fFineVertex->GetEntries() < 2) {
    Vtx->SetStatus(TPhVertex::kStatSmallVtx);
    return Vtx;
  }

  // Find the Z value with the highest count
  maximum = 0;
  for(binx = 1; binx <= fCoarseVertex->GetNbinsX(); binx++) {
    value = fCoarseVertex->GetBinContent(binx);
    if (value > maximum) {
      maximum = value;
      MaximumBin = fCoarseVertex->GetBinCenter(binx);
    }
  }

  // Fit the histogram with a gauss and a constant
  TF1 *fcn = new TF1("VertexFitFcn", "gaus + pol0(3)");

  // Set up the fit parameters
  fcn->SetParameter(0, fCoarseVertex->GetMaximum());
  fcn->SetParameter(1, MaximumBin);
  fcn->SetParameter(2, 0.05);
  fcn->SetParameter(3, fCoarseVertex->GetMinimum());

  // We only fit in a range of +-2cm around the peak in the histo
  fcn->SetRange(MaximumBin - 2, MaximumBin + 2); 

  // Use the coarse vertex as a seed for the finer vertex 
  fCoarseVertex->Fit("VertexFitFcn", "QNR");
  fCoarseVertex->GetListOfFunctions()->Add(fcn);
  LargestChi2 = fcn->GetChisquare();

  fcn->SetParameter(0, fFineVertex->GetMaximum());
  fcn->SetParameter(3, fFineVertex->GetMinimum());
  fFineVertex->Fit("VertexFitFcn", "QNR");
  LargestChi2 = 
    (LargestChi2 < fcn->GetChisquare()) ? fcn->GetChisquare() : LargestChi2;

  fFineVertex->GetListOfFunctions()->Add(fcn);

  //  Float_t Amp = fcn->GetParameter(0);
  Float_t Z = fcn->GetParameter(1);
  Float_t SigZ = fcn->GetParameter(2);

  // Check if Z lies outside of the range of the vertex detector, it would
  // be very surprising if |Z| > 24, but mark claims that he has seen it!
  if(TMath::Abs(Z) > 24)
    Vtx->SetStatus(TPhVertex::kOutsideRangeVtx);
  else if(fcn->GetNDF() < 1) // number of degrees is very small
    Vtx->SetStatus(TPhVertex::kStatSmallVtx);
  else if(LargestChi2 / fcn->GetNDF() < 0.00001) 
    Vtx->SetStatus(TPhVertex::kBadChi2Vtx); // Bad Chi2
  else {
    // we have a good vertex, so mark it as good!
    Vtx->SetStatus(TPhVertex::kValidVtx);

    // We Set X = 0, Y = 0 for now!
    Vtx->SetX(0);
    Vtx->SetY(0);
    
    // The Z vertex!
    Vtx->SetZ(Z);
    Vtx->SetSigZ(SigZ);
    Vtx->SetPchi2(fcn->GetChisquare()/fcn->GetNDF());
  }
  return Vtx;
}

TPhDataObject *TPhEvent::GetObject(Char_t *Name)
{
// Returns an object in the evebt with name Name

    TPhDataObject *object = NULL; 
    object = (TPhDataObject *) fObjectList->FindObject(Name);
    if(!object && fVerbose)
      Warning("GetObject", "Object %s not found!", Name);
    return object;
}

void TPhEvent::ls(Option_t *option)
{
// List all objects in the event.
// Use option 'C' to list the data containers.
// Use option 'D' to list the data objects in the containers.

    IndentLevel();
    cout << IsA()->GetName() << " Name: " << GetName() << endl;
    TObject::IncreaseDirLevel();

    if(strstr(option, "C") || strstr(option, "c"))
       fObjectList->ls(option);

    if(strstr(option, "D") || strstr(option, "d"))
       fObjectList->ls(option);

    TObject::DecreaseDirLevel();
}

void TPhEvent::Print(Option_t *option)
{
// List all objects in the event.
// Use "number" as an option to indicate how many sublevels should be listed

    printf("\n>>%s<<  %s %s\n\n", IsA()->GetName(), " Name: ", GetName());

    Char_t aoption1[5];
    Int_t level = atoi(option);
    if(level>=0) {
       level -=1;

       sprintf(aoption1,"%4d",level);
    } else return;

    if(level>=0)
       fObjectList->Print(aoption1);
}

Int_t TPhEvent::RemoveObject(TPhDataObject *Object)
{
// Removes an object from the event. 

    fObjectList->Remove(Object);
    if(fVerbose)
    {
        fprintf(stderr,"Remove data Object %s, total # of Objects %d\n",
                Object->GetName(),fObjectList->GetSize());
    }
    return fObjectList->GetSize();
}

//_______________________________________________________________________
void TPhEvent::Streamer(TBuffer &R__b)
{
   // Stream an object of class TPhEvent.


  if (R__b.IsReading()) {
    Version_t R__v = R__b.ReadVersion(); if(R__v) { }
    TPhDataObject::Streamer(R__b);
    R__b >> fVerbose;
    R__b >> fEventHeader;
    R__b >> fObjectList;
  } else {
    R__b.WriteVersion(TPhEvent::IsA());
    TPhDataObject::Streamer(R__b);
    R__b << fVerbose;
    R__b << fEventHeader;
    
// make a temporary hashtable for the objects marked 
// as persistent and store only those in the output 
// buffer

    THashTable *temp = new THashTable();  
    TIter NextObject(fObjectList);
    TPhDataObject *obj;
    while((obj = (TPhDataObject*) NextObject()) != NULL) {
      if(obj->IsPersistent()) 
	temp->Add(obj);
    }
    R__b << temp;
    temp->Clear();
    delete temp;
  }
}

