#include "TPhHitArray.h"

#include "TPhSensor.h"

#include "TH2.h"

ClassImp(TPhHitArray)


///////////////////////////////////////////////////////////////////////
//                                                                   //
//    TPhHitArray                                                    //
//                                                                   //
//    PHOBOS container class for unpacked hits                       //
//                                                                   //
//    TPhHitArray is the container class for the PHOBOS              //
//    unpacked hit data for one PHOBOS sensor.                       //
//                                                                   //
//    It carries ADC information for now, but it could be further    //
//    extended to carry Energy deposition as well.                   // 
//                                                                   //
//                                                                   //
// $Id: TPhHitArray.cxx,v 1.7 1999/02/24 23:37:16 olszewsk Exp $ 
//
///////////////////////////////////////////////////////////////////////


TPhHitArray::TPhHitArray()
{
// Default constructor. Does nothing.
// Don't use this constructor unless you have to and know
// what you are doing
// Use TPhHitArray(Char_t *name) instead.

    fHits    = 0;
    fRows    = 0;
    fColumns = 0;
    fADC      = NULL;

    sprintf(fSensorName,"%s","");
}

TPhHitArray::TPhHitArray(const Char_t *Name, const Char_t *Title)
{
// Constructor. Create the object setting the name
// (and title, if supplied)

   SetName(Name);
   if(Title)
       SetTitle(Title);
   else
       SetTitle(Name);

    fHits    = 0;
    fRows    = 0;
    fColumns = 0;
    fADC      = NULL;

    sprintf(fSensorName,"%s","");
}

TPhHitArray::TPhHitArray(TPhSensor *Sensor, const Char_t *Name, 
			 const Char_t *Title)
{
// Constructor. Create the data container setting the name
// (and title, if supplied)

    SetName(Name);
    if(Title)
        SetTitle(Title);
    else
        SetTitle(Name);
   
    fHits    = 0;
    MakeArrays(Sensor);
}

TPhHitArray::~TPhHitArray()
{
// Destructor. Delete TPhHitArray and all the data objects
// currently owned by TPhHitArray

  if(fADC)
    delete[] fADC;

  SafeDelete(fHist);
}

void TPhHitArray::MakeArrays(TPhSensor *Sensor)
{
  // Allocates the array needed to store the ADC information. 
    fRows     = Sensor->GetRows();
    fColumns  = Sensor->GetColumns();

    fADC      = new Short_t[fRows*fColumns];

    sprintf(fSensorName,"%s",Sensor->GetName());
}

void TPhHitArray::DrawHistogram(Option_t *Option)
{
  // Draw a 2D histogram with the ADC values in the current
  // canvas (pointed to by gPad)

  Int_t Row, Col;

  // Create the histogram
  // We maintain the convention that the rows are the dimension with the 
  // smaller number of elements. 
  TH2S *fHist = new TH2S(GetName(), GetName(), 
			 GetColumns(), -0.5, GetColumns() - 0.5,
			 GetRows(), -0.5, GetRows() - 0.5);


  for(Row = 0; Row < GetRows(); Row++)
    for(Col = 0; Col < GetColumns(); Col++)
      fHist->Fill(Col, Row, GetADC(Row, Col));

  fHist->Draw(Option);
}

Short_t TPhHitArray::GetADC(Int_t row, Int_t column)
{
//mpd Return -1 if the index is out of bounds.
    Int_t i = GetIndex(row,column);

    if(i >= 0)
      return fADC[i];
    else 
      return -1;
}

void  TPhHitArray::SetADC(Int_t row, Int_t column, Int_t val)
{
    Int_t i = GetIndex(row,column);
    if(i >= 0)
        fADC[i] = val;
}

void  TPhHitArray::CalcNHits()
{
  Int_t count = 0;
  for( Int_t i=0; i<fRows*fColumns; i++) if( fADC[i] ) count++;
  
  fHits = count;
  return;
}

Int_t TPhHitArray::GetIndex(Int_t row, Int_t column)
{
// Return -1 if the index is out of bounds.

    if(row < 0 || column < 0 ||
       row >= fRows || column >= fColumns)
        return -1;
    else if(!fADC)
        return -1;
    else if(fRows == 0  || fColumns == 0)
        return -1;
    else
        return row*fColumns + column;
    
}

void TPhHitArray::Print(Option_t *option)
{
   // List the non zero array elements

  Int_t level = atoi((const char*) option);

  if(level>=0) level -=1; else return;

  printf("Hit array: sensor name %6s , %5d rows %5d columns %5d hits\n", 
               fSensorName, fRows, fColumns, fHits);
  if(level>=0) {
    Int_t i, j;
    for(i=0; i<fRows; i++) {
      for(j=0; j<fColumns; j++) {
          if( fADC[i*fColumns+j]!=0 ) 
              printf(" Row: %5d, Column: %5d, Value: %5hd\n", 
		     i, j, fADC[i*fColumns+j] );
      }
    }
  }
}

//____________________________________________________________________
void TPhHitArray::Streamer(TBuffer &R__b)
{
  // Stream an object of class TPhHitArray.

  if (R__b.IsReading()) {
    Version_t R__v = R__b.ReadVersion(); if(R__v) { } 
    TPhDataObject::Streamer(R__b);
    R__b.ReadStaticArray(fSensorName);
    R__b >> fHits;
    R__b >> fRows;
    R__b >> fColumns;
    R__b.ReadArray(fADC);
  } else {
    R__b.WriteVersion(TPhHitArray::IsA());
    TPhDataObject::Streamer(R__b);
    R__b.WriteArray(fSensorName, 32);
    R__b << fHits;
    R__b << fRows;
    R__b << fColumns;
    R__b.WriteArray(fADC, fRows*fColumns);
  }
}

