#include "TPhRawHitFECn0.h"
#include "TPhFEC.h"
#include <string.h>

ClassImp(TPhRawHitFECn0)
/////////////////////////////////////////////////////////////////////////
//
// Implementation of the NON zero-suppressed data.
//
// The data looks like the following (as of March 23, 1998):
//
// |-----trailer info--------|-------------data--------------------|
//
// -------------------     -----     -----             -----     --- ---
// | 1 | 1 | 1 | 1 | 1 ... 1 | 0 ... 0 | 0 ........... 0 | 0 ... 0 |  |
// -------------------     -----     -----             -----     ---  |
// |   |   |   |   |         |                                     |  |
// | T |   |   |   |  16x    |     3bit String Number              |  |
// | O |   |   |   |  7bit   |                                     |  |
// | T |   |   |   |         |---    ----- ........... ------    --|  |
// | A |   |   |   | Analog  |                                     |  |
// | L |   |   |---|  Values |                                     |  |
// |   |   |   |Mo |         |                                     |  |
// | # |   |   |de |         |                                     | 16-bit
// |   |--------------     ---          8 Strings                  |  |
// | o | S | F |   |         |                                     |  |
// | f | h | I | F |  8bit   |      each ADC value is 12bits       |  |
// |   | i | F | E |         |                                     |  |
// | W | f | O | C | Analog  |                                     |  |
// | O | t |   |   |  Values |                                     |  |
// | R | E | E | # |         |                                     |  |
// | D | r | r |   |         |                                     |  |
// | S | r | r |   |         |                                     |  |
// -------------------     ---     ---     --     --     ---     --- ---
//
//                                                        Start ---^
//
// Created by: Patrick Decowski
//
// $Id: TPhRawHitFECn0.cxx,v 1.9 1999/02/24 23:38:42 olszewsk Exp $
//
///////////////////////////////////////////////////////////////////////

TPhRawHitFECn0::TPhRawHitFECn0()
{
  // Default constructor

}

TPhRawHitFECn0::TPhRawHitFECn0(Int_t FECNum) 
  : TPhAbsRawHitFEC(FECNum)
{
  // Constructor when generating fake data
  // 
  // FOR FAKE DATA GENERATION!
  
  SetTitle("Non zero suppressed FEC buffer");
}

TPhRawHitFECn0::~TPhRawHitFECn0()
{
  // Default destructor

  if(fRawHitsPacked)
    delete[] fRawHitsPacked;

}

void TPhRawHitFECn0::PrintPacked(Option_t * /* option */)
{
  // Print some information about the RawBuffer
  
  Int_t String;
  Int_t i;

  CalcTotalWords();
  
  // Prints the packed hits with U for upper byte of the word and L for 
  // lower byte of the word 
  printf("FEC# %d\n Packed Hits:\n", GetFECNum());
  for(i = 0; i < fTotalWords; i += 2) 
    printf("%d: U = %x L = %x\n", i, 
	   fRawHitsPacked[i], fRawHitsPacked[i+1]);

  printf("=============================\n Allocation:\n");
  for(String = 0; String < kNumStrings; String++) 
    printf("String# %d: Buf start @ %p, Lenght = %d\n", String, 
	   fRawHits[String], fRawHitBufSize[String]);
}



void TPhRawHitFECn0::PrintPacked(Option_t *option, Int_t Line)
{
  // Print some information about the RawBuffer
  
  CalcTotalWords();
  
  TString opt = option;
  opt.ToLower();
  if (opt.Contains("line")){
    if(Line < fTotalWords){
      printf("FEC# %d line %d Packed Hits:     ", GetFECNum(), Line );
      printf("U = %x L = %x\n", fRawHitsPacked[Line], fRawHitsPacked[Line+1]);
    }
  }   
}



void TPhRawHitFECn0::Pack()
{
  // Pack the array into the 16-bit string that is coming out of the
  // FEC.
  //
  // FOR FAKE DATA GENERATION!

  Int_t String, Ch;

  CalcTotalWords();

  // Reserve the memory for the packed buffer after testing if 
  // memory has been allocated before. We want to delete the memory, since
  // the number of words might have changed. 
  if(fRawHitsPacked)
    delete[] fRawHitsPacked;
  fRawHitsPacked = new UChar_t[2*fTotalWords];
  
  Int_t Counter = 0;
  UChar_t PackedUpperByte;
  UChar_t PackedLowerByte;

  // Pack the 8 strings first
  for(String = 0; String < kNumStrings; String++) {
    if(fRawHits[String] != NULL) {
      for(Ch = 0; Ch < fRawHitBufSize[String]; Ch++) {
	// Pack the string number and upper 4 bits of ADC into upper word
	PackedUpperByte = ((String << 4) 
			   | (fRawHits[String][Ch].ADC >> 8) & 0x0f) & 0x7f;
	// Lower 8 bits of ADC go into lower word
	PackedLowerByte = fRawHits[String][Ch].ADC & 0x00ff;
	fRawHitsPacked[Counter++] = PackedUpperByte;
	fRawHitsPacked[Counter++] = PackedLowerByte;

#ifdef DEBUG
	printf("String = %x ADC = %x ;; U = %x, L = %x\n",
	       String, fRawHits[String][Ch].ADC, PackedUpperByte, 
	       PackedLowerByte);
#endif

      }
    }
  }

  // Total number of words in the FEC (including trailer)
  fTrailer[kTotalWordsPos] = fTotalWords;

  // Pack the trailer
  Int_t i;
  for(i = 0; i < kTrailerLength; i++) {
    // Set highest bit
    SETBIT(fTrailer[i], 15);
    fRawHitsPacked[Counter++] = (fTrailer[i] & 0xff00) >> 8;
    fRawHitsPacked[Counter++] = fTrailer[i] & 0x00ff;
  }

}

void TPhRawHitFECn0::UnPack()
{
  // Unpack the Packed Hits as they are coming out of the FEC into
  // the hit buffers

  Int_t i; 
  Int_t TrailerIndex = 0;
  Int_t CountCh = 0;

  Int_t String = 0;
  Int_t OldString = 0;

  UChar_t PackedUpperByte;
  UChar_t PackedLowerByte;

  Int_t ADC;
  Int_t Counter = 0;

  // Check fRawHitsPacked pointer
  if( !fRawHitsPacked ) {
    Warning("Unpack","no buffer with packed hits, cannot unpack!\n");
    return;
  }

  // Loop over all entries in the Packed buffer
  for(i = 0; i < fTotalWords; i++) {
    PackedUpperByte = fRawHitsPacked[Counter++];
    PackedLowerByte = fRawHitsPacked[Counter++];
    // Test the upper bit to see if we are dealing with the trailer or
    // with the data in the strings
    if(TESTBIT(PackedUpperByte, 7)) {
      // This is a trailer
      fTrailer[TrailerIndex++] = (PackedUpperByte << 8) + PackedLowerByte;
      fRawHitBufSize[OldString] = CountCh;
    }
    else {
      // This is sensor data
      String = (PackedUpperByte >> 4) & 0x07;
      if(String != OldString) {
	// save the length of the previous buffer
	fRawHitBufSize[OldString] = CountCh;
	CountCh = 0;
      }

      if(fRawHits[String] == NULL)
	// we make the length equal to the longest string and keep track of
	// the size in fRawHitBufSize (this saves copying the whole string)
      	fRawHits[String] = new FECRawHit_t[kMaxLength];

      fRawHits[String][CountCh].Channel = CountCh;
      ADC = ((PackedUpperByte << 8) + PackedLowerByte) & 0x0fff;
      fRawHits[String][CountCh].ADC = ADC;
      CountCh++;
    }
    
    OldString = String;
  }

}

void TPhRawHitFECn0::Streamer(TBuffer &R__b)
{
  // Stream an object of class TPhRawHitFECn0.
  // 
  // This Streamer was modified by hand!

  if (R__b.IsReading()) {
    Version_t R__v = R__b.ReadVersion(); if (R__v) { }
    TPhAbsRawHitFEC::Streamer(R__b);

    // Read the packed data in and unpack it
    R__b >> fTotalWords;

    fRawHitsPacked = new UChar_t[2*fTotalWords];
    R__b.ReadFastArray(fRawHitsPacked, 2*fTotalWords);
    UnPack();
#ifdef DEBUG
Int_t i;
for(i = 2*(fTotalWords - 20); i < 2*fTotalWords; i += 2)
printf("%x %x ", fRawHitsPacked[i], fRawHitsPacked[i+1]);
printf("\n");
#endif //DEBUG
  } else {
    R__b.WriteVersion(TPhRawHitFECn0::IsA());
    TPhAbsRawHitFEC::Streamer(R__b);

    // Pack the buffers and write them out
    Pack();
    R__b << fTotalWords;
    R__b.WriteFastArray(fRawHitsPacked, 2*fTotalWords);
#ifdef DEBUG
for(i = 2*(fTotalWords - 20); i < 2*fTotalWords; i += 2)
printf("%x %x ", fRawHitsPacked[i], fRawHitsPacked[i+1]);
printf("\n");
#endif //DEBUG

  }
}
