#include "TPhRawHitFECs0.h"
#include <string.h>

ClassImp(TPhRawHitFECs0)
/////////////////////////////////////////////////////////////////////////
//
// Implementation of the zero-suppressed data.
//
// This currently implements the following format (P. Sarin: Nov. 28, '98)
//
//
// |--CMN block--|-------------data-------------------|
//
// ---         -----            ---------       -----------  ---
// | 0 ....... 0 | 0 .......... 0 | 1 | 0 ..... 0 | 1 | 1 |   |
// -----------------            ---------       -----------  2-bit
// | 0 ....... 0 | 1 .......... 1 | 0 | 1 ..... 1 | 0 | 1 |   |
// ---         -----            ---------       -----------  ---
// |             |                |Str|           |Str|   |   |
// |             |                | # |           | # | F |   |
// |             |                |   |           |   | E |   |
// |  32x        |   Up to        |----  10bit    |---- C |   |
// |  11bit      |     8 Strings  | # |  Channel  | # |   |   |
// |   CMN Val   |                | o |   Number  | o | # |   |
// |             |                | f |           | f |---|   |
// |             |                |   |           |   |   |   |
// |             |                | W |           | W | I |   |
// |             |                | O |           | O | n |   |
// |             |-- .......... --| R |-- ..... --| R | f |   |
// |-- ....... --|                | D |           | D | o |  22-bit
// |             |                | S |           | S |   |   |
// |             |                |   |           |   | W |   |
// |             |                | t |           | t | o |   |
// |             |                | o |           | o | r |   |
// |  32x        |                |   | 12bit     |   | d |   |
// |  11bit      |                | f |  ADC      | f |   |   |
// |   CMN Value |                | o |   Value   | o | 1 |   |
// |             |                | l |           | l | 6 |   |
// |             |                | l |           | l | b |   |
// |             |                | o |           | o | i |   |
// |             |                | w |           | w | t |   |
// --- ....... ----- .......... --------- ..... -----------  ---
//
//                                               Start ---^
//
// Created by: Patrick Decowski
//
// $Id: TPhRawHitFECs0.cxx,v 1.3 1999/02/24 23:39:13 olszewsk Exp $
//
///////////////////////////////////////////////////////////////////////

TPhRawHitFECs0::TPhRawHitFECs0()
{
  // Default constructor

}

TPhRawHitFECs0::TPhRawHitFECs0(Int_t FECNum) : TPhAbsRawHitFEC(FECNum)
{
  // Constructor when generating fake data
  // 
  // FOR FAKE DATA GENERATION!
  
  Int_t i;
  for(i = 0; i < kCMNLength; i++)
    fCMN[i] = 0;

  SetTitle("ZERO suppressed FEC buffer");
}

TPhRawHitFECs0::~TPhRawHitFECs0()
{
  // Default destructor

  if(fZeroSupPacked)
    delete[] fZeroSupPacked;

}

void TPhRawHitFECs0::PrintPacked(Option_t * /* option */) const
{
  // Print some information about the RawBuffer
  
  Int_t String;
  Int_t i;
  Char_t Buffer1[50], Buffer2[50], Buffer3[50];

  if(!fZeroSupPacked) {
    Error("Print", "No Zero Suppressed data structure!");
    return;
  }

  Int_t Length = CalcTotal24Words();

  // Prints the packed hits with U for upper byte of the word and L for 
  // lower byte of the word 
  printf("FEC# %d\n Packed Hits:\n", GetFECNum());
  for(i = 0; i < 3 * Length; i += 3) {
    printf("%5d: U = 0x%2.2x M = 0x%2.2x L = 0x%2.2x -> "\
	   "%%%s | %%%s | %%%s\n", i, 
	   fZeroSupPacked[i], fZeroSupPacked[i+1], fZeroSupPacked[i+2],
	   Conv2Base(fZeroSupPacked[i], Buffer1), 
	   Conv2Base(fZeroSupPacked[i+1], Buffer2), 
	   Conv2Base(fZeroSupPacked[i+2], Buffer3));
  }
  printf("=============================\n Allocation:\n");
  for(String = 0; String < kNumStrings; String++) 
    printf("String# %d: Buf start @ %p, Lenght = %d\n", String, 
	   fRawHits[String], fRawHitBufSize[String]);
}

Int_t TPhRawHitFECs0::CalcTotal24Words() const
{
  // Calculate the number of 24-bit words that need to be stored

  Int_t String;
  Int_t Length = 1; // One for the FEC etc. info

  // figure out what the total number of words is.
  for(String = 0; String < kNumStrings; String++) {
    if(fRawHits[String] != NULL) {
      Length += fRawHitBufSize[String];
      Length++; // Every String has a header
    }
  }
  Length += kCMNLength;
  //mpd FIXME: is this a problem: We ignore the FEC Trailer at the moment!!!! 
  return Length;
}
 
void TPhRawHitFECs0::Pack()
{
  // Pack the array into the 24-bit string that is coming out of the
  // FEC.
  //

  Int_t String, Pos;
  Char_t TempBuffer[20];

  // Reserve the memory for the packed buffer after testing if 
  // memory has been allocated before. We want to delete the memory, since
  // the number of words might have changed. 
  if(fZeroSupPacked)
    delete[] fZeroSupPacked;

  fZeroSupPacked = new UChar_t[3 * CalcTotal24Words()];
  
  Int_t Counter = 0;

  // Make sure that the name reflects the FEC number
  SetName(MakeFECName(fFECNumber, TempBuffer));

  // Pack the FEC header
  fZeroSupPacked[Counter++] = 0xc0 | (fFECNumber & 0x3f); 
  //mpd FIXME: what is the info word???? 
  fZeroSupPacked[Counter++] = (fInfoWord & 0xff00) >> 8; 
  fZeroSupPacked[Counter++] = fInfoWord & 0x00ff; 

  // Pack the 8 strings
  for(String = 0; String < kNumStrings; String++) {
    if(fRawHits[String] != NULL) {
      // Pack the String header
//mpd FIXME: passing zero in bits 17 and 16 of size - is it really that big???
      fZeroSupPacked[Counter++] = 
	0x80 |  (String << 2) | (fRawHitBufSize[String] & 0x0);
      fZeroSupPacked[Counter++] = (fRawHitBufSize[String] & 0xff00) >> 8; 
      fZeroSupPacked[Counter++] = fRawHitBufSize[String] & 0x00ff; 
      
      for(Pos = 0; Pos < fRawHitBufSize[String]; Pos++) {
	fZeroSupPacked[Counter++] =
	  0x40	 | ((fRawHits[String][Pos].Channel & 0x03f0) >> 4);
	fZeroSupPacked[Counter++] = 
	  ((fRawHits[String][Pos].Channel & 0x000f) << 4)
	  | ((fRawHits[String][Pos].ADC & 0x0f00) >> 8);
	fZeroSupPacked[Counter++] = fRawHits[String][Pos].ADC & 0x00ff;

#ifdef DEBUG
	printf("String = %x Ch = %d: ADC = %x ;; U = %x, M = %x, L = %x\n",
	       String, fRawHits[String][Ch].Ch, fRawHits[String][Ch].ADC,  
	       fZeroSupPacked[Counter-3], fZeroSupPacked[Counter-2], 
	       fZeroSupPacked[Counter-1]);
#endif

      }
    }
  }

  Int_t i = 0;
  Int_t CMN1, CMN2;
  for(Pos = 0; Pos < kCMNLength; Pos++) {
    CMN1 =fCMN[i++];
    CMN2 =fCMN[i++];
    fZeroSupPacked[Counter++] = CMN1 & 0xe0 >> 5;
    fZeroSupPacked[Counter++] = (CMN1 & 0x1f << 3) | (CMN2 & 0x0700 >> 8);
    fZeroSupPacked[Counter++] = (CMN2 & 0xff);
  }
}

void TPhRawHitFECs0::UnPack(const UChar_t *ZeroSupBuffer, const Int_t Length)
{
  // Unpack the Packed Hits as they are coming out of the FEC into
  // the hit buffers

  Int_t String = 0;
  Int_t StringSize = 0;

  UChar_t PackedUpperByte;
  UChar_t PackedMiddleByte;
  UChar_t PackedLowerByte;

  Int_t ADC;
  Int_t Ch = 0;
  Int_t Pos = 0;
  Int_t Counter = 0;

  // Check fZeroSupPacked pointer
  if( !ZeroSupBuffer ) {
    Warning("Unpack","no buffer with packed hits, cannot unpack!\n");
    return;
  }
  if(Length < 6) {
    Error("Unpack", "Length of buffer too small!");
    return;
  }

  PackedUpperByte = ZeroSupBuffer[Counter++];
  if(!(TESTBIT(PackedUpperByte, 7) && TESTBIT(PackedUpperByte, 6))) {
    Error("UnPack", "Wrong FEC header!");
    return;
  }
  fFECNumber = PackedUpperByte & 0x3f;
  
  PackedMiddleByte = ZeroSupBuffer[Counter++];
  PackedLowerByte = ZeroSupBuffer[Counter++];

  fInfoWord = ((PackedMiddleByte << 8) + PackedLowerByte) & 0x0fff;

  PackedUpperByte = ZeroSupBuffer[Counter++];
  PackedMiddleByte = ZeroSupBuffer[Counter++];
  PackedLowerByte = ZeroSupBuffer[Counter++];

  // Check that the upper byte starts with %10xxxxxx, which indicates that
  // this is a String header 

  while(TESTBIT(PackedUpperByte, 7) 
	&& !TESTBIT(PackedUpperByte, 6)) {
    String = (PackedUpperByte & 0x3c) >> 2;
    //mpd FIXME: ignoring bit 16 & 17 of size again!
    StringSize = ((PackedMiddleByte << 8) + PackedLowerByte) & 0xffff;
#ifdef DEBUG
printf("string %d size %d\n", String, StringSize);
#endif
    fRawHitBufSize[String] = StringSize;
    fRawHits[String] = new FECRawHit_t[StringSize];

    // Loop over all entries in the string
    for(Pos = 0; Pos < StringSize; Pos++) {
      if(Counter > Length - 6) {
	Error("Unpack", "Counter=%d > Length=%d", Counter, Length - 6);
	return;
      }
      PackedUpperByte = ZeroSupBuffer[Counter++];
      PackedMiddleByte = ZeroSupBuffer[Counter++];
      PackedLowerByte = ZeroSupBuffer[Counter++];

      // Make sure that the data makes sense
      if( !(!TESTBIT(PackedUpperByte, 7) && TESTBIT(PackedUpperByte, 6))) {
	Warning("UnPack", "Internal data consistency problem!");
	return;
      }
      Ch = ((PackedUpperByte & 0x3f) << 4) + ((PackedMiddleByte & 0xf0) >> 4);
      ADC = ((PackedMiddleByte << 8) + PackedLowerByte) & 0x0fff;
      fRawHits[String][Pos].ADC = ADC;
      fRawHits[String][Pos].Channel = Ch;
    }
    PackedUpperByte = ZeroSupBuffer[Counter++];
    PackedMiddleByte = ZeroSupBuffer[Counter++];
    PackedLowerByte = ZeroSupBuffer[Counter++];
  }
 
  Int_t i = 0;
  // Read CMN Block in
  for(Pos = 0; Pos < kCMNLength; Pos++) {
      if( !(!TESTBIT(PackedUpperByte, 7) && !TESTBIT(PackedUpperByte, 6))) {
	Warning("UnPack", "Internal data consistency problem in CMN block!");
	return;
      }
      fCMN[i++] = (PackedUpperByte & 0x3f << 5) 
	+ (PackedMiddleByte & 0xfc >> 5);
      fCMN[i++] = (PackedMiddleByte & 0x03 << 8) + PackedLowerByte;
      // read data in for next pass
      PackedUpperByte = ZeroSupBuffer[Counter++];
      PackedMiddleByte = ZeroSupBuffer[Counter++];
      PackedLowerByte = ZeroSupBuffer[Counter++];
  }

}

void TPhRawHitFECs0::Streamer(TBuffer &R__b)
{
  // Stream an object of class TPhRawHitFECs0.
  // 
  // This Streamer was modified by hand!

  Int_t Total24Words;

  if (R__b.IsReading()) {
    Version_t R__v = R__b.ReadVersion(); if (R__v) { }
    TPhAbsRawHitFEC::Streamer(R__b);

    // Read the packed data in and unpack it
    R__b >> Total24Words;

    fZeroSupPacked = new UChar_t[3*Total24Words];
    R__b.ReadFastArray(fZeroSupPacked, 3*Total24Words);
    UnPack(fZeroSupPacked, 3*Total24Words);
  } else {
    // Pack the buffers and write them out
    Pack();
    R__b.WriteVersion(TPhRawHitFECs0::IsA());
    TPhAbsRawHitFEC::Streamer(R__b);
    Total24Words = CalcTotal24Words();
    R__b << Total24Words;
    R__b.WriteFastArray(fZeroSupPacked, 3*Total24Words);

  }
}
