#include <iostream.h>
#include <stdlib.h>
#include "math.h"
#include "TMath.h"
#include "TPhTrack.h"
#include "TPhHit.h"

ClassImp(TPhTrack)

///////////////////////////////////////////////////////////////////////
//                                                                   //
//    TPhTrack                                                       //
//                                                                   //
//    TPhTrack is the basic PHOBOS track class                       //
//                                                                   //
///////////////////////////////////////////////////////////////////////


TPhTrack::TPhTrack()
{
// Constructor. Set fHitList to NULL.

    fHitList = NULL;
    fXProd = 0;
    fYProd = 0;
    fZProd = 0;
    fMass = 0;
    fGPid = 0;
    fHitCounter = 0;
    fPolyLine = NULL;
}

TPhTrack::~TPhTrack()
{

// Destructor. Clears fHitList if applicable, but does
// not delete TPhHit objects.

  if(fHitList)
    fHitList->Clear();

  delete fHitList;

//mpd delete polyline
    if(fPolyLine)
      delete fPolyLine;
}

void TPhTrack::AddHit(TPhHit *hit)
{
// Add one hit to fHitList, create fHitList if necessary.

    if(!fHitList)
        fHitList = new TObjArray();
    fHitList->Add(hit);
}

void TPhTrack::MakeHitList()
{
// Create fHitList if necessary.

    if(!fHitList)
        fHitList = new TObjArray();
}

Int_t TPhTrack::GetNHits()
{
// Get number of hits on the track if applicable

    if(fHitList)
        return fHitList->GetEntries();
    else
        return 0;
}

Float_t  TPhTrack::GetEnergy()
{
// return track energy, assuming mass fMass.

    return sqrt(GetPx()*GetPx()+GetPy()*GetPy()+GetPz()*GetPz()+fMass*fMass);
}

Float_t TPhTrack::GetRap(Float_t mass)
{
// return track energy, assuming mass given in argument.

    Float_t rap;
    Float_t E;

    E = sqrt(mass*mass + GetPz()*GetPz()+GetPx()*GetPx()+GetPy()*GetPy());
    rap = 0.5 * log((E+GetPz())/(E-GetPz()));

    return rap;
}

Float_t TPhTrack::GetTheta()
{
   Float_t  theta, pxy;

   if(GetPz()==0)
   {
      theta = TMath::Pi() / 2.0;
   }
   else
   {
      pxy = sqrt(GetPx()*GetPx() + GetPy()*GetPy());
      theta = atan(pxy / GetPz());
      if(GetPz() < 0.0)  theta += TMath::Pi();
   }
   return theta;
}

Float_t TPhTrack::GetDip()
{
  Float_t pxz, dip;

   pxz = sqrt(GetPx()*GetPx() + GetPz()*GetPz());

   if(pxz==0)
   {
      dip = TMath::Pi() / 2.0;
      if(GetPy() < 0.0)  dip = -dip;
   } 
   else
      dip = atan(GetPy()/pxz);

   return dip;
}

Float_t TPhTrack::GetBeta()
{
    Float_t p;
    Float_t E;

    p = GetP();
    E = sqrt(p*p + fMass* fMass); 
    return p/E;
}

TPhHit *TPhTrack::GetNextHit()
{
// Iterator over hits on track. Note that this loop
// can NOT be nested. Use the GetHitList() call to
// obtain and manipulate the fHitList ObjArray for 
// more complicated loops.
    
    if(fHitList)
    {
        return (TPhHit *) fHitList->At(fHitCounter++);
    }
    else
    {
        return NULL;
    }
}

void TPhTrack::Draw(Option_t */* Option */)
{
// Draw the track. Note that this function is not standalone,
// but should be used in a context as provided by the 
// TPhDetector::Draw() member function

    Int_t   num;
    TPhHit *hit;

    if(fHitList) {
        if(!fPolyLine) {
            fPolyLine = new TPolyLine3D(GetNHits());
            num = 0;
	    int nhits=fHitList->GetEntries();
	    for (int ih=0; ih<nhits; ih++) {
	      hit = (TPhHit *)fHitList->At(ih);
	      fPolyLine->SetPoint(num,hit->GetX(),hit->GetY(),hit->GetZ());
	      num++;
	    }	
        }
        fPolyLine->SetLineColor(3);
        fPolyLine->Draw();
    }
}


void TPhTrack::ListHits()
{
      if(fHitList)
      {
          TIter NextHit(fHitList);
	  TPhHit *hit;
          while((hit = (TPhHit *) NextHit()) != NULL)
          {
	    printf("Hit in sensor %d layer %d at %g %g %g  "\
		   "pad %d %d  with De %g  Dt %g\n",
		   hit->GetSensorLabel(),
		   hit->GetLayer(),
		   hit->GetX(),
		   hit->GetY(),
		   hit->GetZ(),
		   hit->GetPadX(),
		   hit->GetPadY(),
		   hit->GetDe(),
		   hit->GetDt());
          }
      }
}

void TPhTrack::ls(Option_t * /* option */)
{
  // List track object short info 

  Int_t nhits = 0; 
  if( fHitList ) nhits = fHitList->GetEntries();
  IndentLevel();
  cout << IsA()->GetName() << " Name: " << GetName() << "  " << nhits << " hits." << endl;
}


void TPhTrack::Print(Option_t *option)
{
  // List track object info
  // Use "number" as an option to indicate how many sublevels should be listed

  Int_t level = atoi(option);
  if(level>=0) {
     level -=1;
  } else return;

  Int_t nhits = 0;
  if( fHitList ) nhits = fHitList->GetEntries();
  printf(" - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - \n");
  printf("%s - Track at %6.2f %6.2f %6.2f, Hits %4d\n             px py pz: %8.4e %8.4e %8.4e Gpid %3d mass %7.3f charge %3d\n", 
     IsA()->GetName(),fXProd,fYProd,fZProd, nhits,
     this->GetPx(),this->GetPy(),this->GetPz(),fGPid,fMass,this->GetCharge());
  if( nhits>0 && level>=0 ) fHitList->Print(option);
}


