#ifndef _PhDetector_INCLUDED_
#define _PhDetector_INCLUDED_

///////////////////////////////////////////////////////////////////////
//                                                                   //
//    TPhDetector                                                    //
//                                                                   //
//    PHOBOS detector information class                              //
//                                                                   //
//    TPhDetector manages information and operations related to the  //
//    detector geometry                                              //
//                                                                   //
//    Author  : Gunther Roland (rolandg@vxcern.cern.ch)              //
//    Created : 15/01/96                                             //
//    Version : 1.0                                                  //
//    Changed : 30.09.1997 Andrzej Olszewski                         //
//    Changed : 15.10.1997 Gunther Roland                            //
//              Make TPhDetector a singleton class.                  //
//              Access to single instance via                        //
//              TPhDetector::GetInstance() static member function.   //
//              15.06.98   Andrzej Olszewski                         //
//              Creation of TPhDetector object by calls to AddSensor //
//              Subdetector list made and filled by a class creator  //
//              Friend classes are not needed anymore.               //
//              15.12.98   Andrzej Olszewski                         //
//              New function MakeHits to test pedestal, calibration  //
//              creating hits directly from rawhit objects           //
//                                                                   //
// $Id: TPhDetector.h,v 1.10 1999/03/05 11:30:05 olszewsk Exp $
//                                                                   //
///////////////////////////////////////////////////////////////////////

// ROOT classes

#include "TObject.h"
#include "TNode.h"
#include "TFile.h"
#include "TBRIK.h"
#include "THashTable.h"
#include "THashList.h"
#include "TRotMatrix.h"

// Phobos classes
#include "TPhSubDetector.h"
#include "TPhSensorMapDB.h"
#include "TPhFECDetector.h"

class TPhEvent;
class TPhSensor;
class TPhLayer;
class TPhMagnet;
class TPhTranslate;
class TPhASCIIData;
class TPhHit;
                                                                              
class TPhDetector : public TNamed {

  friend class TPhEvent;

public:

  TPhDetector();
  virtual ~TPhDetector();

    static  TPhDetector  *GetInstance();
    static  TPhDetector  *CreateInstance();
    static  TPhDetector  *GetOrCreateInstance();
    static  void          DeleteInstance();
    static  TPhDetector  *ReadLatestGeometry();

    void            Draw(Option_t *Option = "");
    void            Print() const;
    void            ls(Option_t *Option = "") const;

    void            SetVersion(Float_t version) { fVersion = version;  }
    Float_t         GetVersion() const          { return fVersion;     }

    // Functions that save information on 3 separate building blocks of the detector
    // to a file
    void            WriteOutSensorGeometry();  
    void            WriteOutFECGeometry();  
    void            WriteOutDBMap();  
    void            WriteOutSensorGeometry(Char_t *filename);  
    void            WriteOutFECGeometry(Char_t *filename);  
    void            WriteOutDBMap(Char_t *filename);  

    void            WriteOutSensorGeometry(TFile  *tfile);  
    void            WriteOutFECGeometry(TFile  *tfile);  
    void            WriteOutDBMap(TFile  *tfile);  

    // Functions that read information on 3 separate building blocks of the detector
    Bool_t          ReadInSensorGeometry();  
    Bool_t          ReadInFECGeometry();  
    Bool_t          ReadInCalibration();  
    Bool_t          ReadInDBMap();  
    Bool_t          ReadInSensorGeometry(Char_t *filename);  
    Bool_t          ReadInFECGeometry(Char_t *filename);  
    Bool_t          ReadInCalibration(Char_t *filename);  
    Bool_t          ReadInDBMap(Char_t *filename);  
    Bool_t          ReadInSensorGeometry(TFile  *tfile);  
    Bool_t          ReadInFECGeometry(TFile  *tfile);  
    Bool_t          ReadInDBMap(TFile  *tfile);

    // The only means of building sensor geometry part of the detector 
    // Subdetector structure is hard coded
    Bool_t          AddSensor(Int_t detno, Char_t *detname, Int_t shapeno, 
                       Float_t *center, Int_t nvolpar, Float_t *volpars, 
                       TRotMatrix *rotmatp, Int_t padtype, Float_t *padpars); 

    // Functions building FEC structure of the Phobos electronics
    // 1. As a copy of FEC structure in the sensor map data base,
    //    currently connected to the Detector object
    Bool_t          BuildFECDetector();

    THashTable     *GetSubDetectorList() const 
                       { return fSubDetectorList;   }
    TPhSubDetector *GetSubDetector(const Char_t *Name) const;
    TPhSubDetector *GetSubDetectorOfSensor(const Char_t *Name) const;
    TPhSubDetector *GetSubDetectorOfSensor(const Int_t SensNum) const;
	
    THashList      *GetSensorList() const { return fSensorList;   }
    TPhSensor      *GetSensor(const Int_t i) const;
    TPhSensor      *GetSensor(const Char_t *Name) const;

    TPhSensorMapDB *GetSensorMapDB() const  
                      { return fSensorMapDB;   }

    TPhFECDetector *GetFECDetector() const  
                      { return fFECDetector;   }


    TPhEvent       *GetEvent()  const            { return fEvent;       }
    TPhMagnet      *GetMagnet() const            { return fMagnet;      }

    void            SetEvent(TPhEvent *event)    { fEvent = event;      }
    void            SetMagnet(TPhMagnet *magnet) { fMagnet = magnet;    }

    void            AddHit(TPhHit *Hit);
    void            FillWithEventHits(Char_t *HitConName);
    void            FillWithEventHitArrays(Char_t *HitArrConName);

    Int_t           MakeHitArraysOutOfHits(Char_t *HitConName, 
					   Char_t *HitArrayConName);
    Int_t           MakeHitsOutOfHitArrays(Char_t *HitArrayConName,
					   Char_t *HitConName);
    Int_t           MakeHitArrays(Char_t *RawHitFECConName, 
				  Char_t *HitArrayConName);
    Int_t           MakeHitArrays(Char_t *hitArrayConName);
    void            MakeHitLists(const Char_t *HitsConName);
    Int_t           MakeHits(Char_t *RawHitFECn0ConName, Char_t *HitConName,
    			     Char_t *PedFileName, Char_t *ConFileName);
    Int_t           MakeRawHitFECn0(Char_t *HitArrayConName, 
				    Char_t *RawHitFECn0ConName);
    Bool_t          MakeRawHitFECs0(Char_t *RawHitFECn0ConName, 
				    Char_t *RawHitFECs0ConName);
    Bool_t          MakeCalRawHitFECs0(Char_t *RawHitFECs0ConName, 
				       Char_t *CalRawHitFECs0ConName);

    void            DeleteHits();
    void            ClearHits();

    Bool_t          CMNProcess(Char_t *rawContainerName);
    Bool_t          NoiseProcess(Char_t *rawContainerName);
    Bool_t          PedPreProcess(Char_t *rawContainerName);
    Bool_t          PedProcess(Char_t *rawContainerName);
    Bool_t          ZeroSuppress(Char_t *rawContainerName);
    Bool_t          Calibrate(Char_t *rawContainerName);
    Bool_t          CalibrateOnStrings(Char_t *rawContainerName);

    Bool_t          CompatibilityCheck(Char_t *rawContainerName);
    Bool_t          CompatibilityCheck(TPhRawHitFECn0* rawhit);

    //              Setting/getting parameters for pedestal/noise string routines 
    void            SetMIN_NUM_PED_EVENTS(Int_t value) 
                          { fFECDetector->SetMIN_NUM_PED_EVENTS(value);   }
    void            SetMIPADC_CUTOFF(Int_t value)             
                          { fFECDetector->SetMIPADC_CUTOFF(value);        }
    void            SetMIN_CHANNELS_CMN_EST(Int_t value) 
                          { fFECDetector->SetMIN_CHANNELS_CMN_EST(value); }
    void            SetMIN_NUM_NOISE_EVENTS(Int_t value) 
                          { fFECDetector->SetMIN_NUM_NOISE_EVENTS(value); }
    void            SetTHRESH_CMN(Float_t value)  
                          { fFECDetector->SetTHRESH_CMN(value);    }
    void            SetTHRESH_MULTIPLIER(Float_t value)  
                          { fFECDetector->SetTHRESH_MULTIPLIER(value);    }


    void            RemoveCMNArrays()     { fFECDetector->RemoveCMNArrays(); }
    void            RemoveNoiseArrays()   { fFECDetector->RemoveNoiseArrays(); }
    void            RemovePedArrays()     { fFECDetector->RemovePedArrays(); }
    void            RemoveSuppArrays()    { fFECDetector->RemoveSuppArrays(); }
    void            RemoveZZArrays()      { fFECDetector->RemoveZZArrays(); }

    Int_t           GetVerbose() const { return fVerbose; }
    void            SetVerbose(Int_t verbose) { fVerbose = verbose; }
    void            SetSensorsVerbose(Int_t verbose);

private: 
    Bool_t          BuildSensorList();
    void            SetSensorList(THashList *sl) 
      { fSensorList = sl; }
    void            SetSubDetectorList(THashTable *hs) 
      { fSubDetectorList = hs; }
    void            AddSubDetector(TPhSubDetector *sub)  
                         { fSubDetectorList->Add(sub); }
    void            ClearHitArrays();

    static Bool_t   fgDetIsInitialized;  //  Singleton initialization flag   

    Float_t         fVersion;            //  Geometry version number
    THashTable     *fSubDetectorList;    //  HashTable of all SubDetectors
    THashList      *fSensorList;         //  HashList of all TPhDetector sensors
    TPhFECDetector *fFECDetector;        //! TPhFECDetector object created automatic
    TPhSensorMapDB *fSensorMapDB;        //! TPhSensorMapDB object created automatic
    TPhEvent       *fEvent;              //! PHOBOS event object
    TPhMagnet      *fMagnet;             //! PHOBOS Magnet Object
    TNode          *fNode;               //! Mother Node to Draw the Detector 
    TBRIK          *fBrik;               //! Shape of Mother Node
    Int_t           fVerbose;            //! Verbosity flag

public:

  ClassDef(TPhDetector,1) // The whole PHOBOS geometry

};

#endif



