#ifndef _TPhFECDetector_
#define _TPhFECDetector_

/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//                                                                         //
//    TPhFECDetector                                                       //
//                                                                         //
//    Phobos interface class to properties of Phobos electronics           //
//                                                                         //
//    FECDetector class gives access to the properties of the Phobos       //
//    electronics. It stores a list of TPhFEC objects describing channels  //
//    within strings of a single FEC connector.                            //
//                                                                         //
//    Author  : Andrzej Olszewski (andrzej.olszewski@ifj.edu.pl)           //
//    Created : 28/12/98                                                   //
//    Version : 1.0                                                        //
//    Changed :                                                            //
//                                                                         //
// $Id: TPhFECDetector.h,v 1.2 1999/03/05 11:30:05 olszewsk Exp $
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

// ROOT classes
#include "TNamed.h"
#include "TObjArray.h"

// Phobos classes
#include "TPhFEC.h"
#include "TPhRawHitFECn0.h"
#include "TPhHashContainer.h"

enum EDescriptionFECDetector {
  kMaxFECs =   64         // The (maximum) number of FECs
};

class TPhFECDetector : public TNamed {

public:
  TPhFECDetector();
  TPhFECDetector(Char_t *Name);
  virtual ~TPhFECDetector();

  TPhFEC*    GetFEC(const Int_t i) const;
  TPhString* GetString(const Int_t fecno, const Int_t strno) const;

  Bool_t     AddFEC(TPhFEC *fec);
  Bool_t     IsThere(const Int_t i);
  void       SetVersion(Float_t version) { fVersion = version; }

  Bool_t     CompatibilityCheck(TPhHashContainer *container);
  Bool_t     CompatibilityCheck(TPhRawHitFECn0 *rawhit);

  Bool_t     PedPreProcess(TPhRawHitFECn0 *fecevent);     
  Bool_t     PedProcess(TPhRawHitFECn0 *fecevent);        
  Bool_t     CMNProcess(TPhRawHitFECn0 *fecevent);  
  Bool_t     NoiseProcess(TPhRawHitFECn0 *fecevent);      
  Bool_t     ZeroSuppress(TPhRawHitFECn0 *fecevent);
  void       Calibrate(Bool_t &status);
  Bool_t     Calibrate(TPhAbsRawHitFEC *fecevent);
  Bool_t     CalibrateOnStrings(TPhAbsRawHitFEC *fecevent);

  Bool_t     ReadCalibrationTB98File(Char_t *filename);
  Bool_t     ReadinStringCalibrationTB98File(Char_t *filename);
  Bool_t     CalibrationCheck();

  void       MakeHitArrays(TPhHashContainer *HitArrayCont)
                                 { fFECList->ForEach(TPhFEC,MakeHitArrays)(HitArrayCont); }
  void       DeleteCalChannelArrays()
                                 { fFECList->ForEach(TPhFEC,DeleteCalChannelArrays)(); }
  void       DeleteCalStringObjects()
                                 { fFECList->ForEach(TPhFEC,DeleteCalStringObjects)(); }

  void       RemovePedArrays()   { fFECList->ForEach(TPhFEC,RemovePedArrays)();       }
  void       RemoveCMNArrays()   { fFECList->ForEach(TPhFEC,RemoveCMNArrays)();       }
  void       RemoveNoiseArrays() { fFECList->ForEach(TPhFEC,RemoveNoiseArrays)();     }
  void       RemoveZZArrays()    { fFECList->ForEach(TPhFEC,RemoveZZArrays)();        }
  void       RemoveSuppArrays()  { fFECList->ForEach(TPhFEC,RemoveSuppArrays)();      }

  void       SetDoHistograms(Bool_t onoff) 
                                 { fFECList->ForEach(TPhFEC,SetDoHistograms)(onoff);  }
  void       DeleteHistograms()
                                 { fFECList->ForEach(TPhFEC,DeleteHistograms)();      }

  //         Setting/getting parameters for pedestal/noise string routines 
  void       SetMIN_NUM_PED_EVENTS(Int_t value) 
                          { fFECList->ForEach(TPhFEC,SetMIN_NUM_PED_EVENTS)(value);   }
  void 	     SetMIPADC_CUTOFF(Int_t value)             
                          { fFECList->ForEach(TPhFEC,SetMIPADC_CUTOFF)(value);        }
  void       SetMIN_CHANNELS_CMN_EST(Int_t value) 
                          { fFECList->ForEach(TPhFEC,SetMIN_CHANNELS_CMN_EST)(value); }
  void       SetMIN_NUM_NOISE_EVENTS(Int_t value) 
                          { fFECList->ForEach(TPhFEC,SetMIN_NUM_NOISE_EVENTS)(value); }
  void       SetTHRESH_CMN(Float_t value)  
                          { fFECList->ForEach(TPhFEC,SetTHRESH_CMN)(value);    }
  void       SetTHRESH_MULTIPLIER(Float_t value)  
                          { fFECList->ForEach(TPhFEC,SetTHRESH_MULTIPLIER)(value);    }

  void       Print() const;
  void       ls(Option_t *Option = "") const;

  Int_t      GetVerbose() const        { return fVerbose; }
  void       SetVerbose(Int_t verbose) { fVerbose = verbose; }
  void       SetVerboseDown(Int_t verbose) 
                                 { fVerbose = verbose;
				   fFECList->ForEach(TPhFEC,SetVerboseDown)(verbose); }

private:
  Float_t      fVersion;     // Version of the FEC structure
  TObjArray   *fFECList;     // List of all the FEC objects                 
  TObjArray   *fStringList;  //! (temporary) list of all the String objects
			     //  not filled at them moment
  Int_t        fVerbose;     //! Verbose flag

public:

  ClassDef(TPhFECDetector,1)  // SubDetectors containing detector layers

};

#endif // _TPhFECDetector_
