#include "TPhBoxSensor.h"

#include "TMath.h"
#include "TGeometry.h"

#include <iostream.h>

ClassImp(TPhBoxSensor)
//////////////////////////////////////////////////////////////////////////////
// TPhBoxSensor is the base class for all non-Si BOX shaped sensors such 
// as TPhTOFSensor. 
// 
// The Si detectors have their own counterpart with TPhBoxSiSensor. 
// This division had to be made because of inheritance rules. 

TPhBoxSensor::TPhBoxSensor()
{
  // Default constructor. 

  fX0 = 0;
  fX1 = 0;
  fY0 = 0;
  fY1 = 0;
  fZ0 = 0;
  fZ1 = 0;
  fAlpha = 0;
}

TPhBoxSensor::TPhBoxSensor(const Char_t* Name, Int_t Label, Int_t Layer, 
			   Double_t X0, Double_t X1, Double_t Y0, Double_t Y1, 
			   Double_t Z0, Double_t Z1)
{
  // Normal constructor. 

  SetName(Name);
  SetLabel(Label);
  SetLayer(Layer);

  fX0 = X0;
  fX1 = X1;
  fY0 = Y0;
  fY1 = Y1;
  fZ0 = Z0;
  fZ1 = Z1;
}

TPhBoxSensor::~TPhBoxSensor()
{
  // Destructor
  
  SafeDelete(fShape);
}

void TPhBoxSensor::SetDim(Float_t dx, Float_t dy, Float_t dz)
{
  // Set the half-dimensions (a la GEANT) of the Box volume. 

  Char_t name[40];
  sprintf(name, "Brik_%s", GetName());
  fShape = new TBRIK(name, name, "", dx, dy, dz);
  fDx = dx; fDy = dy; fDz = dz;
}

void TPhBoxSensor::CalcBoxPoints()
{
  // Calculates the outer points of the sensor. These are the points 
  // that can later be retrieved by GetX0() etc. It also sets the angle 
  // between the beam pipe and the normal of the sensor, used by the 
  // track-reconstruction software for the spectrometer sensors. 
  // Make sure to call SetRotMtx(), SetPosition() and SetDim() first. 

  if(!fMatrix) {
    Error("CalcBoxPoints", 
          "fMatrix = NULL. You have to do SetRotMtx() first!\n");
    return;
  }

  Double_t *matrix = fMatrix->GetMatrix();

  // Calculate the edge points of the box
  // in the middle of the plane thickness
  fX0 = fX - fDx*matrix[0] - fDy*matrix[3];
  fY0 = fY - fDx*matrix[1] - fDy*matrix[4];
  fZ0 = fZ - fDx*matrix[2] - fDy*matrix[5];

  fX1 = fX + fDx*matrix[0] + fDy*matrix[3];
  fY1 = fY + fDx*matrix[1] + fDy*matrix[4];
  fZ1 = fZ + fDx*matrix[2] + fDy*matrix[5];

  CalcAlpha();
}

void TPhBoxSensor::CalcAlpha()
{
  // Calculate the angle between the normal of the detector and the beamaxis. 

  // Not tested!
  // -----------
  fAlpha = fMatrix->GetTheta() *(TMath::Pi()/180);
  // GetTheta() always positive
  if((fX1-fX0)<0) fAlpha *= -1.0;
  // to get normal
  fAlpha += TMath::Pi()/2.0;
}

void TPhBoxSensor::Print()
{
  // Print some information about this sensor
  // option is unused.

  cout << GetName() << " is a " << GetTitle() << " BOX shape 1st corner = ("
       << GetX0() << ", " << GetY0() << ", " << GetZ0() << ") 2nd corner = (" 
       << GetX1() << ", " << GetY1() << ", " << GetZ1() << ")" << endl;
}

