#include "TPhBoxSiSensor.h"

#include "TPhHit.h"
#include "TPhHitArray.h"
#include "TPhObjectContainer.h"
#include "TPhEDigi.h"
#include "TPhEvent.h"
#include "TPhDetector.h"

#include "TRandom.h"

#include <iostream.h>

ClassImp(TPhBoxSiSensor)
/////////////////////////////////////////////////////////////////////////////
//
// TPhBoxSiSensor is the base class for all the BOX shaped Silicon sensors 
// such as TPhSpectSensor and TPhVTXSensor
//
// $Id: TPhBoxSiSensor.cxx,v 1.12 1999/02/25 00:36:13 olszewsk Exp $
//
/////////////////////////////////////////////////////////////////////////////

TPhBoxSiSensor::TPhBoxSiSensor()
{
  // Default constructor. 

  fX0 = 0;
  fX1 = 0;
  fY0 = 0;
  fY1 = 0;
  fZ0 = 0;
  fZ1 = 0;
  fAlpha = 0;
}

TPhBoxSiSensor::TPhBoxSiSensor(const Char_t* Name, Int_t Label, Int_t Layer, 
		          Double_t X0, Double_t X1, Double_t Y0, Double_t Y1, 
		          Double_t Z0, Double_t Z1)
{
  // Normal constructor. 

  SetName(Name);
  SetLabel(Label);
  SetLayer(Layer);

  fX0 = X0;
  fX1 = X1;
  fY0 = Y0;
  fY1 = Y1;
  fZ0 = Z0;
  fZ1 = Z1;
}

TPhBoxSiSensor::~TPhBoxSiSensor()
{
  // Destructor
  
  SafeDelete(fMatrix);
  SafeDelete(fShape);
}

void TPhBoxSiSensor::SetDim(Float_t dx, Float_t dy, Float_t dz)
{
  // Set the half-dimensions (a la GEANT) of the Box volume. 

  Char_t name[40];
  sprintf(name, "Brik_%s", GetName());
  fShape = new TBRIK(name, name, "", dx, dy, dz);
  fDx = dx; fDy = dy; fDz = dz;
}

void TPhBoxSiSensor::CalcBoxPoints()
{
  // Calculates the outer points of the sensor. These are the points that 
  // can later be retrieved by GetX0() etc. It also sets the angle 
  // between the beam pipe and the normal of the sensor, used by the 
  // track-reconstruction software for the spectrometer sensors. 
  // Make sure to call SetRotMtx(), SetPosition() and SetDim() first. 

  if(!fMatrix) {
    Error("CalcBoxPoints", 
          "fMatrix = NULL. You have to do SetRotMtx() first!\n");
    return;
  }

  Double_t *matrix = fMatrix->GetMatrix();

  // Calculate the edge points of the box, 
  // in the middle of the plane thickness
  fX0 = fX - fDx*matrix[0] - fDy*matrix[3];
  fY0 = fY - fDx*matrix[1] - fDy*matrix[4];
  fZ0 = fZ - fDx*matrix[2] - fDy*matrix[5];

  fX1 = fX + fDx*matrix[0] + fDy*matrix[3];
  fY1 = fY + fDx*matrix[1] + fDy*matrix[4];
  fZ1 = fZ + fDx*matrix[2] + fDy*matrix[5];

  CalcAlpha();
}

void TPhBoxSiSensor::CalcAlpha()
{
  // Calculate the angle between the normal of the detector and the beamaxis. 

  // Not tested!
  // -----------
  fAlpha = fMatrix->GetTheta() *(TMath::Pi()/180);
  // GetTheta() always positive
  if((fX1-fX0)<0) fAlpha *= -1.0;
  // to get normal
  fAlpha += TMath::Pi()/2.0;
}

void TPhBoxSiSensor::Print()
{
  // Print some information about this sensor
  // option is unused.

  cout << GetName() << " is a " << GetTitle() << " BOX shape 1st corner = ("
       << GetX0() << ", " << GetY0() << ", " << GetZ0() << ") 2nd corner = (" 
       << GetX1() << ", " << GetY1() << ", " << GetZ1() << ")" << endl;
}

TPhHit *TPhBoxSiSensor::Pixelize(TPhHit *hit)
{
// Pixelise position of hit based on pixel information
// known to TPhBoxSiSensor object. Also handles hit merging
// for hits falling into same pixel.

  Int_t    flag;
  Double_t d_old;
  Double_t dx_old;
  Double_t dz_old;
  Double_t d_new;
  Double_t dx_new;
  Double_t dz_new;
  Double_t dx;
  Double_t dy;
  Double_t dz;
  Double_t old_x;
  Double_t old_z;
  Double_t ps_xz;
  Double_t ps_y;
  TPhHit *OldHit;

  // remove the implementation of fEfficiency, since the efficiency isn't 
  // treated on sensor level anymore (jk)
  // if(gRandom->Rndm() > fEfficiency)
  //  {
  //    if(fVerbose)
  //   printf("Delete hit at %g %g %g\n",hit->GetX(),hit->GetY(),hit->GetZ());
  //    delete hit;
  //    return NULL;
  //  }
  flag = kFALSE;
#ifdef DEBUG
  printf("Before: %g %g %g\n",hit->GetX(),hit->GetY(),hit->GetZ());
#endif

  old_z = hit->GetZ();
  old_x = hit->GetX();
  ps_xz = GetPxlWidth();
#ifdef DEBUG
  printf("LayerNumber = %d, det = %d ==> %g\n",
	 GetLayer(),GetLabel(),ps_xz);
#endif
  dx_old = hit->GetX() - GetX0();
  dz_old = hit->GetZ() - GetZ0();
  d_old  = sqrt(dx_old*dx_old+dz_old*dz_old);
  d_new  = ps_xz * (TMath::Nint(d_old/ps_xz) + (gRandom->Rndm() - 0.5));
  if(d_old == 0)
    {
      fprintf(stderr,"TPhBoxSiSensor::Pixelize(): d_old = %g (%g,%g %g,%g)\n",
              d_old,hit->GetX(),GetX0(),hit->GetZ(),GetZ0());
      d_old = gRandom->Gaus(0.0,0.001);
      fprintf(stderr,"Set d_old to %g\n",d_old);
    }
  dz_new  = d_new/d_old * dz_old;
  dx_new  = d_new/d_old * dx_old;
  hit->SetX(GetX0() + dx_new);
  hit->SetZ(GetZ0() + dz_new);
  dx = hit->GetX() - old_x;  
  dz = hit->GetZ() - old_z;  
#ifdef DEBUG
  printf("d_old,d_new: %g %g\n",d_old,d_new);
  printf("dx_old,dz_old,dx_new,dz_new: %g %g %g %g\n",
          dx_old,dz_old,dx_new,dz_new);
  printf("dx,dz,dx_old - dx_new,dz_old-dz_new: %g %g %g %g  \n",
	 dx,dz,dx_old - dx_new,dz_old-dz_new);
  printf("d_old - d_new,sqrt(dx*dx+dz*dz): %g %g\n",
         d_old - d_new,sqrt(dx*dx+dz*dz));
#endif

  ps_y = GetPxlHeight();
  if(GetRows() % 2 != 0)
    {
      if(fabs(hit->GetY()) < ps_y/2.0)
        {
	  hit->SetY(0);
        }
      else
        {
	  if(hit->GetY() >= 0)
	    hit->SetY(ps_y * TMath::Nint(fabs(hit->GetY()/ps_y)));
	  else
	    hit->SetY(-(ps_y * TMath::Nint(fabs(hit->GetY()/ps_y))));
        }
    }
  else
    {
      if(hit->GetY() >= 0)
	hit->SetY(ps_y/2.0 + ps_y * (Double_t)(int) fabs(hit->GetY()/ps_y));
      else
	hit->SetY(-(ps_y/2.0 + ps_y * (Double_t)(int)fabs(hit->GetY()/ps_y)));
    }

#ifdef DEBUG
  printf("After:  %g %g %g\n",hit->GetX(),hit->GetX(),hit->GetX());
#endif

  //mpd 980526: this if statement was in the code, what is the reason?
  //mpd : i commented the if statement out...
  //  if(kFALSE)
  //{
      if(!fHitList)
        fHitList = new TObjArray();
  
      TIter NextHit(fHitList);
      while((OldHit = (TPhHit *) NextHit()) != NULL)
        {
	  dx = hit->GetX() - OldHit->GetX();
	  dy = hit->GetY() - OldHit->GetY();
	  dz = hit->GetZ() - OldHit->GetZ();
	  if(sqrt(dx*dx+dy*dy) * kMrgFak < GetPxlWidth() &&
	     fabs(dz) * kMrgFak < GetPxlHeight())
            {
#ifdef DEBUGM
	      printf("Row %d: Merge %g %g %g and %g %g %g\n",
		     LayerNumber,
		     OldHit->GetX(), 
		     OldHit->GetY(), 
		     OldHit->GetZ(), 
		     hit->GetX(),
		     hit->GetY(),
		     hit->GetZ());
#endif
	      OldHit->SetDe(hit->GetDe());
	      delete hit;
	      hit = OldHit;
	      flag = kTRUE;
	      break;
	    }
	}
      fHitList->Add(hit);
      return hit;
      //    } //mpd: end of if(kFALSE)
}

Bool_t TPhBoxSiSensor::MakeHitList(Char_t* HitConName)
{
  // Make a HitList from the HitArray. The hitarray should have previously 
  // been added to this Sensor. 

  Double_t X, Y, Z;
  Int_t i, j;

  Short_t Adc;
  Double_t E;

  TPhEDigi Digitizer;
  TPhHit *Hit;

  if(!fHitList)
    fHitList = new TObjArray();

  TPhEvent *event = fDetector->GetEvent();
  // See if we have an event
  if(!event) {
    if(fVerbose > 0)
      Warning("MakeHitList", "No Event connected to Detector! \n");
    return kFALSE;
  }

  // return if we don't have a HitArray
  if(!fHitArray) 
    return kFALSE;

  // Check if we have Hit container in event. if not make it.
  TPhObjectContainer *HitCont;
  HitCont = (TPhObjectContainer*) event->GetObject(HitConName);
  if(!HitCont) {
     HitCont = new TPhObjectContainer(HitConName);
     event->AddObject(HitCont);
  }

  // FIXME: this is very very naive and needs to be fixed!!!
  // FIXME: i think that this has to move to a new class taking care
  // of the calibration stuff for the ADCs at least

  // this is an inefficient way of doing this, but it seems to 
  // currently be the only way. 
  for (i = 0; i < fHitArray->GetRows(); i++) {
      for (j = 0; j < fHitArray->GetColumns(); j++) {
	Adc = fHitArray->GetADC(i, j);

	if(Adc > 0) {
	  E = Digitizer.ADCtoE(Adc);
	  Hit = new TPhHit();
	  // Get Coords of pixel i, j
	  GetCoords(i, j, &X, &Y, &Z);
	  Hit->SetX(X);
	  Hit->SetY(Y);
	  Hit->SetZ(Z);
	  Hit->SetDe(E);
          Hit->SetPadX(j);
          Hit->SetPadY(i);
	  
	  Hit->SetSensorLabel(GetLabel());
	  Hit->SetLayer(GetLayer());
	  
	  // Add hit to hitlist
	  fHitList->Add(Hit);
          HitCont->Add(Hit);
	}
      }
  }

  return kTRUE;
}

Int_t TPhBoxSiSensor::MakeHitsOutOfHitArray(Char_t *HitArrayConName,
                                            Char_t *HitConName)
{
  // Make Hits from the HitArray.
  // All the checks for containers has been done at the detector level 
  // and are not repeated
  Double_t X, Y, Z;
  Int_t i, j, hitcounter=0;

  Short_t Adc;
  Double_t E;

  TPhHit *Hit;

  if(!fHitList)
    fHitList = new TObjArray();
  else fHitList->Clear();

  TPhEvent *event = fDetector->GetEvent();
  // See if we have an event
  if(!event) {
    if(fVerbose > 0)
      Warning("MakeHitsOutOfHitArray", "No Event connected to Detector! \n");
    return kFALSE;
  }

  // Try to get the hit array from the container 
  if(!fHitArray) {
    TPhHashContainer *HitArrCont;
    HitArrCont = (TPhHashContainer*) event->GetObject(HitArrayConName);
    if(HitArrCont) fHitArray = (TPhHitArray *) HitArrCont->GetObject(GetName());
    // return if we still don't have a HitArray
    if(!fHitArray) return 0;
  }

  // Check if we have Hit container in event. if not make it.
  TPhObjectContainer *HitCont;
  HitCont = (TPhObjectContainer*) event->GetObject(HitConName);

  // this is an inefficient way of doing this, but it seems to 
  // currently be the only way. 
  Int_t nrows = fHitArray->GetRows();
  Int_t ncols = fHitArray->GetColumns();
  for (i = 0; i < nrows; i++) {
      for (j = 0; j < ncols; j++) {
	Adc = fHitArray->GetADC(i, j);
	// The hit array should contain pedestal corrected, 
	// zero suppressed data
	// drop negative signals
	if( Adc <= 0 ) continue;

	E = kChargeToEnergy *Adc;
	Hit = new TPhHit();
	// Get Coords of pixel i, j
	GetCoords(i, j, &X, &Y, &Z);
	Hit->SetX(X);
	Hit->SetY(Y);
	Hit->SetZ(Z);
	Hit->SetDe(E);
	Hit->SetPadX(j);
	Hit->SetPadY(i);
	
	Hit->SetSensorLabel(GetLabel());
	Hit->SetLayer(GetLayer());
	
	// Add hit to hitlist
	fHitList->Add(Hit);
	HitCont->Add(Hit);
	hitcounter++;
      }
  }
  return hitcounter;
}

Bool_t TPhBoxSiSensor::GetCoords(Int_t Row, Int_t Column, 
			    Double_t *X, Double_t *Y, Double_t *Z)
{
// Calculate the X, Y, Z position for a given pixel Row and Column

  Double_t *matrix = fMatrix->GetMatrix();
  Double_t dx, dy;

  dx = GetPxlWidth() * Column + GetPxlWidth()/2;
  dy = GetPxlHeight() * Row + GetPxlHeight()/2;

  *X = fX + (dx - fDx)*matrix[0] + (dy - fDy)*matrix[3];
  *Y = fY + (dx - fDx)*matrix[1] + (dy - fDy)*matrix[4];
  *Z = fZ + (dx - fDx)*matrix[2] + (dy - fDy)*matrix[5];
  
  return kTRUE;
}

