#include <iostream.h>
#include "TPhFEC.h"

ClassImp(TPhFEC)

/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// Description of FEC structure and channel properties.                    //
// FEC is divided into strings of channels, that may come from different   //
// sensors. A string may come from from several chips, each of them        //
// may be connected to a various number of pads.                           //
//                                                                         //
// $Id: TPhFEC.cxx,v 1.1 1999/02/24 22:48:06 olszewsk Exp $
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

TPhFEC::TPhFEC()
{
  // Default constructor

  fFECNo = 0;
  fNumStrings = 0; 
  fStrings = NULL;
  fVerbose = 0;
}

TPhFEC::TPhFEC(Int_t FECNum, 
	       Int_t numstrings, Int_t *stringsizes, Int_t *chipnos, Int_t *chipsizes) 
{
  // Normal constructor 
  Int_t i, choff=0;
  Char_t Name[20];

  sprintf(Name, "FEC_%d", FECNum);
  SetName(Name);
  sprintf(Name,"FEC with %d strings", numstrings);
  SetTitle(Name);

  fVerbose = 0;
  fFECNo = FECNum;
  fNumStrings = numstrings;
  fStrings = new TObjArray(numstrings);
  for(i=0; i<numstrings; i++) {
    // count current total of chips
    fStrings->AddAt(new TPhString(FECNum, i, stringsizes[i], 
		                  chipnos[i], chipsizes+choff),i);
    choff += chipnos[i];
  }
}

TPhFEC::~TPhFEC()
{
  // Default destructor

  SafeDelete(fStrings);
}

TPhString *TPhFEC::GetString(Int_t i)  
{ 
  // return pointer to string at position i

  if(fStrings) {
    if(fStrings->GetSize()<=i) return NULL;
    return ((TPhString *) fStrings->At(i));
  }
  return NULL; 
}

void TPhFEC::RemovePedArrays()
{
  // Delete arrays storing pedestal mean and sigma 
  // from string objects for memory saving reasons
  TPhString *string;
  TIter nextString(fStrings);

  while((string = (TPhString *) nextString())) string->DeletePedArrays();
}

void TPhFEC::RemoveCMNArrays()
{
  // Delete arrays with CMN mean value and validity flag
  // from string objects for memory saving reasons
  TPhString *string;
  TIter nextString(fStrings);

  while((string = (TPhString *) nextString())) string->DeleteCMNArrays();
}

void TPhFEC::RemoveNoiseArrays()
{
  // Delete arrays with noise levels
  // from string objects for memory saving reasons
  TPhString *string;
  TIter nextString(fStrings);

  while((string = (TPhString *) nextString())) string->DeleteNoiseArrays();
}

void TPhFEC::RemoveZZArrays()
{
  // Delete ZZ arrays 
  // from string objects for memory saving reasons
  TPhString *string;
  TIter nextString(fStrings);

  while((string = (TPhString *) nextString())) string->DeleteZZArrays();
}

void TPhFEC::RemoveSuppArrays()
{
  // Delete support arrays 
  // from string objects for memory saving reasons
  TPhString *string;
  TIter nextString(fStrings);

  while((string = (TPhString *) nextString())) string->DeleteSuppArrays();
}

Bool_t  TPhFEC::PedPreProcess(TPhRawHitFECn0 *fecevent)
{
  // Add pedestal estimation info from all strings of a single event 
  TPhString *string;
  FECRawHit_t *hitstring;
  Int_t hitstrlength,lowchannel,i;
  Bool_t status=kTRUE;
  
  if(!fecevent) return kFALSE;
  for(i=0;i<kNumStrings;i++) {
    hitstrlength = fecevent->GetRawHitArraySize(i);
    if(!hitstrlength) continue;
    hitstring = fecevent->GetRawHitArray(i);
    string = (TPhString *) fStrings->At(i);
    if(string) {
      lowchannel = string->PedPreProcess(hitstring);
      status = status && !((Bool_t)lowchannel);
      if(fVerbose>2) printf("PedPreProcess: FEC no %d String no %d " \
			    "Channel no %d statistics too low\n", fFECNo, i,lowchannel);
    }			    
    else {
      Error("PedPreProcess","Unknown string structure in FEC no %d\n",fFECNo);
      return kFALSE;
    }
  }
  return status;
}

Bool_t  TPhFEC::PedProcess(TPhRawHitFECn0 *fecevent)
{
  // Add pedestal info from all strings of a single event 
  TPhString *string;
  FECRawHit_t *hitstring;
  Int_t hitstrlength,lowchannel,i;
  Bool_t status=kTRUE;
  
  if(!fecevent) return kFALSE;
  for(i=0;i<kNumStrings;i++) {
    hitstrlength = fecevent->GetRawHitArraySize(i);
    if(!hitstrlength) continue;
    hitstring = fecevent->GetRawHitArray(i);
    string = (TPhString *) fStrings->At(i);
    if(string) {
      lowchannel = string->PedProcess(hitstring);
      status = status && !((Bool_t)lowchannel);
      if(fVerbose>2) printf("PedProcess: FEC no %d String no %d " \
			    "Channel no %d statistics too low\n", fFECNo, i,lowchannel);
    }			    
    else {
      Error("PedProcess","Unknown string structure in FEC no %d\n",fFECNo);
      return kFALSE;
    }
  }
  return status;
}

Bool_t  TPhFEC::CMNProcess(TPhRawHitFECn0 *fecevent)
{
  // Calculate CMN for all strings of a single event 
  TPhString *string;
  FECRawHit_t *hitstring;
  Int_t hitstrlength,badchips,i;
  Bool_t status=kTRUE;  

  if(!fecevent) return kFALSE;
  for(i=0;i<kNumStrings;i++) {
    hitstrlength = fecevent->GetRawHitArraySize(i);
    if(!hitstrlength) continue;
    hitstring = fecevent->GetRawHitArray(i);
    string = (TPhString *) fStrings->At(i);
    if(string) {
      badchips = string->CMNProcess(hitstring);
      status = status && !((Bool_t)badchips);
      if(fVerbose>2) printf("CMNProcess: FEC no %d String no %d " \
			    "No of chips with irregular CMN %d", fFECNo, i, badchips);
    }			    
    else {
      Error("CMNProcess","Unknown string structure in FEC no %d\n",fFECNo);
      return kFALSE;
    }
  }
  return status;
}

Bool_t  TPhFEC::NoiseProcess(TPhRawHitFECn0 *fecevent)
{
  // Add noise info from all strings of a single event 
  TPhString *string;
  FECRawHit_t *hitstring;
  Int_t hitstrlength,lowchannel,i;
  Bool_t status;

  if(!fecevent) return kFALSE;
  for(i=0;i<kNumStrings;i++) {
    hitstrlength = fecevent->GetRawHitArraySize(i);
    if(!hitstrlength) continue;
    hitstring = fecevent->GetRawHitArray(i);
    string = (TPhString *) fStrings->At(i);
    if(string) {
      lowchannel = string->NoiseProcess(hitstring);
      status = status && !((Bool_t)lowchannel);
      if(fVerbose>2) printf("NoiseProcess: FEC no %d String no %d " \
			    "Channel no %d statistics too low\n", fFECNo, i,lowchannel);
    }			    
    else {
      Error("NoiseProcess","Unknown string structure in FEC no %d\n",fFECNo);
      return kFALSE;
    }
  }
  return status;
}

Bool_t  TPhFEC::ZeroSuppress(TPhRawHitFECn0 *fecevent)
{
  // Zero suppress all strings of a single event 
  TPhString *string;
  FECRawHit_t *hitstring;
  Int_t hitstrlength,i;
  Bool_t status=kTRUE;
  
  if(!fecevent) return kFALSE;
  for(i=0;i<kNumStrings;i++) {
    hitstrlength = fecevent->GetRawHitArraySize(i);
    if(!hitstrlength) continue;
    hitstring = fecevent->GetRawHitArray(i);
    string = (TPhString *) fStrings->At(i);
    if(string) status = status && 
		 ( (string->ZeroSuppress(hitstring))>=0 );
    else {
      Error("ZeroSuppress","Unknown string structure in FEC no %d\n",fFECNo);
      return kFALSE;
    }
  }
  return status;
}

Bool_t  TPhFEC::Calibrate(TPhAbsRawHitFEC *fecevent)
{
  // Calibrate all strings of a single event 
  TPhString *string;
  FECRawHit_t *hitstring;
  Int_t hitstrlength,i;
  Bool_t status=kTRUE;
  
  if(!fecevent) return kFALSE;
  for(i=0;i<kNumStrings;i++) {
    hitstrlength = fecevent->GetRawHitArraySize(i);
    if(!hitstrlength) continue;
    hitstring = fecevent->GetRawHitArray(i);
    string = (TPhString *) fStrings->At(i);
    if(string) status = status && string->Calibrate(hitstring, hitstrlength);
    else {
      Error("Calibrate","Unknown string structure in FEC no %d\n",fFECNo);
      return kFALSE;
    }
  }
  return status;
}

Bool_t  TPhFEC::CalibrateOnStrings(TPhAbsRawHitFEC *fecevent)
{
  // Calibrate all strings of a single event 
  TPhString *string;
  FECRawHit_t *hitstring;
  Int_t hitstrlength,i;
  Bool_t status=kTRUE;
  
  if(!fecevent) return kFALSE;
  for(i=0;i<kNumStrings;i++) {
    hitstrlength = fecevent->GetRawHitArraySize(i);
    if(!hitstrlength) continue;
    hitstring = fecevent->GetRawHitArray(i);
    string = (TPhString *) fStrings->At(i);
    if(string) status = status && string->CalibrateOnString(hitstring, hitstrlength);
    else {
      Error("Calibrate","Unknown string structure in FEC no %d\n",fFECNo);
      return kFALSE;
    }
  }
  return status;
}

void TPhFEC::Calibrate(Bool_t &status)
{ 
  //
  Bool_t strstatus = kTRUE;

  fStrings->ForEach(TPhString,Calibrate)(strstatus); 
  status = status && strstatus;                               
  
  return;
}

void  TPhFEC::ls(Option_t *option) 
{
  //
    IndentLevel();
    cout << IsA()->GetName() << " Name: " << GetName() << endl;
    TObject::IncreaseDirLevel();

    if(strstr(option, "C") || strstr(option, "c"))
       fStrings->ls();

    if(strstr(option, "D") || strstr(option, "d"))
       fStrings->ls(option);

    TObject::DecreaseDirLevel();
}

void  TPhFEC::Print(Option_t *option) 
{
  //

}


