#include "TPhSiSensor.h"
#include "TPhMCHit.h"
#include "TPhHitArray.h"
#include "TPhEvent.h"
#include "TPhDetector.h"
#include "TPhEDigi.h"
#include "TPhObjectContainer.h"
#include "TPhHashContainer.h"

#include "TMatrix.h"
#include "TCanvas.h"
#include "TPad.h"
#include "TPaveLabel.h"

ClassImp(TPhSiSensor)
/////////////////////////////////////////////////////////////////////////////
// TPhSiSensor is the base class for all Phobos Silicon sensors. 
// It is meant to handle all the general silicon stuff. 
// 
/////////////////////////////////////////////////////////////////////////////

TPhSiSensor::TPhSiSensor()
{
  // Default constructor
  fThickness  = 0.03;
  fPadWidthX = 0;
  fPadWidthY = 0;
  fRows = 0;
  fColumns = 0;
  fHitArray   = NULL;
}

TPhSiSensor::~TPhSiSensor()
{
  // Default destructor
  SafeDelete(fHitArray);
}

void TPhSiSensor::ClearHitArray()
{
  // Zero the HitArray member BUT DON'T DELETE it! Be careful with this 
  // function, it is a potential memory leak!

  fHitArray = NULL;
}

Int_t TPhSiSensor::MakeHitsOutOfHitArray(Char_t *HitArrayConName, 
					  Char_t *HitConName)
{
  // Not done, but put to make SiSensor class non-abstract

  return 0;
}

Int_t TPhSiSensor::MakeHitArraysOutOfHits(Char_t *HitConName, 
					  Char_t *HitArrayConName)
{
// Generate HitArray for a sensor from the hits. 
// The data container with Hits is called HitConName
// The data container with HitArrays is called HitArrayConName
// 
//  Return: Number of HitArrays created (always 1 hitarray for 1 sensor)

  const Char_t *sensorName = GetName();

  TPhEvent *event = fDetector->GetEvent();
  if(!event) {
    if(fVerbose > 0)
      Warning("MakeHitArraysOutOfHits", "No Event connected to Detector!\n");
    return 0;
  }

  TPhObjectContainer *HitContainer = NULL;
  HitContainer = (TPhObjectContainer*) event->GetObject(HitConName);
  if(!HitContainer) {
    if(fVerbose > 0) 
      Warning("MakeHitArraysOutOfHits", 
	      "Sensor: %s --> no %s hit container in the event\n", 
	      sensorName,HitConName);
    return 0;
  }

  // Check if we have HitArrayConName container in event. if not, make it.
  TPhHashContainer *hitArrCon = NULL;
  hitArrCon = (TPhHashContainer*) event->GetObject(HitArrayConName);
  if(!hitArrCon) {
    hitArrCon = new TPhHashContainer(HitArrayConName,
				     "HitArrays made from MonteCarlo hits");
    event->AddObject(hitArrCon);
    if(fVerbose > 0) 
      Warning("MakeHitArraysOutOfHits", 
	      "Creating a new HitArrayContainer %s\n", 
	      HitArrayConName);
  }

  // Check if we have HitArray named like the sensor in the container. 
  // if yes, return
  TPhHitArray *hitArr = NULL;

  hitArr = (TPhHitArray *) hitArrCon->GetObject(sensorName);
  if(!hitArr) hitArr = new TPhHitArray(this, sensorName, sensorName);
  else {
    if(fVerbose > 0) 
      Warning("MakeHitArraysOutOfHits", 
	      "Sensor: %s --> HitArray for this sensor already present "\
	      "in the container\n", sensorName);
    fHitArray = hitArr;
    // cannot allow to overwrite existing HitArray
    return 1;
  }

  // Create a temporary array where the energy from individual hits 
  // will be summed up
  Int_t nrows, ncolumns;
  nrows = GetRows();
  ncolumns = GetColumns();
  if(fVerbose>1) 
     printf("Processing sensor %s, nrows = %d, ncolumns = %d\n", 
	    sensorName, nrows, ncolumns);
  TMatrix he(nrows,ncolumns);

  // Make an iterator and iterate over the sensor's Hit List
  TIter NextHit(GetHitList());
  TPhMCHit *hit;

  // sum the hit energies
  Int_t hitcount = 0;
  while ((hit = (TPhMCHit *) NextHit()) != NULL) {

     Float_t E    = hit->GetDe();
     //unused for hitarrays:     Float_t Dt   = hit->GetDt();
     Int_t   PadX = hit->GetPadX();
     Int_t   PadY = hit->GetPadY();
     if(fVerbose>1) 
       printf("Temporary settings, nrows = %d, ncolumns = %d\n", PadY, PadX);

     // PMC convention is Y pixel direction for negative arm 
     // already changed in MC Hits
     he(PadY,PadX) += E; 
     hitcount++;
  }

  // If there are any hits, then connect this HitArray to the event, 
  // otherwise forget about it. 
  if(hitcount > 0) {
    Short_t De;
    TPhEDigi *digi = new TPhEDigi();

    hitcount = 0;
    for(Int_t rs=0; rs<nrows; rs++) {
      for(Int_t cs=0; cs<ncolumns; cs++) {
        if (he(rs,cs)>1.0e-8) {
	  // Energy to ADC conversion  
          De = (Short_t) digi->EtoADC( he(rs,cs) ); 
          hitArr->SetADC( rs,cs,De ); 
          hitcount++;
          if(fVerbose>1) 
	    printf("Final settings, nrows = %d, ncolumns = %d\n", rs, cs);
        }
      }
    }
    delete digi;
    
    // number of real hits (not the number of occupied cells) 
    // is stored in TPhHitArray fHits data member
    
    hitArr->SetNHits(hitcount);
    hitArrCon->Add(hitArr);
    fHitArray = hitArr;
  } else {
    delete hitArr;
    if(fVerbose>1) 
      printf("--> Sensor %s, no hits\n",sensorName);
  }
  return 1;
}

////////////////////////////////////////////////////
void TPhSiSensor::HitDisplay()
{
  // Draw single event hit distribution in a given sensor; the 
  // data have to be connected to the geometry by TPhSensor::AddHitArray()

  TPhEvent *event = fDetector->GetEvent();
  // See if we have an event
    if(!event) {
      if(fVerbose > 0)
	Warning("HitDisplay", "No Event connected to Detector!\n");
      return;
    }
    
    Int_t ii,jj;
    Float_t xlow, ylow, xup, yup;
    
//* Draw outer bounds of Si Sensor
    
    Float_t scale;
    Float_t xtot =  fPadWidthY * fRows;
    Float_t ytot =  fPadWidthX * fColumns;
    // increase scale for display purpose
    if(xtot > ytot) 
      {
	scale = 800 / xtot; 
	xtot = scale * xtot + 1;
	ytot = scale * ytot + 1;
      }
    else 
      {
	scale = 700 / ytot; 
	xtot = scale * xtot + 1;
	ytot = scale * ytot + 1;	
      }

    Int_t xcanvas = (Int_t) xtot + 20;
    Int_t ycanvas = (Int_t) ytot + 80;
    gPad  = new TCanvas("view", "Sensor",xcanvas, ycanvas);

    // draw canvas title

    Char_t title[256];
    //unused:    Char_t sname[256];
    sprintf(title,"evt %d,run %d, %s\n",event->GetNEvent(), 
	    event->GetNRun(),fName.Data());

    // draw pads (blue = no hit; red  = hit)
    xlow = 10./(xcanvas);
    ylow = 10./ycanvas;
    xup = xtot / xcanvas;
    yup = ytot / ycanvas;

    TPad *fPadSensor = new TPad("fPadSensor",title, 
				xlow, ylow, xup, yup, 4, 2);
    fPadSensor->Draw();

    Float_t xpavelow = xlow;
    Float_t ypavelow = yup + 0.02;
    Float_t xpaveup  = xup;
    Float_t ypaveup  = 0.98;
    //unused:    Float_t dxpave = xpaveup - xpavelow;
    TPaveLabel *fPadTitle = new TPaveLabel(xpavelow,ypavelow,
					   xpaveup,ypaveup,title);
    fPadTitle->SetTextFont(142);
    fPadTitle->SetTextSize(1.);
    fPadTitle->Draw();

    fPadSensor->Divide(fRows,fColumns,0.,0.,0);
    Int_t padind;
    //unused:    Int_t n = fHitArray->GetNHits();
    Short_t adc;
    for (ii=0; ii<fRows; ii++)
      {
	for (jj=0; jj<=fColumns; jj++)
	  {
	    adc = fHitArray->GetADC(ii, jj);
	    if (adc >0) 
	      {
		// calculate padindex from (ii,jj)->padind see TPad docu
		padind = (jj - 1) * fRows + ii;
		fPadSensor->cd(padind);
		gPad->SetFillColor(2);
		gPad->SetBorderSize(0);
		gPad->SetBorderMode(0);
		gPad->Draw();
	      }
	  }
      } 
}





