#include "TH1.h"
#include "TH2.h"
#include "TIterator.h" // ?
#include "TObject.h"
#include "TObjArray.h"
#include "TPhDept.h"
#include "TPhDetector.h"
#include "TPhDispDef.h"
#include "TPhEDigi.h"
#include "TPhEvent.h"
#include "TPhFsec.h"
#include "TPhHashContainer.h"
#include "TPhHitArray.h"
#include "TPhMCHit.h"
#include "TPhMCTrack.h"
#include "TPhMul.h"
#include "TPhObjectContainer.h"
#include "TPhOpts.h"
#include "TPhPadMap.h"
#include <math.h> // For GetData and Recon methods
#include "TPhSensDef.h"
#include "TPhSensor.h"
#include "TPhVertex.h"
#include "TProfile.h"
#include <stdlib.h> // For rand and srand in GetData method diagnostic mode

ClassImp(TPhMul)
///////////////////////////////////////////////////////////////////////////
//
// TPhMul is a multiplicity reconstruction class that creates and displays
// arrays of reconstructed values of dN/dPad and d2N/dEtadPhi for selectable
// combinations of sensors.  Minimally, the user instantiates a TPhMul
// object and a TPhEvent object, and calls TPhMul::Process(event, 0).
// Other calls allow you to reset operating parameters after reading an event;
// see Process.
//
// The destructors and streamers are not yet fully implemented.
//
// File: TPhMul.cxx
//
// Date     Vers  Description
// 09-Feb-99 4.0  Fix bug in clearing 1-ev displays; ignore "-1" in name;
//                move fSensDefs and fDispDefs from stack to heap (RV)
//
// $Id: TPhMul.cxx,v 1.17 1999/02/10 04:22:20 verdier Exp $
//
///////////////////////////////////////////////////////////////////////////

// 15-Jan-99 4.0  Change "MCHits" -> "PMCHits", "MCTracks" -> "PMCTracks",
//                "HitArraysFromHits" -> "PMCHitArraysFromHits" following
//                unannounced changes in Phat; make names public; remove
//                include for newly obsolete TPhDataContainer (RV)
// 02-Jan-99 4.0  Update GetGraph and GetHist calls (RV)
// 16-Dec-98 4.0  Remove unused arguments to avoid gcc warnings (RV)
//                See casting problem at line 1686 ff
// 06-Dec-98 4.0  Change default HitArrayName again, remove OLDTHMS (RV)
// 29-Nov-98 3.0  Use new TPh hit and track methods unless OLDTHMS is defined;
//                add fDet to provide access to detector object;
//                add fHitArrayName, fHitArrayId, fDataFormat (RV)
// 10-Nov-98 1.0  Correct for OSF1 scope problem in Release (RV)
// 20-Oct-98 1.0  Correct phi range in GetData (RV)
// 30-Sep-98 1.0  Correct hit-track relation using TPhMCTrack::GetHitList (RV)
// 25-Sep-98 1.0  Normalize and limit EdepPrimMC; add nMCTrack stats; 
//                count MC track segments depositing no energy; 
//                correct layer initialization (RV)
// 10-Sep-98 1.0  Allow accumulation of sensor maps for diagnostics (RV)
// 02-Sep-98 1.0  Change energy deposition diagnostics to E vs Z (RV)
// 28-Aug-98 1.0  Make sensor data error message limits and HitArray 
//                names public (RV)
// 21-Aug-98 1.0  Correct isens upper limit in Process method (RV)
// 15-Aug-98 1.0  Correct FSecRMSMC calculation and parent track syntax (RV)
// 11-Aug-98 1.0  UseHitArrays -> fDataSource = EData; add SetVertex (RV)
// 08-Aug-98 1.0  nCharLayer= 2 for Vxxx; tally bad indices, unknown
//                sensors, inactive sensors, used and unused hits;
//                use TPhObjectContainers for tracks and hits (RV)
// 21-Jul-98 1.0  Remove obsolete include for TPhRun.h (RV)
// 14-Jul-98 1.0  Add GetGraph and GetHist and remove Display methods (RV)
// 07-Jul-98 1.0  Add fNTracksUse for diagnostic tracing, ixy+1->ixy (RV)
// 26-Jun-98 1.0  Enable use of MCHitArrays (RV)
// 23-Jun-98 1.0  Access GetData diagnostics through public variables (RV)
// 22-Jun-98 1.0  Do diagnostics only for real hits; change TraceTracks (RV)
// 17-Jun-98 1.0  Add Diagnostic dE/dX distributions in GetData (RV)
// 11-Jun-98 1.0  Add GetDisplayIndex and GetSensMapIndex, convert Edep
//                to normal incidence (RV)
// 02-Jun-98 1.0  Save process mode to avoid re-initialization (RV)
// 01-Jun-98 1.0  Implement MaxEvents in PadMaps (RV)
// 12-May-98 0.3  Remove det argument from Process (RV)
// 09-May-98 0.3  Add GetSensMap(name) (RV)
// 07-May-98 0.3  Fix GetDisplay and Process for g++ 'for' scoping (RV)
// 06-May-98 0.3  Add GetDisplay(name), ListDisplays, RL's (RV)
// 01-May-98 0.2  TPhPadMap::GetNcells -> GetNCells for consistency (RV)
// 12-Mar-98 0.2  Corrected syntax errors & use kTRUE, for Abacus use (RV)
// 25-Feb-98 0.2  Corrected error in GetId call (RV)
// 19-Feb-98 0.2  Recon creates display list using TPhPadMap::Select (RV)
// 15-Feb-98 0.1  Add fDispDefs for access to display definitions (RV)
// 29-Jan-98 0.1  Add GetDeptP and GetFsecP methods for access to 
//                Dept and Fsec classes for fitting (RV)
// 18-Jan-98 0.1  Add GetSensMap method for access to sensor data (RV)
// 15-Jan-98 0.1  Add diagnostic pointer access methods for histos (RV)
// 12-Jan-98 0.1  Add access to SensDefs; generalize Process mode arg (RV)
// 28-Dec-97 0.1  GetData (RV)
// 04-Nov-97 0.1  Creation (Robin Verdier)

// The constant ints kBarrel = 1 and kRing = 2 are defined in TPhMul.h.
// They could also be enumerated types.

// Deposited energy limit for one pad.  This should not be a single global
// value, but should instead depend on the average occupancy for each part
// of the detector.  The value taken here is 6.5 times the PDH mip = 117 keV.
// At present we default to the value 500 keV used in the old analysis.

Float_t TPhMul::feMax = 500.0e-6;
Int_t   fInitMode; // 0 at instantiation, 1 after initialization.

// Initialize numbers of active sensors and displays; they will be reset later
Int_t nDispAct = kNDispMax;
Int_t nSensAct = kNSensMax;

TPhMul::TPhMul()
  : RLiprint(0), RLNPrint(20), RLGlobal(1), RLProcess(1), 
    RLGetData(1), RLGetF(1), RLRecon1(1), RLRecon2(1)
{ fDataFormat       = 0; // 0 for .root, 1 for .apx
  fDataSource       = EMCData; // EData=0, EMCData=1, EMCHits=2, EDiagnostic=3 
  fDet              = (TPhDetector *)0;
// Handle capricious changes in PhAT object names.
  fHitArrayName[0]  = "PMCHitArraysFromHits"; // after 30-Dec-98
  fHitArrayName[1]  = "MCHitArrays";
  fHitArrayName[2]  = "HitArraysFromHits"; // before 30-Dec-98
  fGDNTracksTrace   = 0;
  fGDTrackTrace     = -1;
  fGDTraceSkip      = 0;
  for (Int_t nn = 0; nn < kNDiagMax; nn++)
    fGDNDiagBins[nn] = 0;
  fHitArrayId       = 0;
  fInitMode         = 0;
  fMaxBadIndexMsgs  = 0; // index discrepancy message limit (MC)
  fMaxInactSensMsgs = 0; // Inactive sensor message limit
  fMaxUnkSensMsgs   = 0; // unknown sensor message limit
  fMCHitsName       = "PMCHits"; // "MCHits" before 30-Dec-98
  fMCTracksName     = "PMCTracks"; // MCTracks before 30-Dec-98
  fNTracksUse       = 999999;
  fXIndexErrTol     = 1; // Maximum allowed Z-type pad index discrepancy (MC)
  fYIndexErrTol     = 1; // Maximum allowed Phi-type pad index discrepancy (MC)
}

TPhMul::~TPhMul()
{ Release();
}

////////////////////////////////////////////////////////////////////////////

void TPhMul::GetData(Int_t isens, TPhEvent *event, TPhPadMap **Maps, 
                     TPhDept &de, TPhFsec &fs)
{
// Get raw hit data for sensor isens or all sensors.
// In MC mode, this gets data from MCHits in the event object.
// In diagnostic mode, this generates random data of roughly the expected
// distribution in Eta.  It needs vertex dependence.
// Date     Vers  Description
// 23-Jun-98 1.0  Access tracing facilites through public variables (RV)
// 28-Dec-97 0.1  Correct MC track counting for multi-layer hits (RV)
// 10-Dec-97 0.1  Correct MC track counting for path length (RV)
// 26-Nov-97 0.1  Get MC tracks and hits (RV)
// 20-Nov-97 0.1  Get data from event object (RV)
// 10-Nov-97 0.1  Include secondary fraction (RV)
// 07-Nov-97 0.1  Explicitly init pseudorandom sequence (RV)
// 05-Nov-97 0.1  Correct +z ring calculation (RV)
// 23-Oct-97 0.1  Creation (Robin Verdier)

// Declare general local variables:

  Float_t        dEta;
  Float_t        dEPad; // Energy deposited in hit pad
  Float_t        dist = Maps[isens]->GetDist(); // Normal distance to plane
  Int_t          dloc[3]; // Loc pointer increments in map coords
static           // Making dtdr automatic changes the calculated output
  Float_t        dtdr; // Obliquity factor, (sin|cos)(theta) for (barr|ring)
  Float_t        dy = Maps[isens]->GetSize(1); // Y-cell size
  Float_t        e1; // Global Energy deposit / track for this sensor & eta
  Float_t        edep; // Thrown energy for this pad;
static           // Making epd automatic produces NaN's in output
  Float_t epd[3]; // eta, phi, dtdr returned by GetEtaPhi
  Float_t        etaLim[2];
  Float_t        fTrack; // fraction of a track assigned to read hit
//  Int_t          hitIterMode = 0; // !0 for old (wrong) hit-track association
//  Int_t          iev = 0;
//  Int_t          iHit; // Index of hit, for MCHits mode
  Int_t         &iprint = RLiprint;
  Int_t          is; // Temporary sensor index
//  Int_t          iSeg; // Index of segment (layer), for MCHits mode
// Facilities for locating a crash:
  Int_t         &iTraceSkip = fGDTraceSkip; // #tracks to skip between tracing
  Int_t         &iTrackTrace = fGDTrackTrace; // 1st track to trace debugging
  Int_t          iTrack; // Index of track, for MCHits mode
  Int_t          iTracksTraced; // Number of tracks traced, for MCHits mode
  Int_t          ix;
  Int_t          iy;
const
  Int_t          MaxTrackHits = 30; // Maximum number of hits saved on a track
  Int_t          nBadIndexMsgs; // number of discrepancy messages issued
  Int_t          NHBins[kNDiagMax];
  for (Int_t nn = 0; nn < kNDiagMax; nn++) NHBins[nn] = fGDNDiagBins[nn];
  Int_t          nHits; // Number of hits in current track
  Int_t          nInactSensMsgs; // number of inactive sensor messages issued
  Int_t          nMCHitsInactSens; // Hits on inactive sensors
  Int_t          nMCHitsRead; // Hits read in
  Int_t          nMCHitsTot; // Hits in container
  Int_t          nMCHitsUnkSens; // Hits on unknown sensors
  Float_t        nMCHitsUsed; // Sum of hit fractions entered into Sensor maps
  Int_t          nMCTracks; // MC tracks read
  Int_t          nMCTracksCh; // MC charged tracks read
  Float_t        nMCTracksUsed; // Sum of track fracts entered into Sensor maps
  Int_t         &Nprint = RLNPrint; // Max number of diagnostic lines
  Int_t          nSeg; // Number of segments in track, for MCHits mode
static
  Int_t          nstored; // # hits stored, for MCHits mode
  Int_t          nTracks; // Number of tracks in this event, for MCHits mode,
                          // including tracks with no hits
  Int_t         &nTracksTrace = fGDNTracksTrace;
  Int_t         &nTracksUse = fNTracksUse;
  Int_t          nUnkSensMsgs; // number of unknown sensor messages issued
  Int_t          nx = Maps[isens]->GetNCells(0); 
  Int_t          ny = Maps[isens]->GetNCells(1); 
  Int_t         &ReportLevel = RLGetData;
  Char_t        *sensName;
  Float_t        ylim[2];
  Float_t        ymin = Maps[isens]->GetMin(1); 

  TPhObjectContainer *allHits;
  TPhHashContainer   *hitarrays;
  TPhMCTrack         *track;
  TPhObjectContainer *tracks;

// Declare local variables for generating sham data to simulate occupancy
// dN/dEta = dNdE0 * exp(-(eta-eta0)**2/2/sigeta**2) for 90% at eta=0
// and 5% at eta = 5.0;

  Float_t dNdE0  = 800.; // dN/dEta(0)
  Float_t eta0   = 1.5;  // Center of eta distribution 
  static Bool_t  InitSeed = kTRUE; // Flag to init random number sequence
  const  Float_t Rrmax = RAND_MAX; // Pseudorandom number range
  Int_t          Seed = 1234567 ; // Initial value for pseudorandom sequence
  Float_t sigeta = 1.7;  // Width of dN/dEta distribution 

// Local variables used if CheckEtaCalc is true:
// Compare fast and slow eta calculations, for debugging:
  const  Int_t   CheckEtaCalc = kFALSE;
  static Float_t dx;
  static Float_t eta;
  static Float_t theta;
  static Int_t   type;
  static Float_t x;
  static Float_t xmin;

  if (InitSeed) {
    srand(Seed);
    InitSeed = kFALSE;
  }

  if (CheckEtaCalc)
  { xmin = Maps[isens]->GetMin(0); // Current x-value
    dx = Maps[isens]->GetSize(0); // X-cell size
    type = Maps[isens]->GetType(); // Ring or barrel
// Ring annuli have ~ constant dEta: dEta = (Z * dRho - Rho * dZ) / (R * Rho)
// where Rho*Rho = X*X + Y*Y and R*R = Rho*Rho + Z*Z
    if (type == kRing) dx = dx + 1.0; // Ring annuli have ~ constant dEta
    x = Maps[isens]->GetMin(0);
  }

// Get arrays of MC event data from the event treated as a data container
// This is time-consuming when we use tracks, rather than hit arrays, in
// order to get calibrations for energy deposition and secondary fractions.

// Tests for a valid event and its subcontainers should all return a
// failure code through a single return.  For the present, simply warn
// the caller.  Note a null test for a pointer doesn't tell you if the
// addressed object is valid, and neither PhAT nor ROOT offers a solution.

  if (!event && ReportLevel)
  { printf ("TPhMul/GetData: null event, exiting.\n");
    return;
  }
  if (!isens && fDataSource == EMCData)
  { hitarrays = (TPhHashContainer*)
      event->GetObject(fHitArrayName[fHitArrayId]);
    if (!hitarrays && ReportLevel)
    { printf ("TPhMul/GetData: no %s hitarray in event, exiting.\n",
        fHitArrayName[fHitArrayId]);
      printf ("Use the checkevents macro for more info.\n");
      return;
    }
    TPhHitArray *ha;
    TPhEDigi ed;
    for (is = 0; is < nSensAct; is++)
    { TPhPadMap *sens = GetSensMap(is);
      if ((ha = (TPhHitArray*) hitarrays->GetObject(sens->GetName())))
      { Int_t nX = ha->GetColumns();
        if (nX != sens->GetNCells(0))
        { if (ReportLevel >= 2 && iprint < Nprint)
          { printf ("TPhMul/GetData: hitarray for %s has %d X-cells,"
             " padmap has %d\n.", sens->GetName(), nX, sens->GetNCells(0));
            iprint++;
          }
          if (nX > sens->GetNCells(0)) nX = sens->GetNCells(0);
        }
        Int_t nY = ha->GetRows();
        if (nY != sens->GetNCells(1))
        { if (ReportLevel >= 2 && iprint < Nprint)
          { printf ("TPhMul/GetData: hitarray for %s has %d Y-cells,"
             " padmap has %d\n.", sens->GetName(), nY, sens->GetNCells(1));
            iprint++;
          }
          if (nY > sens->GetNCells(1)) nY = sens->GetNCells(1);
        }
        for (ix = 0; ix < nX; ix++)
         for (iy = 0; iy < nY; iy++)
// Note the order of indices in GetADC(iy, ix).  GetADC returns -1
// for an illegal index, ADCtoE happily converts it => FIX.
	 { dEPad = ed.ADCtoE(ha->GetADC(iy, ix));
// Accumulate the energy.  Later we may want to clear it here.
           *sens->GetLoc(ESEdep,ix,iy) += dEPad;
           if (strcmp(sens->GetName(),"MOA3") == 0
               && ReportLevel >= 10 && iprint < Nprint) // dEPad )
           { printf ("TPhMul/GetData: sensor %s[ix=%d,iy=%d] = %10.4e.\n", 
               sens->GetName(), ix, iy, dEPad);
             iprint++;
           }
	 }
      } // if (ha
      else if (ReportLevel >= 2 && iprint < Nprint)
      { printf ("TPhMul/GetData: No hitarray found for %s.\n", 
          Maps[is]->GetName());
        iprint++;
      }
    }
  } // if (!isens && fDataSource == EMCData

  if (isens == 0 && fDataSource == EMCHits)

  { iTracksTraced = 0;
    nMCHitsInactSens = 0;
    nMCHitsRead = 0;
    nMCHitsUnkSens = 0;
    nMCHitsUsed = 0.;
    nMCTracks = 0;
    nMCTracksCh = 0;
    nMCTracksUsed = 0.;
    nBadIndexMsgs = 0;
    nInactSensMsgs = 0;
    nUnkSensMsgs = 0;

    tracks = (TPhObjectContainer*) event->GetObject(fMCTracksName);
// Get access to HitArrays for statistics printout only
    hitarrays = (TPhHashContainer*)
      event->GetObject(fHitArrayName[fHitArrayId]);
    allHits = (TPhObjectContainer*) event->GetObject(fMCHitsName);

// Define diagnostic histograms and save the pointers in diagnostic
// storage for externally-controlled plotting.  If mm is the TPhMul
// object, use 
//   mm.ShowDiag(0) 
// to show an index of stored objects, and
//   TH1F *hh = (TH1F*)mm.GetDiag(n)
// to retrieve it.  Use hh->SetXTitle, hh->SetMarkerStyle, ... and
// hh->SetLabelFont(131,"X" or "Y"), hh->Draw("alpe1") or "pesame"...
//
    
// Distributions in eta of all, primary, and silicon-hitting charged tracks
    static TH1F *diag[kNDiagMax];
    Stat_t diagWt[kNDiagMax];
    if (NHBins[0] && !diag[0])
    { diag[0] = new TH1F ("dNdEtaMC",
        "dNch/dEta generated", NHBins[0], -6.25, 6.25);
      SetDiag(0, "TH1F *dNdEtaMC", diag[0]);
// Set the weight to 1 / (Eta-bin).  These must still be scaled 
// to the number of events using hh->Scale(1/Nevents), and the errors
// shown are not scaled.
// Set weights to 0 until I find out how to scale them.
      diagWt[0] = 1.0; // NHBins[0] / 12.5;
    }
    if (NHBins[1] && !diag[1])
    { diag[1] = new TH1F ("dNdEtaPrimMC", 
       "dNch/dEta generated, primaries", NHBins[1], -6.25, 6.25);
      SetDiag(0, "TH1F *dNdEtaPrimMC", diag[1]);
      diagWt[1] = 1.0; // NHBins[1] / 12.5;
    }
    if (NHBins[2] && !diag[2])
    { diag[2] = new TH1F ("dNdEtaPrimHitMC", 
       "dNch/dEta generated, primaries hitting silicon", NHBins[2], 
         -6.25, 6.25);
      SetDiag(0, "TH1F *dNdEtaPrimHitMC", diag[2]);
      diagWt[2] = 1.0; // NHBins[2] / 12.5;
    }
// Distributions in energy deposited in .3 mm silicon of all and primary tracks
    Float_t dEscale = 3.0e4; // 300 u * GeV/keV (Geant units)
    Float_t dElim = 1.0e6 * feMax;
    if (NHBins[3] && !diag[3])
    { diag[3] = new TH2F ("dE300MC", "MC dE in .3 mm Si",
        NHBins[3], 0., dElim, 27, -54.0, 54.0);
      SetDiag(0, "TH2F *dE300MC", diag[3]);
    }
    if (NHBins[4] && !diag[4])
    { diag[4] = new TH2F ("dE300PrimMC", "MC dE in .3 mm Si for primaries",
        NHBins[4], 0., dElim, 25, -50.0, 50.0);
      SetDiag(0, "TH2F *dE300PrimMC", diag[4]);
    }
// Profile distributions in energy deposited in .3 mm silicon of all and 
// primary tracks
//    Float_t dEscale = 3.0e4; // 300 u * GeV/keV (Geant units)
//    Float_t dElim = 1.0e6 * feMax;
/* Fails to compile
    if (NHBins[5] && !diag[5])
// The following fails with
// "  type `TH1F' is not a base type for type `TProfile"
    { diag[5] = new TProfile ("dE300MCp", "MC dE in .3 mm Si",
        NHBins[5], -50.0, 50.0);
      SetDiag(0, "TH2F *dE300MCp", diag[5]);
    }
    if (NHBins[6] && !diag[6])
    { diag[6] = new TProfile ("dE300PrimMCp", 
        "MC dE in .3 mm Si for primaries",
        NHBins[6], -50.0, 50.0);
      SetDiag(0, "TH2F *dE300PrimMCp", diag[6]);
    }
*/
    nMCHitsTot = (allHits != 0)? allHits->GetEntries() : 0;
    if (ReportLevel >= 3 || (ReportLevel >=1 && !nMCHitsTot))  
      printf("TPhMul/GetData: event %d has %d tracks, %d hits, "
        "%d hitarrays.\n", event->GetNEvent(), tracks->GetEntries(),
        nMCHitsTot, hitarrays->GetEntries());
    TIter TrackIter (tracks->GetIterator());
// The HitIter iterator must be outside the track loop, else it will
// restart at the beginning of the event buffer for each track
    TIter nextEventHit(allHits->GetIterator());
    TPhMCHit *Hits[MaxTrackHits];
    iprint = 0;
    nstored = 0; // Diagnostics only

// Loop on MC tracks
//
// The structure here will be either a loop over hits gotten from 
// as-yet unavailable calibrated hit arrays, for data or MC, or from 
// tracks, for MC only.  The tracks have to be separated into segments
// associated with different layers, and tracks crossing more than one
// pad in each layer weighted by path length in order to count MC
// hits correctly for calibrating energy deposition and secondary 
// fractions.

    nTracks = tracks->GetEntries();
// Limit  tracks for diagnostics
    if (nTracks > nTracksUse) nTracks = nTracksUse;
    for (iTrack = 0; iTrack < nTracks; iTrack++)

// Get a track and the number of hits it produced, limited to MaxTrackHits
    { track = (TPhMCTrack*) TrackIter();
      nMCTracks++;
      nHits = track->GetNHits();
// Select charged tracks.  This should probably be an if clause rather
// than setting the # hits to 0; but they should be the same in Geant.
      if (track->GetCharge() == 0) nHits = 0;
      if (nHits) 
      { nMCTracksCh++;
        nMCHitsRead += nHits;
      }
        
// Find the segments of the track that lie in distinct layers.

      const Int_t MaxSegs = 4; // Arbitrary
      Float_t lHit[MaxTrackHits];
      Float_t lSeg[MaxSegs];
//      Int_t layerNumber;
      Int_t iStart[MaxSegs];
      Int_t iSeg = -1;
      const Int_t kMaxSensChars = 4; // # of characters defining a sensor
      Char_t Layer[kMaxSensChars+1];
      Char_t CurrLayer[kMaxSensChars+1];
      Int_t iHit;
// Init layer name
      strncpy(Layer, "0000", kMaxSensChars);

//      iTrackTrace = -1; // SPECIAL TRACE

// Loop on hits produced by this track

      TPhMCHit *hit;
      Int_t nHitsUsed = 0;

      for (iHit = 0; iHit < nHits; iHit++)

      { hit = (TPhMCHit*) nextEventHit();
        sensName = "????";
        if (hit)
// The following statement assumes fDet points to a valid TPhDetector.
        { TPhSensor *ss = fDet->GetSensor(hit->GetSensorLabel());
          if (ss) sensName = (Char_t *)ss->GetName();
// Ignore empty hits, hits in inactive or non-existent sensors, and
// hits outside the storage limit
        }
        TPhPadMap *sens = 0;
        if (hit) sens = GetSensMap(sensName);
        if (sens && sens->GetId() && nHitsUsed < MaxTrackHits)
	{ strncpy(CurrLayer, sensName, kMaxSensChars);
          Hits[nHitsUsed] = hit;
//          if (!strncmp(CurrLayer, "MOBB", kMaxSensChars)) // SPECIAL TRACE
//            iTrackTrace = iTrack;
// A layer is characterized by the first 2 characters of the sensor name 
// for the octagon and 3 for the rings, vertex, and spectrometer.
// SENSOR NAMES ARE NOT GUARANTEED TO HAVE THIS PROPERTY.
// An error in the algorithm can either overcount or undercount the number
// of MC tracks.
// The hit's layer number starts with 0 for each group and so is not usable 
// for this purpose.
          Int_t NCharLayer = (!strncmp(CurrLayer, "MO", 2))? 2 : 3;
// Initialize new segment if layer id has changed unless there are too many
	  if (strncmp(CurrLayer, Layer, NCharLayer) && iSeg < MaxSegs-1)
          { strncpy (Layer, CurrLayer, NCharLayer);
            iSeg++;
            iStart[iSeg] = nHitsUsed;
            lSeg[iSeg] = 0.;
          }
// Get the hit's path length in silicon.
          lHit[nHitsUsed] = sqrt (pow (hit->GetDx(), 2) 
            + pow (hit->GetDy(), 2) + pow (hit->GetDz(), 2));
          lSeg[iSeg] += lHit[nHitsUsed];
          nHitsUsed++;
        } // if (sens &&
        else
        { if (!hit && ReportLevel >= 2 && iprint < Nprint)
          { printf ("TPhMul/Getdata: event %d, hit %d, track %d is empty.\n",
              event->GetNEvent(), nHitsUsed, iTrack);
            iprint++;
          }
// The following should be fixed to indicate ONLY hits on active sensors
// that were rejected because of the MaxTrackHits limit
          if (nHitsUsed == MaxTrackHits && ReportLevel >= 2 && iprint < Nprint)
          { printf ("TPhMul/GetData: track %d has %d hits, only %d used.\n",
              iTrack, nHits, nHitsUsed);
            iprint++;
          }
        }
      } // for (iHit

      nSeg = iSeg + 1;
      iSeg = 0;
      for (iHit = 0; iHit < nHitsUsed; iHit++)
      {
// Find segment containing this hit
//        while (iSeg < nSeg - 1 && iHit == iStart[iSeg+1]) iSeg++;
        if (iSeg < nSeg - 1 && iHit == iStart[iSeg+1]) iSeg++;

        hit = Hits[iHit];
	if (hit)
        { ix = hit->GetPadX();
          iy = hit->GetPadY();
          dEPad = hit->GetDe();

// Take the fractional number of tracks as the ratio of the pathlength
// of this hit to the total pathlength of the segment.
// If the pathlength and energy deposited are zero, ignore the hit.
// The real data analog occurs when zero suppression fails.

          if (lHit[iHit] && lSeg[iSeg])

          { fTrack = lHit[iHit] / lSeg[iSeg];

// Update diagnostic energy-deposited-per-track histograms
// These are presently for octagon only.

            TPhSensor *ss = fDet->GetSensor(hit->GetSensorLabel());
            if (ss) sensName = (Char_t *)ss->GetName();
            if ((diag[3] || diag[4]) && sensName[0] == 'M'
              && sensName[1] == 'O')
            { Float_t deNorm = dEPad / lHit[iHit] * dEscale;
              if (diag[3]) diag[3]->Fill(deNorm, hit->GetZ(), 1.0);
              if (diag[4] && track->GetParentNo() == 0)
                diag[4]->Fill(deNorm, hit->GetZ(), 1.0);
            }
            if ((diag[5] || diag[6]) && sensName[0] == 'M'
              && sensName[1] == 'O')
            { Float_t deNorm = dEPad / lHit[iHit] * dEscale;
              if (diag[5]) diag[5]->Fill(hit->GetZ(), deNorm);
              if (diag[6] && track->GetParentNo() == 0)
                diag[6]->Fill(hit->GetZ(), deNorm);
            }
          }

          else 

          { fTrack = 0.;
            if (dEPad && ReportLevel >= 2 && iprint < Nprint)
            { printf ("TPhMul/GetData: dE =%10.3e, dL/lSeg = %6.4f/%6.4f,"
                " for hit %d/%d,\n track %d, %s[%d][%d]\n", dEPad, 
                lHit[iHit], lSeg[iSeg], iHit,  nHitsUsed, iTrack, 
                sensName, iy, ix);
              iprint++;
            }
          }

// Diagnostics:

          if (iTrack == iTrackTrace
              || nBadIndexMsgs < fMaxBadIndexMsgs
              || nInactSensMsgs < fMaxInactSensMsgs
              || nUnkSensMsgs < fMaxUnkSensMsgs
              || (ReportLevel >= 5  && iprint < Nprint))
	  { static Int_t prevTrack = -1;
            Float_t DegPerRad = 57.29578;
            if (iTrack != prevTrack)
            { prevTrack = iTrack;
// Get the polar production angles.  Note the track->GetTheta() returns
// px/pz and GetDip returns py/pz, the Cartesian angles, which are not
// what we need for the octagon and ring geometry.
              Float_t px = track->GetPx();
              Float_t py = track->GetPy();
              Float_t pz = track->GetPz();
// Note the C++ atan2 doesn't try to get the quadrant correct; it just
// returns a value in [-pi/2, +pi/2]. 13-Jan-99: B. Wosiek says [-pi, pi].

              Float_t theta = DegPerRad * atan2 (sqrt(px*px + py*py), pz);
              if (theta < 0.) theta += 180.;
              Float_t phi = DegPerRad * atan2 (py, px);
              if (px < 0.) phi += 180.;
              else if (py < 0.) phi += 360.;
              if ((iTrack == iTrackTrace || (ReportLevel >= 5)
                && iprint < Nprint))
              { printf ("TPhMul/GetData evt %d, track %d, Gid %d,"
                  " E=%9.3e GeV, %d hits in %d segs,\n",
                  event->GetNEvent(), iTrack, track->GetGPid(),
                  track->GetEnergy(), nHitsUsed, nSeg);
                printf (" produced by parent track %d at XYZ = (%5.2f,"
                  "%5.2f,%7.2f) cm,\n", track->GetParentNo(),
                  track->GetXProd(), track->GetYProd(), track->GetZProd());
                printf (" theta =%7.2f (eta =%6.3f), phi =%7.2f deg:\n",
                  theta, -log(tan(0.5 / DegPerRad * theta)), phi);
                iprint++;
              }
            }
// Find the sensor and show the hit
            TPhPadMap *sens = GetSensMap(sensName);
            if (sens)
	    { Float_t etaPhi[2][3];
              sens->GetEtaPhi(ix + 0.5, iy + 0.5, &etaPhi[0][0]);
              sens->GetEtaPhi(ix + 1.5, iy + 1.5, &etaPhi[1][0]);
              if (iTrack == iTrackTrace
                  || (ReportLevel >= 5  && iprint < Nprint))
                printf (" Hit %d, segment %d in %s, dE=%7.1f keV,"
                  " dL=%6.3f/%5.3f cm, wt%5.2f track;\n", iHit, iSeg, 
                  sensName, 1.0e6 * dEPad, lHit[iHit], lSeg[iSeg], fTrack);
// Check the assigned bins for consistency with sens's geometry
              Int_t type = sens->GetType(); // Ring or barrel
              Float_t dx, dy, x, xmin, y, ymin;
// Ring annuli have ~ constant dEta: dEta = (Z * dRho - Rho * dZ) / (R * Rho)
// where Rho*Rho = X*X + Y*Y and R*R = Rho*Rho + Z*Z
              Int_t ixEx;
              xmin = sens->GetMin(0);
              if (type != kRing)
              { x = hit->GetZ();
                dx = sens->GetSize(0);
                if (dx) ixEx = int ( (x-xmin) / dx);
                else ixEx = -999;
	      }
              else
              { x = sqrt (hit->GetX() * hit->GetX() 
                        + hit->GetY() * hit->GetY());
                dx = log (1.0 + sens->GetSize(0));
                if (dx) ixEx = int (log (x / xmin) / dx);
                else ixEx = -999;
              }
              y = DegPerRad * atan2(hit->GetY(), hit->GetX());
              ymin = sens->GetMin(1);
              dy = sens->GetSize(1);
// Convert phi to [-22.5,-337.5] degrees for octagon, [0.360] for rings.
// This is for checking assigned indices only, and is a KLUDGE.
              if (y < 0. && strncmp(sens->GetName(), "MOA", 3)) y += 360.;
              Float_t yoff = y - ymin;
              Int_t iyEx = int( yoff / dy);
// Flag a discrepancy in indices.  This must be expanded to flag a discrepancy
// in track and sensor pad eta and phi, but that's trickier.
              Bool_t discrep = abs(ix - ixEx) > fXIndexErrTol
                            || abs(iy - iyEx) > fYIndexErrTol;
              if (discrep && nBadIndexMsgs < fMaxBadIndexMsgs)
              { printf ("TPhMul/GetData index error for track %d,"
                 " hit %d, on %s [%d/%d]:\n", iTrack, iHit, sensName,
                 nBadIndexMsgs+1, fMaxBadIndexMsgs);
                printf (" True XYZ = (%5.2f,%5.2f,%7.2f) cm, assigned/exp "
                  "ix = %d/%d, iy = %d/%d;\n",
                  hit->GetX(), hit->GetY(), hit->GetZ(), ix, ixEx, iy, iyEx);
                printf (" Pad center at eta=%6.3f, phi=%6.1f, pad dedp="
                  "%7.4f, dtdr = %7.4f.\n", etaPhi[0][0], etaPhi[0][1],
                  (etaPhi[1][0]-etaPhi[0][0]) * (etaPhi[1][1]-etaPhi[0][1]),
                   etaPhi[0][2]);
                nBadIndexMsgs++;
              }
            }
            else
// Tally inactive and unknown sensors
            { if (!sens)
              { nMCHitsUnkSens++;
                if (nUnkSensMsgs < fMaxUnkSensMsgs)
                { printf (" TPhMul/GetData: sensor %s not known [%d/%d].\n",
                  sensName, nUnkSensMsgs+1, fMaxUnkSensMsgs);
                  nUnkSensMsgs++;
                }
              }
              else
              { nMCHitsInactSens++;
                if (nInactSensMsgs < fMaxInactSensMsgs)
                { printf (" TPhMul/GetData: sensor %s is inactive [%d/%d].\n",
                  sensName, nInactSensMsgs+1, fMaxInactSensMsgs);
                  nInactSensMsgs++;
                }
              }
            } // if (sens
          } // if (iTrack == iTrackTrace

// Find the sensor map[s] in which this hit should be entered.

// Diagnostic: limit the number of hits stored
          /* if (nstored >= 1) dEPad = 0.; */

// Search for sensor by name.  This may eventually use a HitArray and
// a translation table from PhAT to Mul sensor index.

          TPhPadMap *sens = GetSensMap(sensName);
          if (sens && sens->GetId() != 0  && lHit[iHit])

// Accumulate the deposited energy if the indices are valid, the sensor 
// is active, and the path length is non-zero.
// If we were reading a true data stream, a non-zero existing pad energy 
// here would signal a data error.

          { if ( ix >= 0 && ix < sens->GetNCells(0) 
              && iy >= 0 && iy < sens->GetNCells(1)
              && sens->GetId())

            { *sens->GetLoc(ESEdep,ix,iy) += dEPad;
              nstored++;

// If data source is MC hits,

              if (fDataSource == EMCHits)

// ... accumulate MC tracks

              { *sens->GetLoc(ESNTracksMC,ix,iy) += fTrack;
                nMCHitsUsed++;
                nMCTracksUsed += fTrack;

// ... accumulate # of and energy deposited by primary MC tracks 
// (parent track id = 0)

                if (track->GetParentNo() == 0)
                { *sens->GetLoc(ESNTracksPrimMC,ix,iy) += fTrack;
                  *sens->GetLoc(ESEdepPrimMC,ix,iy) += dEPad;
                }

              } // if (fDataSource
        
            }

            else

// Issue warning if indices are invalid.  We MUST tabulate these.

              if (ReportLevel >= 2 && iprint < Nprint)
              { printf ("Mul/GetData: ignored bad data index: %s[%d][%d].\n",
                  sensName, ix, iy);
                iprint++;
              }
          } // if (sens
        } // if ((hit
      } // for iHit

// Update diagnostic Eta-distribution histo
// These need to be normalized
// Note that these are for all tracks; if we select tracks producing
// hits, the shape changes drastically

      if ((diag[0] || diag[1] || diag[2]) && track->GetCharge())
      { Float_t Eta = -log (tan (0.5 *
        atan2 ( sqrt(track->GetPx() * track->GetPx() + 
          track->GetPy() * track->GetPy()), track->GetPz())));
        if (diag[0]) diag[0]->Fill(Eta, diagWt[0]);
        if (diag[1] && track->GetParentNo() == 0) 
          diag[1] ->Fill(Eta, diagWt[1]);
        if (diag[2] && track->GetParentNo() == 0 && nHitsUsed > 0) 
          diag[2] ->Fill(Eta, diagWt[2]);
        if (ReportLevel >= 10 && iprint < Nprint) 
	{ printf ("Track %d: eta = %5.2f.\n", iTrack, Eta);
          iprint++;
        }
      }
      if (iTrack == iTrackTrace && iTracksTraced < nTracksTrace)
      { iTrackTrace += 1 + iTraceSkip;
        iTracksTraced++;
      }
    } // for (itrack

// Diagnostic summary: this should include a note of neutrals ignored.

    if (ReportLevel >= 3)
      printf ("Mul/GetData: %d MC tracks read, %d w/hits, %6.1f entered in "
        "sensor maps.\n             %d MC hits: %d read, %d inactive, %d "
        "unknown, %6.1f used.\n",
        nMCTracks, nMCTracksCh, nMCTracksUsed, nMCHitsTot, nMCHitsRead, 
        nMCHitsInactSens, nMCHitsUnkSens, nMCHitsUsed);

// Special diagnostic:

    if (ReportLevel >= 9)
    { int imap = GetSensMapIndex("MOB6");
      int ibin = 10;
      int nbins = 10;
      for (int ib = ibin; ib < ibin + nbins; ib++)
      { printf ("Mul/GetData: %s[0:3][%d] have", Maps[imap]->GetName(), ib);
        for (int kk=0; kk<4; kk++)
          printf (" %4.2f", *Maps[imap]->GetLoc(ESNTracksMC, ib, kk));
        printf (" MC hits.\n");
      }
    }
  } // if (isens && fDataSource

// The following will change to a loop over all listed sensors and 
// the isens argument will be removed:

  if (fDataSource == EDiagnostic)
  { Maps[isens]->GetEtaPhi (0., 0., epd);
    etaLim[1] = epd[0];
    Float_t *Lx, *Lxy;
// Loop over all x-pads
    Lx = Maps[isens]->GetDLoc (0, 0, 0, dloc);
    for (ix = 0; ix < nx; ix++)
    { Lxy = Lx;
      Lx += dloc[1];    
      etaLim[0] = etaLim[1];
      Maps[isens]->GetEtaPhi (ix+1.0, 0., epd);
      etaLim[1] = epd[0];
      if (CheckEtaCalc)
        if (ReportLevel >= 2 && iprint < Nprint)
        { theta = (type==kBarrel)? atan2(dist,x) : atan2(x,dist);
          eta = -log(tan(0.5 * theta));
          printf ("For %s bin %d, eta =%8.5f, slow calc =%8.5f.\n",
            Maps[isens]->GetName(), ix, etaLim[0], eta); 
          x = (type==kBarrel)? x + dx : x * dx;
          iprint++;
        }
      dEta = fabs ( (Double_t)(etaLim[1] - etaLim[0]));
      dtdr = epd[2];
// Calculate sham dNdy, in 1/degrees, corrected for oblique path length
      const Float_t  PhiTot = 360.; // Full range of y
      Double_t abseta = fabs( (Double_t) etaLim[0]);
      Float_t arg = (abseta - eta0) / sigeta; 
      Float_t dNdy = dNdE0 * exp( -0.5 * arg * arg) * dEta / dtdr / PhiTot;

// Loop over all y-pads for this x assuming dt/dr is independent of phi(y) 
      ylim[1] = ymin;
      for (iy = 0; iy < ny; iy++)
      { ylim[0] = ylim[1];
        ylim[1] = ylim[0] + dy;
// Throw total # hits for this pad, including secondaries
        Float_t nhits0 = dNdy * fabs(dy) * fs.GetFsec(isens, etaLim[0]);
// Approximate a Poisson width with a square width, limited so that it 
// doesn't skew nhits.  This horrendous approximation should be flaged.
        Float_t u = rand() / Rrmax;
        Float_t nhits
          = (nhits0 > 3.0)? nhits0 + (u - 0.5) * sqrt(12.0 * nhits0)
          : 2.0 * u * nhits0;
        if (ReportLevel >= 1 && iprint < Nprint && nhits < 0.)
        { printf ("Generated%5.2f hits for %s[%3d,%3d] at eta =%6.2f,",
            nhits, Maps[isens]->GetName(), ix, iy, epd[0]);
          printf (" phi =%6.2f, dtdr =%6.4f\n", epd[1], epd[2]);
          printf (" nhits0 = %f, Fsec = %f\n", nhits0, 
            fs.GetFsec(isens, etaLim[0]));
          iprint++;
        }
        e1 = de.GetDept(isens, etaLim[0]);
        edep = nhits * e1;
        //*Maps[isens]->GetLoc(0,ix,iy) = edep; // Slow form
        *Lxy = edep;
        if (ReportLevel >= 1 && iprint < Nprint && edep < 0)
        { printf ("Mul/GetData error: edep = %f for %s[%d][%d].\n", edep,
            Maps[isens]->GetName(), ix, iy);
          iprint++;
        }
        Lxy += dloc[2]; 
      } // for iy
    } // for ix
  } // if (fDataSource
}

////////////////////////////////////////////////////////////////////////////

//void *TPhMul::GetDiag(Int_t id) { return fDiagnostics[id-1]; }
  // Return diagnostic pointer # id = 1, 2, ...

////////////////////////////////////////////////////////////////////////////

//Char_t *TPhMul::GetDiagName(Int_t id) { return fDiagNames[id-1]; }
  // Return name of diagnostic # id = 1, 2, ...

////////////////////////////////////////////////////////////////////////////

TPhPadMap *TPhMul::GetDisplay (Int_t idisp)
{ return (idisp >= 0 && idisp < kNDispMax)? Disps[idisp] : (TPhPadMap *)0;
  // note that (TPhPadMap *)NULL returns 0 under CInt but not in gcc;
}

////////////////////////////////////////////////////////////////////////////

TPhPadMap *TPhMul::GetDisplay (Char_t *name)
{ Int_t idisp = GetDisplayIndex(name);
  return (idisp >= 0)? Disps[idisp] : (TPhPadMap *)0;
}

////////////////////////////////////////////////////////////////////////////

Int_t TPhMul::GetDisplayIndex (Char_t *name)
{ Int_t idisp;
  for (idisp = 0; idisp < nDispAct; idisp++)
    if (Disps[idisp] && !strcmp(name, Disps[idisp]->GetName())) break;
  return (idisp < nDispAct)? idisp : -1;
}

////////////////////////////////////////////////////////////////////////////

Float_t TPhMul::GetF(Char_t *name, Int_t arrIndex)
{ Int_t &ReportLevel = RLGetF;
  Int_t CaseSensitive = 0;
  Float_t *vp;
  if (!strcmp(name, "feMax") || (!CaseSensitive && !strcmp(name, "femax")))
    { vp = &feMax + sizeof(float) * (arrIndex - 1); // WRONG
    if (ReportLevel >= 3)
      printf ("Mul/GetF: %s[%d] = %12.4e.\n", name, 0, *vp);
    return *vp;
  }
  else
  { if (ReportLevel >= 1) printf ("Mul/GetF: variable %s unknown.\n", name);
    return 0;
  }
}

////////////////////////////////////////////////////////////////////////////

TGraphErrors *TPhMul::GetGraph (Char_t *typeName, Char_t *selector,
  Int_t iVar, Int_t iVarErr, Float_t norm, Int_t nXBins, Int_t nYBins,
  Int_t nXMinSkip, Int_t nYMinSkip, Int_t nXMaxSkip, Int_t nYMaxSkip)
{ TPhPadMap **ms;
  Int_t nMaps;
  if (typeName[0] == 's' || typeName[0] == 'S')
  { ms = Maps;
    nMaps = nSensAct;
  }
  else
  { ms = Disps;
    nMaps = nDispAct;
  }
  return ms[0]->GetGraph(ms, nMaps, typeName, selector, -1, iVar, iVarErr,
     norm, nXBins, nYBins, nXMinSkip, nYMinSkip, nXMaxSkip, nYMaxSkip);
}

////////////////////////////////////////////////////////////////////////////

TGraphErrors *TPhMul::GetGraph (Char_t *typeName, Int_t iMap,
  Int_t iVar, Int_t iVarErr, Float_t norm, Int_t nXBins, Int_t nYBins,
  Int_t nXMinSkip, Int_t nYMinSkip, Int_t nXMaxSkip, Int_t nYMaxSkip)
{ TPhPadMap **ms;
  Int_t nMaps;
  if (typeName[0] == 's' || typeName[0] == 'S')
  { ms = Maps;
    nMaps = nSensAct;
  }
  else
  { ms = Disps;
    nMaps = nDispAct;
  }
  return ms[0]->GetGraph(ms, nMaps, typeName, "", iMap, iVar, iVarErr,
     norm, nXBins, nYBins, nXMinSkip, nYMinSkip, nXMaxSkip, nYMaxSkip);
}

////////////////////////////////////////////////////////////////////////////

TH2F *TPhMul::GetHist (Char_t *typeName, Char_t *selector,
  Int_t iVar, Int_t iVarErr, Float_t norm, Int_t nXBins, Int_t nYBins,
  Int_t nXMinSkip, Int_t nYMinSkip, Int_t nXMaxSkip, Int_t nYMaxSkip)
{ TPhPadMap **ms;
  Int_t nMaps;
  if (typeName[0] == 's' || typeName[0] == 'S')
  { ms = Maps;
    nMaps = nSensAct;
  }
  else
  { ms = Disps;
    nMaps = nDispAct;
  }
  return ms[0]->GetHist(ms, nMaps, typeName, selector, -1, iVar, iVarErr,
  norm, nXBins, nYBins, nXMinSkip, nYMinSkip, nXMaxSkip, nYMaxSkip);
}

////////////////////////////////////////////////////////////////////////////

TH2F *TPhMul::GetHist (Char_t *typeName, Int_t iMap,
  Int_t iVar, Int_t iVarErr, Float_t norm, Int_t nXBins, Int_t nYBins,
  Int_t nXMinSkip, Int_t nYMinSkip, Int_t nXMaxSkip, Int_t nYMaxSkip)
{ TPhPadMap **ms;
  Int_t nMaps;
  if (typeName[0] == 's' || typeName[0] == 'S')
  { ms = Maps;
    nMaps = nSensAct;
  }
  else
  { ms = Disps;
    nMaps = nDispAct;
  }
  return ms[0]->GetHist(ms, nMaps, typeName, "", iMap, iVar, iVarErr,
  norm, nXBins, nYBins, nXMinSkip, nYMinSkip, nXMaxSkip, nYMaxSkip);
}

////////////////////////////////////////////////////////////////////////////

//Int_t TPhMul::GetNDiag() { return fNumberDiags; } 
 // Number of diagnostic entries

////////////////////////////////////////////////////////////////////////////

TPhPadMap *TPhMul::GetSensMap (Int_t isens)
{ return (isens >= 0 && isens < kNSensMax)? Maps[isens] : (TPhPadMap *)0;
}

////////////////////////////////////////////////////////////////////////////

TPhPadMap *TPhMul::GetSensMap (Char_t *name)
{ Int_t isens = GetSensMapIndex(name);
  return (isens >= 0)? Maps[isens] : (TPhPadMap *)0;
}

////////////////////////////////////////////////////////////////////////////

Int_t TPhMul::GetSensMapIndex (Char_t *name)
{ Int_t isens;
  for (isens = 0; isens < nSensAct; isens++)
    if (Maps[isens] && !strcmp(name, Maps[isens]->GetName())) break;
  return (isens < nSensAct)? isens : -1;
}

////////////////////////////////////////////////////////////////////////////

TPhVertex *TPhMul::GetVertex (void) {return fVertex; }

////////////////////////////////////////////////////////////////////////////

void TPhMul::ListDisplays (Int_t idMin, Int_t idMax)
{ const Int_t nPerLine = 6;
  Int_t nn = 0;
  if (idMin < 0) idMin = 0;
  if (idMax > kNDispMax) idMax = kNDispMax;
  for (Int_t idisp = idMin; idisp < idMax; idisp++)
  { TPhPadMap *disp = Disps[idisp];
    if (disp)
    { printf ("%4d: %7s", idisp, disp->GetName());
      if (nn++ > 0 && nn % nPerLine == 0) printf ("\n"); 
    }
  }
  if (nn % nPerLine != 0) printf ("\n"); 
}

////////////////////////////////////////////////////////////////////////////

void TPhMul::ListSensors (Int_t idMin, Int_t idMax)
{ const Int_t nPerLine = 8;
  Int_t nn = 0;
  if (idMin < 0) idMin = 0;
  if (idMax > kNSensMax) idMax = kNSensMax;
  for (Int_t isens = idMin; isens < idMax; isens++)
  { TPhPadMap *sens = Maps[isens];
    if (sens)
    { printf ("%4d: %s", isens, sens->GetName());
      if (nn++ > 0 && nn % nPerLine == 0) printf ("\n"); 
    }
  }
  if (nn % nPerLine != 0) printf ("\n"); 
}

////////////////////////////////////////////////////////////////////////////

void TPhMul::Process (TPhEvent *event, Int_t mode)
{
// Process multiplicity-derived quantities for one event.
// Specify mode = 
//    0 to initialize if first call and then process event data, or
//   -1 to initialize but not process event data (this mode allows 
//      the caller to modify parameters after initialization),
//    1 to process event data after bypassing initialization.
//
// Date     Vers  Description
// 02-Jun-98 0.1  Generalize mode argument to allow geometry updates (RV)
// 08-Nov-97 0.1  Add pre-clearing of single-event data (RV)
// 03-Nov-97 0.1  Creation (Robin Verdier)

// Define local variables:

  TPhEvent *fEvent = event; // Not used at present

// Number of variables in each sensor pad map.  For data, we keep only 
// the deposited energy, but may later accumulate energies over many 
// events.  For MC, we keep also the number of hits from primary and 
// from all tracks and the energy deposition per primary track.

  Int_t NVars = ESNVmax;

// Number of variables for display.  The variables are enumerated in
// TPHMul.h.

  Int_t NDVars = EDNVmax;

// Range of sensors to be processed, in 0...kNSensMax-1, where 
// const Int_t kNSensMax is defined in TPhMul.h

  static Int_t FirstSensor = 0;
  static Int_t LastSensor = kNSensMax-1;

// Report level: 0 for no output, 1 for fatal error warnings only, 
// 2, 3, ... for increasing amounts of diagnostic printout.

  Int_t &ReportLevel = RLProcess;
  Int_t &iprint = RLiprint;
  Int_t  ncells; // Return flag for cleared displays
  Int_t &Nprint = RLNPrint;

// Get the detector object, if not already located

  if (!fDet)
    if (!(fDet = TPhDetector::GetInstance()))
      fDet = TPhDetector::ReadLatestGeometry();

// Create a sensor definition object.  This gets geometry parameters 
// either from the detector object or from defaults in a TPhSensGeo
// object.
// I would prefer to instantiate de and fs inside the initialization 
// block, but they would then be out of scope later.  ==> Study this.

  if (ReportLevel >= 5 && iprint < Nprint)
  { printf ("Mul/Process: ready to instantiate TPhSensDefs.\n");
    iprint++;
  }

  fSensDefs = new TPhSensDef;
  TPhSensDef &sd = *fSensDefs;
  nSensAct = sd.GetNDefsAct();

// Instantiate dE-per-track and fraction-of-secondaries classes
// THE FOLLOWING INSTANTIATIONS ARE NOTED BY P. DECOWSKI TO BE
// VERY DANGEROUS, SINCE THEY COULD GO OUT OF SCOPE BEFORE DELETION.

  static TPhDept de(sd);
  fDept = &de;

  static TPhFsec fs(sd);
  fFsec = &fs;

// Initialize if mode = -1 or if not already done.

  if (mode < 0 || fInitMode == 0)
  {

// Instantiate the set Maps of sensor pad map objects, named like
// their corresponding sensors.  They will be zeroed later.

    Int_t nn; // Avoid scope difficulties
    for (nn = 0; nn < nSensAct; nn++)
    { Maps[nn] = new TPhPadMap(nn, sd.GetName(nn), &sd, NVars, 0);
// Set the maximum number of events, fMaxEvents, to accumulate before 
// clearing each sensor map to the "Id" field in the SensGeo; the former 
// convention that fMaxEvents is set to 1 if the name ends with -1 is no 
// longer honored.  fMaxEvents = -1 means accumulate throughout the run.
      Maps[nn]->SetMaxEvents(Maps[nn]->GetId());
// Force clearing on the first event
      Maps[nn]->SetNEvents(Maps[nn]->GetMaxEvents());
    }

// Create a display definitions object.  This gets display parameters 
// from a variant form of the SensGeo object.  It will be extended
// to include ranges of detectors and other information.

    fDispDefs = new TPhDispDef;
    TPhDispDef &dd = *fDispDefs;
    nDispAct = dd.GetNDefsAct();

// Create and zero a set of the maximum number of physics displays allowed
// by storage limits, defined in TPhDispDef.h.
// These may eventually be loadable from an Options object.
// See the above remark concerning fMaxEvents.

    for (nn = 0; nn < nDispAct; nn++)
    { Disps[nn] = new TPhPadMap(nn, dd.GetName(nn), &dd, NDVars, 0);
      Disps[nn]->SetMaxEvents(Disps[nn]->GetId());
      Disps[nn]->SetNEvents(Disps[nn]->GetMaxEvents());
    }

    if (ReportLevel >= 4 && iprint < Nprint)
    { printf ("Mul/Process: %d / %d sensors and %d / %d displays active.\n",
        nSensAct, sd.GetNDefsMax(), nDispAct, dd.GetNDefsMax());
      iprint++;
    }

// Diagnostic: test GetEtaPhi function
    if (ReportLevel >= 4)
    { Float_t ep[3];
      Int_t isens = 11;
      Maps[isens]->GetEtaPhi(3.4, 5.7, ep);
      printf ("Mul/Process: for %s cells [3.4,5.7], eta =%7.4f, phi =%7.2f"
        ", dtdr =%11.4e.\n", Maps[isens]->GetName(), ep[0], ep[1], ep[2]);
    }

    fInitMode = 1;

  } // if (mode

  if (mode >= 0)

  {

// Process event data in mode 0 or 1:

// Set the vertex position and Zero sensor data maps.
// Each map specifies how many events it should accumulate before being 
// cleared; the sensor data maps should not be accumulated for multiple
// events unless the vertices are fixed or known and used to shift 
// to vertex-0 equivalent storage locations..

    if (ReportLevel >= 5 && iprint < Nprint)
    { printf ("Mul/Process: ready to clear %d sensor TPhPadMaps"
       " in mode %d.\n", nSensAct, mode);
      printf ("Maps[0] = %p, maps[%d] = %p\n", Maps[0], nSensAct-1,
        Maps[nSensAct-1]);
      printf ("ready to execute Maps[0]->GetNEvents()\n");
      printf ("Result was %d\n", Maps[0]->GetNEvents());
      iprint++;
    }

    Int_t isens;
    for (isens = 0; isens < nSensAct; isens++)
    { Int_t nev = Maps[isens]->GetNEvents();
      if (nev == Maps[isens]->GetMaxEvents())
      { ncells = Maps[isens]->SetData(0.);
        nev = 0;
      }
      Maps[isens]->SetNEvents(nev+1);
      Maps[isens]->SetVertex(fVertex);
    }

// Set the vertex position and zero per-event display maps.  

    if (ReportLevel >= 5 && iprint < Nprint)
    { printf ("Mul/Process: ready to clear %d display TPhPadMaps"
       " in mode %d.\n", nDispAct, mode);
      iprint++;
    }

    for (Int_t idisp = 0; idisp < nDispAct; idisp++)
    { Int_t nev = Disps[idisp]->GetNEvents();
      if (nev == Disps[idisp]->GetMaxEvents())
      { ncells = Disps[idisp]->SetData(0.);
        nev = 0;
      }
      Disps[idisp]->SetNEvents(nev+1);
      Disps[idisp]->SetVertex(fVertex);
    }

    if (ReportLevel >= 5 && iprint < Nprint)
    { printf ("Mul/Process: ready to get sensor data in mode %d.\n",
        mode);
      iprint++;
    }

// Loop over specified (normally all) sensors

    for (isens = FirstSensor; isens <= LastSensor; isens++)

// Get energy deposited in this sensor and store it in the first 
// variable layer of the map.  The Dept and Fsec arguments de and 
// fs are for diagnostic use only.

    { GetData (isens, fEvent, Maps, de, fs);

// Map reconstructed variables onto physics displays

      Recon (isens, Maps, de, fs, Disps);

    }

// Normalize the reconstructed occupancy and multiplicity for the 
// event

    if (ReportLevel >= 9 && iprint < Nprint)
    { printf ("Mul/Process: ready to reconstruct data in mode %d.\n",
        mode);
      iprint++;
    }

    Recon (Disps);

  } // if (mode
}

////////////////////////////////////////////////////////////////////////////

void TPhMul::Recon (Int_t isens, TPhPadMap **Maps, TPhDept &de, 
                    TPhFsec &fs, TPhPadMap **Disps)
{
// Reconstruct Phobos occupancy and multiplicity for sensor # isens
// from raw energies in sensor map Maps, and store physics variables
// in all relevant display maps in the list Disps.
// Date     Vers  Description
// 11-Aug-98 1.0  Pass vertex position to Disps (RV)
// 11-Jun-98 1.0  Normalize Edep for use in fitting (RV)
// 19-Feb-98 0.2  Create display list using TPhPadMap::Select (RV)
// 22-Jan-98 0.2  Get display list from TPhPadMap (RV)
// 10-Dec-97 0.1  Correct feMax cut (RV)
// 26-Nov-97 0.1  Add MC variables (RV)
// 30-Oct-97 0.1  Creation (Robin Verdier)

// Global variables:
 // TEMPORARY! THESE MUST BE GLOBALS, ARGUMENTS, OR DATA MEMBERS OF CLASSES

  const Int_t NDVars = EDNVmax;

// Local variables:

// The vector Vars will become a struct.
// Indices of variables that can be mapped to display bins such as
// dEtadPhi and d2NdEtadPhi, are enumerated in TPhMul.h; they may
// be replaced by data fields in the struct.

  Float_t   Vars[NDVars];

//  Float_t  *dd = Maps[isens]->GetLoc(0,0,0); // Pointer to data
//  Float_t   dist = Maps[isens]->GetDist(); // Normal distance to plane
  Float_t   dtdr; // Normal incidence correction = sin (cos) Theta
//  Float_t   dy   = Maps[isens]->GetSize(1); // Y-cell size
  Float_t   eta;
  Float_t   Cell[2][2][2]; // [XY|EtaPhi][X/Eta min|max][Y/Phi min|max] 
  Float_t   EtaPhi[3];
//  Int_t     ic, nc; // Cell index variables
  Int_t     ix, iy; // Cell index variables
  Int_t     ny   = Maps[isens]->GetNCells(1); 
  Int_t    &ReportLevel = RLRecon1;
  Float_t   xc; // Indices for addressing (possibly non-integer) X-
  Float_t   yc; // and Y-cell positions
//  Float_t   ymin = Maps[isens]->GetMin(1); 
  Int_t    &iprint = RLiprint;
  Int_t    &Nprint = RLNPrint;

// Create list dispUp of displays to be updated with this sensor's data,
// and set the vertex position for eta-calculations
// A sensor updates a display if the sensor's name is accepted by the
// display name and the display number-of-events-to-accumulate is non-zero.
// TPhPadMap::Select describes the acceptance rules.  The number of 
// events called its Id for historical reasons.

  Int_t dispUp[kNDispMax+1]; // List of display maps to update
  Int_t nDisps = fDispDefs->GetNDefsAct(); // Number of active displays
  Int_t nDispUp = 0; // Number of display maps to update
  Char_t *SensName = Maps[isens]->GetName(); // Sensor name

  for (Int_t idisp = 0; idisp < nDisps; idisp++)
    if (TPhPadMap::Select(SensName, Disps[idisp]->GetName())
      && Disps[idisp]->GetId() != 0)
    { dispUp[++nDispUp] = idisp;
    }
  dispUp[0] = nDispUp;

  if (ReportLevel >= 6 && iprint < Nprint)
  { printf ("Recon: sensor %s updates %d displays: ", SensName, nDispUp);
    for (int nn = 1; nn <= nDispUp; nn++) printf ("%s, ",
      Disps[dispUp[nn]]->GetName());
    printf ("\n");
    iprint++;
  }

// Loop over X-cells:

  xc = 0.;
  Int_t nx = (nDispUp)? Maps[isens]->GetNCells(0) : 0;

  for (ix = 0; ix < nx; ix++)

  { Vars[nPads] = 1;

// Loop over Y-cells

    yc = 0.;

    for (iy = 0; iy < ny; iy++)

    {
      if (ReportLevel >= 5 && iprint < Nprint && ix == nx-1 && iy == ny-1)
        printf ("Mul/Recon calc for %s[%d][%d].\n",
        Maps[isens]->GetName(), ix, iy);

// Get eta and dt/dr for this pad, or X-Y cell, as a function of both 
// X- and Y-variables.
// The GetEtaPhi method will be optimized later to re-use recently- 
// -calculated values of eta, to avoid unnecessary recalculation of
// logarithms.  The implementation here is very inefficient, and 
// also ugly; Maps should contain all the mapping methods.

// Calculate both cell limits at the initial point.

      Maps[isens]->GetEtaPhi (xc, yc, EtaPhi);
      eta = EtaPhi[0];
      dtdr = EtaPhi[2];
      Cell[1][0][0] = EtaPhi[0]; // Eta
      Cell[1][1][0] = EtaPhi[1]; // Phi
      yc += 1.0;
      Maps[isens]->GetEtaPhi (xc+1., yc, EtaPhi);
      Cell[1][0][1] = EtaPhi[0];
      Cell[1][1][1] = EtaPhi[1];
// Take eta and dtdr at bin center
      eta = 0.5 * (eta + EtaPhi[0]);
      dtdr = 0.5 * (dtdr + EtaPhi[2]);

    if (ReportLevel >= 8 && iprint < Nprint && isens <= 1)
    { printf ("Mul/Recon: %s[%3.0f][%1.0f] has eta=[%7.4f,%7.4f],",
        Maps[isens]->GetName(), xc, yc, Cell[1][0][0], Cell[1][0][1]);
      printf (" phi=[%5.2f,%5.2f], dtdr=%6.4f.\n",
        Cell[1][1][0], Cell[1][1][1], dtdr);
      iprint++;
    }

// Get the phase space dEta * dPhi for this cell.

      Vars[dEtadPhi] = fabs((Cell[1][0][1] - Cell[1][0][0]) 
        * (Cell[1][1][1] - Cell[1][1][0]));

// Get the total energy deposited in this sensor pad, scaled to normal
// incidence and limited to feMax [which will later be adjusted for occupancy.
// The locations in the sensor pad maps are enumerated in TPhMul.h.

      Float_t eNorm = *Maps[isens]->GetLoc(ESEdep,ix,iy) * dtdr;
      Float_t eLim = (eNorm <= feMax)? eNorm : feMax;
      Vars[Edep] = eLim;

// The index lookups using GetLoc will be replaced by pointer increments 
// using GetDLoc to index safely and efficiently over all pads in a sensor.

// Get the global energy deposition per track, E1, for this sensor and eta:

      Vars[E1] = de.GetDept(isens, eta);

// Reconstruct the # of hits on this pad = Edep / E1, converted to normal 
// incidence by multiplying by dtdr, which assumes that the trajectory 
// points directly away from the vertex.

      if (Vars[E1] > 0) 
      { Vars[nTracks] = eNorm / Vars[E1];

// Reconstruct the # of hits for energy loss limited to feMax.  Note we
// could alternatively ignore pads with large energy.

        Vars[nTracksLim] = eLim / Vars[E1];
      }
      else if (ReportLevel >= 2 && iprint < Nprint)
      { printf ("Mul/Recon found Edep = %5.2f for %s[%3d,%3d] at eta ="
          "%6.2f.\n", Vars[E1], Maps[isens]->GetName(), ix, iy, eta);
        iprint++;
       }

// Global fraction of total hits / primary hits:

      Vars[FSec] = fs.GetFsec(isens, eta);

// Report negative values as a warning

      if (ReportLevel >= 4 && iprint < Nprint && Vars[FSec] < 0.)
      { printf ("Mul/Recon found FSec = %5.2f for %s[%3d,%3d] at eta ="
          "%6.2f.\n", Vars[FSec], Maps[isens]->GetName(), ix, iy, eta);
        iprint++;
       }

// Reconstruct the # of primary hits = nTracksLim / FSec:

      Vars[nTracksPrim] = Vars[nTracksLim] / Vars[FSec];

// Weight the global averages by phase space:

      Vars[E1] *= Vars[dEtadPhi];
      Vars[FSec] *= Vars[dEtadPhi];

// Monte Carlo variables:

      if (fDataSource != EData)

      { Vars[nTracksMC] = *Maps[isens]->GetLoc(ESNTracksMC,ix,iy);
        Vars[nTracksPrimMC] = *Maps[isens]->GetLoc(ESNTracksPrimMC,ix,iy);
// Normalize and limit the primary energy deposition for comparison with 
// the total
        Float_t eNorm = *Maps[isens]->GetLoc(ESEdepPrimMC,ix,iy) * dtdr;
        Float_t eLim = (eNorm <= feMax)? eNorm : feMax;
        Vars[EdepPrimMC] = eLim;
      }

// Normalized variables, such as occupancy, dNdEta, and their errors, must
// be calculated by the Map method after reconstructed data has been 
// accumulated for all pads because the dNdEta ranges of sensors overlap.

// Diagnostic printout: use to trace progress for numerical errors

      if (ReportLevel >= 5 && iprint < Nprint    //  && Vars[Edep]) 
          && isens >= 0 && ix == 0 && iy == 0)
      { eta = 0.5 * (Cell[1][0][0] + Cell[1][0][1]);
        printf ("Mul/Recon found %5.2f hits for %s[%3d,%3d] at eta =%6.2f.\n",
          Vars[nTracksLim], Maps[isens]->GetName(), ix, iy, eta);
        iprint++;
      }
      if (ReportLevel >= 5 && iprint < Nprint    //  && Vars[Edep]) 
          && isens >= 0 && ix == 0 && iy == 0)
      { printf ("Mul/Recon: vars =%9.3e, %7.1e, %9.3e, %9.3e, %9.3e, %9.3e,\n",
          Vars[dEtadPhi], Vars[nPads], Vars[Edep], Vars[E1], Vars[nTracks],
          Vars[nTracksLim]);
        printf ("  %9.3e, %9.3e, %9.3e, %9.3e, %9.3e\n", 
          Vars[FSec], Vars[nTracksPrim], Vars[nTracksMC], Vars[nTracksPrimMC],
		Vars[EdepPrimMC]);
        iprint++;
       }

// Project this pad onto physics displays

      Maps[isens]->Map (Vars, isens, ix, iy, Cell, dispUp, Disps);

    } // end of Y-loop

    xc += 1.;

  } // end of X-loop

}

////////////////////////////////////////////////////////////////////////////

void TPhMul::Recon (TPhPadMap **Disps)
{
// Normalize the reconstructed Phobos occupancy and multiplicity
// Date     Vers  Description
// 18-Oct-97 0.1  Correct EdepPTMC = MC dE / track (RV)
// 28-Oct-97 0.1  Creation (Robin Verdier)

// Global variables:
// TEMPORARY! THESE MUST BE GLOBALS, ARGUMENTS, OR DATA MEMBERS OF CLASSES
//  Int_t NDVars = EDNVmax;

// Local variables:

// Indices of variables that can be mapped to display bins such as
// dEtadPhi and d2NdEtadPhi, are enumerated in TPhMul.h

  Int_t &iprint = RLiprint;
  Int_t &Nprint = RLNPrint;
  Int_t &ReportLevel = RLRecon2;

  Int_t idisp, ieta, iphi, nEta, nPhi;

  Float_t *dd, *de, *dn, *dp, *dq, *dt;
  Float_t nt;

// Normalizing will only be done for variables that exist, using a 
// dynamically specified list for each display
 
// Get the number of active displays from TPhDispDef::GetNDefsAct.

  Int_t nDisps = fDispDefs->GetNDefsAct();

// Loop over displays

  for (idisp = 0; idisp < nDisps; idisp++)
  { nEta = Disps[idisp]->GetNCells(0);
    nPhi = Disps[idisp]->GetNCells(1);

// Loop over eta bins

    for (ieta = 0; ieta < nEta; ieta++)
    {
// Loop over phi bins
      for (iphi = 0; iphi < nPhi; iphi++)
      {
// Normalize...
// ... average energy deposited per pad
        dn = Disps[idisp]->GetLoc(Edep, ieta, iphi);
        dp = Disps[idisp]->GetLoc(nPads, ieta, iphi); // will be used below
        if (dp != 0 && dn != 0 && *dp && *dn)
        { dq = Disps[idisp]->GetLoc(EdepPPad, ieta, iphi);
          if (dq) *dq = *dn / *dp;
        }

// ... average global energy per track weighted by phase space
        dn = Disps[idisp]->GetLoc(E1, ieta, iphi);
        dd = Disps[idisp]->GetLoc(dEtadPhi, ieta, iphi);
        if (dd != 0 && dn != 0 && *dd && *dn)
        { dq = Disps[idisp]->GetLoc(E1Avg, ieta, iphi);
          if (dq) *dq = *dn / *dd;
        }

// ... average total / primary fraction weighted by phase space
        dn = Disps[idisp]->GetLoc(FSec, ieta, iphi);
        if (dd != 0 && dn != 0 && *dd && *dn)
        { dq = Disps[idisp]->GetLoc(FSecAvg, ieta, iphi);
          if (dq) *dq = *dn / *dd;
        }

// ... occupancy = tracks per pad
        dn = Disps[idisp]->GetLoc(nTracksLim, ieta, iphi);
        if (dp != 0 && dn != 0)
          if (*dp)
          { nt = *dn;
            dq = Disps[idisp]->GetLoc(Occ, ieta, iphi);
            if (dq) *dq = *dn / *dp; 

// ... Poisson error in occupancy
            dq = Disps[idisp]->GetLoc(OccRMS, ieta, iphi);
            if (dq) *dq = sqrt(nt) / *dp;
          } 

// ... Multiplicity = sum(primary tracks) / sum(dEta*dPhi)
        dn = Disps[idisp]->GetLoc(nTracksPrim, ieta, iphi);
        dd = Disps[idisp]->GetLoc(dEtadPhi, ieta, iphi);
        if (dd != 0 && dn != 0)
          if (*dd)
          { nt = *dn;
            dq = Disps[idisp]->GetLoc(d2NdEtadPhi, ieta, iphi);
            if (dq) *dq = *dn / *dd; 

// ... Poisson error in multiplicity
            dq = Disps[idisp]->GetLoc(d2NdEtadPhiRMS, ieta, iphi);
            if (dq) *dq = sqrt(nt) / *dd;
          } 

// For MC data only, 

        if (fDataSource != EData)
	{

// ... MC occupancy = MC tracks per pad
          dn = Disps[idisp]->GetLoc(nTracksMC, ieta, iphi);
          if (dp != 0 && dn != 0)
            if (*dp)
            { nt = *dn;
              dq = Disps[idisp]->GetLoc(OccMC, ieta, iphi);
              if (dq) *dq = *dn / *dp; 

// ... Poisson error in occupancy
              de = Disps[idisp]->GetLoc(OccRMSMC, ieta, iphi);
              if (de) *de = sqrt(nt) / *dp;
            } 

// ... MC energy deposited per track
          dn = Disps[idisp]->GetLoc(Edep, ieta, iphi);
          dt = Disps[idisp]->GetLoc(nTracksMC, ieta, iphi);
          if (dt != 0 && dn != 0)
          { nt = *dt;
            if (nt) 
            { dq = Disps[idisp]->GetLoc(EdepPTMC, ieta, iphi);
              if (dq) *dq = *dn / nt; 
// ... Poisson error in energy per track
              de = Disps[idisp]->GetLoc(EdepPTRMSMC, ieta, iphi);
              if (de) *de = *dq / sqrt(nt);
            }
	  }

// ... MC average energy deposited per primary track
          dn = Disps[idisp]->GetLoc(EdepPrimMC, ieta, iphi);
          dt = Disps[idisp]->GetLoc(nTracksPrimMC, ieta, iphi);
          if (dt != 0 && dn != 0)
          { nt = *dt;
            if (nt)
            { dq = Disps[idisp]->GetLoc(EdepPPTMC, ieta, iphi);
              if (dq) *dq = *dn / nt; 
// ... Poisson error in energy per primary track
              de = Disps[idisp]->GetLoc(EdepPPTRMSMC, ieta, iphi);
              if (de) *de = *dq / sqrt(nt);
            }
	  }

// ... MC Total / primary track fraction
          dn = Disps[idisp] -> GetLoc(nTracksMC, ieta, iphi);
          dt = Disps[idisp]->GetLoc(nTracksPrimMC, ieta, iphi);
          if (dt != 0 && dn != 0)
          { nt = *dt;
	    if (nt)
            { dq = Disps[idisp] -> GetLoc (FSecMC,ieta,iphi);
              if (dq) *dq = *dn / nt; 
// ... Binary error in total / primary track fraction
// This should not fail since nTracksMC must always exceed NTracksPrimMC
              de = Disps[idisp] -> GetLoc(FSecRMSMC,ieta,iphi);
              if (de) *de = *dq * sqrt((*dq - 1.) / *dn);
            }
          } 

// ... MC d2N / dEta / dPhi for charged primary tracks
          dn = Disps[idisp] -> GetLoc(nTracksPrimMC,ieta,iphi);
          dt = Disps[idisp] -> GetLoc(dEtadPhi,ieta,iphi);
          if (dt != 0 && dn != 0)
          { nt = *dn;
            if (nt)
            { dq = Disps[idisp] -> GetLoc(d2NdEtadPhiMC,ieta,iphi);
              if (dq) *dq = *dn / *dt; 
// ... Poisson error in energy per primary track
              de = Disps[idisp] -> GetLoc(d2NdEtadPhiRMSMC,ieta,iphi);
              if (de) *de = *dq / sqrt(nt);
            }
	  }

	} // if DataSource 

// Diagnostic printout

        if (ReportLevel >= 5 && iprint < Nprint /*&& idisp == 64*/)
        { printf 
          ("Mul/Recon: %s[%d][%d]: dedp =%10.3e, %6.3f tracks,",
            Disps[idisp]->GetName(), ieta, iphi,
            *Disps[idisp]->GetLoc(dEtadPhi, ieta, iphi),
            *Disps[idisp]->GetLoc(nTracksLim, ieta, iphi));
          printf (" d2NdEdP =%10.3e.\n",
            *Disps[idisp]->GetLoc(d2NdEtadPhi, ieta, iphi));
          iprint++;
        } // ReportLevel
      } // for (iphi
    } // for (ieta
  } // for (idisp

}

////////////////////////////////////////////////////////////////////////////

void TPhMul::Release ()
// Release heap storage
{ Int_t nn; // OSF1 scope problem
  for (nn = 0; nn < kNSensMax; nn++)
    if (Maps[nn]) Maps[nn]->Release();
  for (nn = 0; nn < kNDispMax; nn++)
    if (Disps[nn]) Disps[nn]->Release();
}

////////////////////////////////////////////////////////////////////////////

Int_t TPhMul::SetDiag (Int_t id, Char_t *name, void* ptr)
// Name and store a diagnostic entry in position id if in limits.
// If id is 0, assign the next available id.
// Return id > 0 if successful, else id = 0.
// If the entries are pointers to dynamic storage, it is the 
// responsibility of the user to delete the storage.
{ if (id <= 0 && fNumberDiags < kNDiagMax) id = ++fNumberDiags;
  if (id > 0 && id <= kNDiagMax)
  { fDiagNames[id-1] = name;
    fDiagnostics[id-1] = ptr;
  }
  else id = 0;
  return id;    
}

////////////////////////////////////////////////////////////////////////////

Int_t TPhMul::SetF(Char_t *name, Int_t arrIndex, Float_t value)
{ Int_t &ReportLevel = RLGetF;
  Int_t CaseSensitive = 0;
  Float_t *vp;
  if (!strcmp(name, "feMax") || (!CaseSensitive && !strcmp(name, "femax")))
  { vp = &feMax + sizeof(float) * (arrIndex - 1);
    if (ReportLevel >= 3) 
      printf ("Mul/GetF: %s[%d] changed from %12.4e to %12.4e.\n",
        name, 0, *vp, value);
    *vp = value;
    return 0;
  }
  else
  { printf ("Mul/GetF: variable %s unknown.\n", name);
    return 1;
  }
}

////////////////////////////////////////////////////////////////////////////

void TPhMul::SetOpts(TPhOpts *op) {fOpts = op; }

////////////////////////////////////////////////////////////////////////////

void TPhMul::SetVertex (TPhVertex *v) {fVertex = v; }

////////////////////////////////////////////////////////////////////////////

void TPhMul::ShowDiag (Int_t id)
// Show stored objects on stdout
{ Int_t ilim = (id)? id : fNumberDiags;
  Int_t imin = (id)? id-1 : id;
  printf ("Mul/ShowDiag: %d entries stored.\n", fNumberDiags); 
  for (Int_t ii = imin; ii < ilim; ii++)
//  printf ("id %d: \"%s\" = %s.\n", ii+1, fDiagNames[ii],
//    (TH2F *)fDiagnostics[ii]->GetName()); // 16-Dec-98
  printf ("id %d: \"%s\" = %p.\n", ii+1, fDiagNames[ii],
    fDiagnostics[ii]); // To get going
}

// End of TPhMul implementation
////////////////////////////////////////////////////////////////////////////
