#include "TPhOnString.h"

// ROOT classes

#include "TH1.h"

// Phat classes

#include "TPhDetector.h"
#include "TPhOnString.h"
#include "TPhSensorMapDBFECItem.h"
#include "TPhSensorMapDB.h"

ClassImp(TPhOnString)
  ///////////////////////////////////////////////////////////////////////////
  //
  // This online class contains the set of histograms for a single String 
  // and member functions to operate them. The histograms are noise, pedestal
  // calibration constant, status all per channel vs channel nr.
  // To work with groups of Strings (e.g. monitor all Strings of a FEC) use 
  // the class TPhOnFEC.
  // 
  // Warning: the filling is not yet done, since the necessary information 
  // on data format is missing. the class as it is shows the bookkeeping and 
  // handling of the online histograms.
  ///////////////////////////////////////////////////////////////////////////


TPhOnString::TPhOnString()
{
 // Default constructor. Does nothing use TPhOnString(const Char_t *name) 
 // instead.
  fLabel = 0;  
  fNChannel = 0;   
  fHPedes = NULL;  
  fHNoise = NULL;   
  fHStatus = NULL;  
  fHConst = NULL;   
  fSensorMapDB = NULL;
}

//*********************************************************************
//*********************************************************************

TPhOnString::TPhOnString(Int_t FECNr, Int_t StringNr):TNamed()
{
  // constructor, set name, label etc. to 
  // initialize the histograms use Init()
  if(StringNr>7){
    Warning("TPhOnString","string nr out of range - return");
    return;
  }
  if(FECNr>32){
    Warning("TPhOnString","fec nr out of range - return");
    return;
  }
  Char_t name[256];
  sprintf(name,"FEC%d_string%d",FECNr, StringNr);
  SetName(name);
  fLabel = FECNr* 8 + StringNr;  
  TPhDetector *fdet = TPhDetector::GetInstance();
  if(!fdet) {
    Warning("TPhOnString(name)","detector not connected   - return");
    return;
  }  
  fSensorMapDB = fdet->GetSensorMapDB();
  TPhSensorMapDBFECItem *fecitem = fSensorMapDB->GetFEC(FECNr);
  fNChannel = fecitem->GetStringLength(FECNr);
  fHPedes = NULL;  
  fHNoise = NULL;   
  fHStatus = NULL;  
  fHConst = NULL;   
}

//*********************************************************************
//*********************************************************************

TPhOnString::~TPhOnString()
{
  // destructor; deletes histograms
 if(fHPedes) {
   delete fHPedes;
 }
 if(fHNoise) {
   delete fHNoise;
 }
 if(fHStatus) {
   delete fHStatus;
 }
 if(fHConst) {
   delete fHConst;
 }
};

void TPhOnString::Init()
{
  // Initialize all histograms for this String. Sofar, one 
  // can't choose the histograms to be booked. If it turns 
  // out to be useful, the code can be changed (this means a lot of 
  // bookeeping).

  if(fNChannel == 0) {
    Warning("Init","fNChannel not correctly set - return");
    return;
  }
  Int_t fecnr, stringnr;
  fecnr = (Int_t) fLabel/8;
  stringnr = fLabel - fecnr * 8;
  Char_t hname[256];
  Char_t htitle[256];
  sprintf(hname,"PedesFEC%ds%d",fecnr,stringnr);
  sprintf(htitle,"PedesFEC%ds%d",fecnr,stringnr);
  fHPedes = new TH1S(hname,htitle,fNChannel, 0.,fNChannel-1);
   sprintf(hname,"NoiseFEC%ds%d",fecnr,stringnr);
  sprintf(htitle,"NoiseFEC%ds%d",fecnr,stringnr);
  fHNoise = new TH1S(hname,htitle,fNChannel, 0.,fNChannel-1);
  sprintf(hname,"StatusFEC%ds%d",fecnr,stringnr);
  sprintf(htitle,"StatusFEC%ds%d",fecnr,stringnr);
  fHStatus = new TH1S(hname,htitle,fNChannel, 0.,fNChannel-1);
  sprintf(hname,"ConstFEC%ds%d",fecnr,stringnr);
  sprintf(htitle,"ConstFEC%ds%d",fecnr,stringnr);
  fHConst = new TH1S(hname,htitle,fNChannel, 0.,fNChannel-1);
}

void TPhOnString::Draw(Option_t *option)
{
// draw single histogram; the options are "pedestal","noise", "calib", "status"

  TString opt = option;
  opt.ToLower();
  if (opt.Contains("all")){
    Warning("Draw","only single histogram plotting is implemented use TPhOnStringContainer::Draw(String) instead");
    return;
  }
  // ******** pedestal histo**************
  else if (opt.Contains("pedes")){
    if(!fHPedes) {
      Warning("Draw","pedestal histogram doesn't exist - return");
      return;
    }
    fHPedes->Draw();
  }
  // ******** noise histo**************
  else if (opt.Contains("noise")){
    if(!fHNoise) {
      Warning("Draw","noise histogram doesn't exist - return");
      return;
    }
    fHNoise->Draw();
  }
  // ******** calibration histo**************
  else if (opt.Contains("calib")){
    if(!fHConst) {
      Warning("Draw","calibration const histogram doesn't exist - return");
      return;
    }
    fHConst->Draw();
  }
  // ******** status histo**************
  else if (opt.Contains("status")){
    if(!fHStatus) {
      Warning("Draw","status histogram doesn't exist - return");
      return;
    }
    fHStatus->Draw();
  }
  else {
    Warning("Draw","option doesn't exist");
  }
}

TH1S* TPhOnString::GetHistogram(Option_t *option)
{
// return single histogram; the options are "pedestal","noise", "calib", 
// "status"

  TString opt = option;
  opt.ToLower();
  // ******** pedestal histo**************
  if (opt.Contains("pedes")){
    return fHPedes;
  }
  // ******** noise histo**************
  else if (opt.Contains("noise")){
    return fHNoise;
  }
  // ******** calibration histo**************
  else if (opt.Contains("calib")){
    return fHConst;
  }
  // ******** status histo**************
  else if (opt.Contains("status")){
    return fHStatus;
  }
  return NULL;
}
//-------------------------------------------------------------
Double_t TPhOnString::GetEntries(Option_t *option)
{
// return entries of histogram; the options are "pedestal","noise", "calib", 
// "status"

  TString opt = option;
  opt.ToLower();
  // ******** pedestal histo**************
  if (opt.Contains("pedes")){
    return fHPedes->GetEntries();
  }
  // ******** noise histo**************
  else if (opt.Contains("noise")){
    return fHNoise->GetEntries();
  }
  // ******** calibration histo**************
  else if (opt.Contains("calib")){
    return fHConst->GetEntries();
  }
  // ******** status histo**************
  else if (opt.Contains("status")){
    return fHStatus->GetEntries();
  }
  return 0;
}

void TPhOnString::Reset()
{
  // reset all histograms 
  if(!fHPedes || !fHNoise || !fHConst || !fHStatus) {
    Warning("Reset","At least one of the histograms doesn't exist - return");
    return;}
  fHPedes->Reset();
  fHNoise->Reset();
  fHConst->Reset();
  fHStatus->Reset();
}

Int_t TPhOnString::GetStringNr()
{
  // returns string nr
  Int_t fecnr = (Int_t) fLabel /8;
  Int_t stringnr =  (fLabel-8*fecnr);
  return stringnr;

}

void TPhOnString::Fill()
{
  // does nothing sofar. The variables to be filled into the histograms 
  // should come from the mercury system. The data format is not yet 
  // defined.
}





