#ifndef _TPhPMC_INCLUDED_
#define _TPhPMC_INCLUDED_

//////////////////////////////////////////////////////////////////////////
//                                                                      //
//    TPhPMC                                                            //
//                                                                      //
//    Phobos PMC interface class                                        //
//                                                                      //
//    TPhPMC provides interface to the Phobos Geant detector simulation //
//    code called PMC.                                                  //
//                                                                      //
//                                                                      //
//    Author  : Andrzej Olszewski (andrzej.olszewski@ifj.edu.pl)        //
//    Created : 20/02/98                                                //
//    Version : 1.0                                                     //
//    Changed : 20/06/98 A.O. TPhDetector filled                        //
//              PMC parameters access added                             //
//              Functions SwimFileEvent, SwimMemoryEvent, SwimTrack     //
//              03/12/98 A.O. Run and event parameters and notes        //
//              saved as TPhParKey and TPhNotes objects in the lists    //
//              of TPhParManager and TPhNoteManager classes             //
//              TPhRun object generated.                                //
//                                                                      //
// $Id: TPhPMC.h,v 1.4 1999/03/03 09:09:16 olszewsk Exp $  
//                                                                      //
//////////////////////////////////////////////////////////////////////////

#include "TString.h"

#include "TPhObjectContainer.h"
#include "TPhNotes.h"

class TPhTrack;
class TPhMCTrack;
class TPhRun;
class TPhEvent;
class TPhDetector;

class TPhPMC : public TObject {
    
  public: 
    TPhPMC();
    virtual ~TPhPMC();  
    static  TPhPMC  *GetInstance();
    static  TPhPMC  *CreateInstance();
    static  TPhPMC  *GetOrCreateInstance();

    Bool_t  SetPathToFile(Char_t *filesymbol, Char_t *filename);
    void    AddPMCNotes(Char_t *notes);

    // Most PMC parameters (including GEANT parameters that control physics)
    // can be set interactively.
    Char_t *GetPMCParName(Int_t *no);
    Bool_t  GetPMCParValue(Char_t *pname, Int_t *number, Float_t *values);
    Bool_t  SetPMCParValue(Char_t *pname, Int_t *number, Float_t *values);
    void    PrintPMCPars();

    Bool_t  Init();
    void    AddPhatPMCRInfo(Char_t *parinfo);

    // Set a pointer to run object. It will receive info on run parameters
    void    SetRunLink(TPhRun *run);

    // Get a value of the current magnetic field set in PMC.
    void    PMCGufld(Float_t *vect, Float_t *b);

    // Swim tracks from Hijet ascii input file.
    TPhEvent *SwimPmcEvent(TPhEvent *outEvent,
                     Char_t *outTrackContName, Char_t *outHitContName);

    // Swim tracks stored in the Phat event container.
    TPhEvent *SwimPhatEvent(TPhEvent *inEvent, Char_t *inTrackContName, 
                     TPhEvent *outEvent, Char_t *outTrackContName, Char_t *outHitContName);

    // Swim one track given as a set of vertex-momentum-(geant)id values.
    TPhMCTrack *SwimPhatTrack(Int_t gid, Float_t vx, Float_t vy, Float_t vz, 
                        Float_t px, Float_t py, Float_t pz,
                        TPhEvent *outEvent, Char_t *outTrackContName, Char_t *outHitContName); 
    // Swim one TPhTrack
    TPhMCTrack *SwimPhatTrack(TPhTrack *inTrack,
                TPhEvent *outEvent, Char_t *outTrackContName, Char_t *outHitContName); 

    // Set starting track number when a new event is built 
    // (for use with SwimPhatTrack methods)
    void    SetTrackBaseNo(Int_t no=0);

    Int_t   GetVerbose()                          { return fVerbose;  }
    void    SetVerbose(Int_t i)                   { fVerbose = i;  }

    // Info on versions of code used.
    Float_t GetGeantVer()                         { return fGeantVer;  }
    Float_t GetPmcVer()                           { return fPmcVer;    }
    Float_t GetPmcDetVer()                        { return fPmcDetVer; }
    void    GetVersions();

    // Should be in private but are used by static functions.
    void    SetDetectorCreated(Bool_t wasCreated) { fgDetectorCreated = wasCreated;  }
    Bool_t  GetDetectorCreated()                  { return fgDetectorCreated;        }

  private:

    Bool_t  InitLinks(Char_t *which);
    void    Close();
    void    PutVTrack(Int_t id,
                       Float_t px, Float_t py, Float_t pz,
                       Float_t vx, Float_t vy, Float_t vz);
    void    PutATrack(Int_t id, 
                       Float_t p, Float_t th, Float_t phi, 
                       Float_t vx, Float_t vy, Float_t vz); 
    void    PutTrack(TPhTrack *track);
    Bool_t  Run(TPhEvent *event, Char_t *outTrackContName, Char_t *outHitContName);

    Bool_t  SymRead(Char_t *symname, Char_t *filename, Int_t blength);
    Bool_t  SymChange(Char_t *symname, Char_t *newfilename);


    static Bool_t     fgPMCExists;          //  Singleton flag 
    static Bool_t     fgPMCInit;            //  Initialization flag
    static Bool_t     fgDetectorCreated;    //  Detector initialization flag

    Int_t        fVerbose;          //! Verbose flag
    Float_t      fGeantVer;         // Geant version 
    Float_t      fPmcVer;           // PMC version
    Float_t      fPmcDetVer;        // PMC geometry version
    TString      fFLDASC;           // path to magnetic field map file
    TString      fGCDEFS;           // path to geometry UGCDATA file
    TString      fGENEVS;           // path to HIJET input data file
    TString      fRHAOU;            // path to output ascii AOU file
    TString      fRHDATA;           // path to control file with PMC parameters
    TString      fRHASC;            // path to console output file
    TString      fRHHST;            // path to output histogram file
    TString      fRHNTU;            // path to ntuple output file for tracks with hits
    TString      fMCNTU;            // path to ntuple output file for all tracks
    TString      fRHZEB;            // path to RZ-format Zebra bank output file
    TPhNotes    *fPMCNotes;         // PMC run info stored in a form of a list of strings

  public:

    ClassDef(TPhPMC,1)  // PHOBOS event generator
};

#endif

