      REAL FUNCTION ACCCHK( TP, IDPART, PLANES, NAMES, NHITSL, LSTPLN)
************************************************************************
*                                                                      *
*   FUNCTION ACCCHK  - ACCEPTANCE CHECK                                *
*                                                                      *
*  9-AUG-93  MDB  Change the definition for e+/e- acceptance.          *
* 28-JUN-93  MDB  Initial Version                                      *
*                                                                      *
* A particle is accepted if it satisfies the following:                *
*                                                                      *
* p < 200 MeV         :  6/6  SPA-SPF  or  6/6  SNA-SNF                *
* 200 MeV < p < 1 GeV : 10/10 SPA-SPJ  or 10/10 SNA-SNJ                *
* p > 1 GeV           : 11/11 SPA-SPK  or 11/11 SNA-SNK                *
*                                                                      *
* SxH = SxH1 .OR. SxH2                                                 *
* SxI = SxI1 .OR. SxI2 .OR. SxI3 .OR. SxI4                             *
* SxJ = SxJ1 .OR. SxJ2 .OR. SxJ3 .OR. SxJ4                             *
* SxK = SxK1 .OR. SxK2 .OR. SxK3 .OR. SxK4                             *
*                                                                      *
* Exception: e+/e- must satisfy the following:                         *
*                                                                      *
* p < 200 MeV         :  7/7  SPC-SPI  or  7/7  SNC-SNI                *
* 200 MeV < p < 1 GeV :  8/8  SPC-SPJ  or  8/8  SNC-SNJ                *
* p > 1 GeV           :  9/9  SPC-SPK  or  9/9  SNC-SNK                *
*                                                                      *
* Input:                                                               *
*   TP        = momentum of particle (in GeV).                         *
*   IDPART    = GEANT Particle ID (Integer)                            *
*   PLANES(i) = .TRUE. iff Plane i was hit.                            *
*   NAMES(i)  = Name of plane i.                                       *
*   NHITSL    = # of silicon planes hit.                               *
*   LSTPLN    = Last plane hit.                                        *
*                                                                      *
* Output:                                                              *
*   ACCCHK    =  0.0 if the particle is NOT accepted at all.           *
*             =  1.0 if the particle is accepted in the POSITIVE wing. *
*             = -1.0 if the particle is accepted in the NEGATIVE wing. *
*             =  2.0 if the particle is accepted in the BOTH wings.    *
*                                                                      *
************************************************************************

      IMPLICIT NONE

C     Arguments
      CHARACTER*4 NAMES(*)
      REAL        TP, NHITSL
      LOGICAL     PLANES(*)
      INTEGER     LSTPLN, IDPART

C     Local variables
      INTEGER    NLAYER, NHITS
      PARAMETER (NLAYER=11)
      INTEGER    I, ILAYER
      LOGICAL POSIT(NLAYER), NEGAT(NLAYER)
      LOGICAL POS6,  POS10, POS11, NEG6,  NEG10, NEG11
      LOGICAL POS7E, POS8E, POS9E, NEG7E, NEG8E, NEG9E
      LOGICAL ACCPOS, ACCNEG, HITSOK, NRMMOD
      CHARACTER*1 SIDE, LAYER

C     First make sure that there are enough hits

      ACCCHK = 0.0

      NHITS = NINT( NHITSL )

      ACCPOS = .FALSE.
      ACCNEG = .FALSE.
      HITSOK = .FALSE.
      IF ( IDPART.GE.5 ) THEN
C        Normal Mode
         NRMMOD = .TRUE.
         HITSOK = ( (NHITS.GE.11)
     >         .OR. (NHITS.GE.10 .AND. TP.LE.1.0)
     >         .OR. (NHITS.GE.6  .AND. TP.LE.0.2) )
      ELSEIF ( IDPART.EQ.2 .OR. IDPART.EQ.3 ) THEN
C        Positron/Electron Mode
         NRMMOD = .FALSE.
         HITSOK = ( (NHITS.GE.9)
     >         .OR. (NHITS.GE.8 .AND. TP.LE.1.0)
     >         .OR. (NHITS.GE.7 .AND. TP.LE.0.2) )
      ENDIF

      IF (HITSOK) THEN

         DO 100, I = 1, NLAYER
            POSIT(I) = .FALSE.
            NEGAT(I) = .FALSE.
  100    CONTINUE

         DO 200 I = 1, LSTPLN
            IF (PLANES(I)) THEN
               IF (NAMES(I)(1:1) .EQ. 'S') THEN
                  SIDE   = NAMES(I)(2:2)
                  LAYER  = NAMES(I)(3:3)
                  ILAYER = ICHAR(LAYER)-64
                  IF (1.LE.ILAYER .AND. ILAYER.LE.NLAYER) THEN
                     IF (SIDE.EQ.'P') THEN
                        POSIT(ILAYER)=.TRUE.
                     ELSEIF (SIDE.EQ.'N') THEN
                        NEGAT(ILAYER)=.TRUE.
                     ENDIF
                  ENDIF
               ENDIF
            ENDIF
  200    CONTINUE


         IF (NRMMOD) THEN
            POS6  = ( POSIT(1) .AND. POSIT(2) .AND. POSIT(3) .AND.
     >                POSIT(4) .AND. POSIT(5) .AND. POSIT(6)      )
            POS10 = ( POS6 .AND. POSIT(7) .AND. POSIT(8)
     >                     .AND. POSIT(9) .AND. POSIT(10)      )
            POS11 = ( POS10 .AND. POSIT(11) )

            NEG6  = ( NEGAT(1) .AND. NEGAT(2) .AND. NEGAT(3) .AND.
     >                NEGAT(4) .AND. NEGAT(5) .AND. NEGAT(6)      )
            NEG10 = ( NEG6 .AND. NEGAT(7) .AND. NEGAT(8)
     >                     .AND. NEGAT(9) .AND. NEGAT(10)      )
            NEG11 = ( NEG10 .AND. NEGAT(11) )

            ACCPOS = ( POS11
     >             .OR. (POS10 .AND. TP.LE.1.0)
     >             .OR. (POS6  .AND. TP.LE.0.2) )

            ACCNEG = (      NEG11
     >             .OR. (NEG10 .AND. TP.LE.1.0)
     >             .OR. (NEG6  .AND. TP.LE.0.2) )

         ELSE

            POS7E = ( .NOT.POSIT(1) .AND. .NOT.POSIT(2) .AND. POSIT(3)
     >               .AND. POSIT(4) .AND. POSIT(5) .AND. POSIT(6)
     >               .AND. POSIT(7) .AND. POSIT(8) .AND. POSIT(9)     )
            POS8E = ( POS7E .AND. POSIT(10) )
            POS9E = ( POS8E .AND. POSIT(11) )

            NEG7E = ( .NOT.NEGAT(1) .AND. .NOT.NEGAT(2) .AND. NEGAT(3)
     >               .AND. NEGAT(4) .AND. NEGAT(5) .AND. NEGAT(6)
     >               .AND. NEGAT(7) .AND. NEGAT(8) .AND. NEGAT(9)     )
            NEG8E = ( NEG7E .AND. NEGAT(10) )
            NEG9E = ( NEG8E .AND. NEGAT(11) )

            ACCPOS = (   POS9E
     >             .OR. (POS8E .AND. TP.LE.1.0)
     >             .OR. (POS7E .AND. TP.LE.0.2) )

            ACCNEG = (   NEG9E
     >             .OR. (NEG8E .AND. TP.LE.1.0)
     >             .OR. (NEG7E .AND. TP.LE.0.2) )

         ENDIF

         IF (ACCPOS) THEN
            ACCCHK = 1.0
            IF (ACCNEG) THEN
               ACCCHK = 2.0
            ENDIF
         ELSEIF (ACCNEG) THEN
            ACCCHK = -1.0
         ELSE
            ACCCHK = 0.0
         ENDIF

      ENDIF

      RETURN
      END
