      SUBROUTINE CYFLD(R, V, BR, BV)
*Documentation_Block
************************************************************************
*
* Description:
*   Calculate the magnetic field from the two dimensional
*   cylindrical field map.
*   The map is stored in common block defined in CYFLDC.CMN
*   and read by GUFLD prior tothe first call to CYFLD
*
*   It is possible to use the magnetic field map calculated for the magnet
*   with smaller radius for other with the radius increased
*   by REXT (local parameter). In this case the field values for R < REXT are
*   identical with those for R=0 for the same value  V (in the original map)
*   This option should however be used only temporarily -
*   the magnetic field created this way is only aproximate
*
* Arguments:
*  *R   - radius of the requested point (cylindrical coordinate 1)
*  *V   - vertical position of the point (cylindrical coordinate 2)
*  BR*  - radial component of the field
*  BV*  - vertical component of the field
*
* Called by:
*  GUFLD
*
* Calls:
*  FINT
*
* Implicit inputs, outputs, side effects:
*  NONE
*
* Filename: CYFLD.FOR
*
* History:
* Date      Name       Description
* ----      ----       -----------
* 2-Dec-93  K.W.        Creation
************************************************************************
*End_Documentation_Block



      IMPLICIT NONE

*KEEP,CYFLDC.
*Include CYBFLD.BLK
*Start_Documentation
***********************************************************************
*                                                                     *
*                     INCLUDE file 'CYBFLD.INC'                       *
*Start_Verbatim                                                       *
*                                                                     *
*  Created:  02-MAR-1993   K. Wozniak                                 *
*                                                                     *
*                 Cylindrical magnetic field map                      *
*                 ------------------------------                      *
*                                                                     *
*  NNRMAX   Parameter - maximal size of the map in radial direction   *
*  NNVMAX   Parameter - maximal size of the map in vertical direction *
*  NNR      actual size of the map in radial direction                *
*  NNV      actual size of the map in vertical direction              *
*  XTAB     radial (XTAB(1)..XTYAB(NNR)  and vertical (XTAB(NNR+1)..  *
*            XTAB(NNR+NNV))  coordiantes of the grid points of the map*
*  BRTAB    table of the radial compoments of the field               *
*  BVTAB    table of the vertical components of the field             *
*           Field value at the point  (XTAB(I), XTAB(NNR+K) is:       *
*             (BRTAB(I+(K-1)*NNR), BVTAB(I+(K-1)*NNR))                *
*                                                                     *
*                                                                     *
*End_Verbatim                                                         *
***********************************************************************
*End_Documentation

      INTEGER NNRMAX, NNR, NNVMAX, NNV

      PARAMETER (NNRMAX = 200, NNVMAX = 150)

      REAL XTAB, BRTAB, BVTAB
      COMMON /CYBFLD/ NNR, NNV, XTAB(NNRMAX+NNVMAX),
     +                BRTAB(NNRMAX*NNVMAX), BVTAB(NNRMAX*NNVMAX)

*KEND.

      REAL R, V, BR, BV,  X(2), P, REXT
      PARAMETER (REXT=0.0)

      INTEGER NN(2)
      REAL FINT
      EXTERNAL FINT

      EQUIVALENCE (NN(1), NNR)

      X(1) = R - REXT
      IF(X(1) .LE. 0.0)  X(1) = 0.0

      X(2) = ABS(V)

      IF(R .LT. 0.0)  THEN
         WRITE(*,*) ' GUFLD: NEGATIVE  R  VALUE'
         GO TO 995
      ENDIF

      IF(X(1) .GT. 2.0*XTAB(NNR))   GO TO 995
      IF(X(2) .GT. 2.0*XTAB(NNR+NNV))   GO TO 995

      IF(X(1) .GT. XTAB(NNR))   X(1) = XTAB(NNR)
      IF(X(2) .GT. XTAB(NNR+NNV))   X(2) = XTAB(NNR+NNV)

      BR = FINT(2, X, NN, XTAB, BRTAB)
      BV = FINT(2, X, NN, XTAB, BVTAB)

      P = 1.0
      IF(R-REXT .GT. XTAB(NNR))  P = (2.0*XTAB(NNR)-R+REXT) / XTAB(NNR)
      IF(ABS(V) .GT. XTAB(NNR+NNV))   P = P *
     +                    (2.0*XTAB(NNR+NNV) - ABS(V)) / XTAB(NNR+NNV)
      BR = BR * P
      BV = BV * P

      IF(V .LT. 0.0)  BR = - BR

      GO TO 999

  995 CONTINUE
      BR = 0.0
      BV = 0.0

  999 CONTINUE
      RETURN
      END
