      SUBROUTINE GUFLD(VECT,B)
*DocumentationBlock
***********************************************************************
*                                                                     *
* Description:                                                        *
*  The Phobos-specific version of GUFLD which returns the magnetic    *
*  field B at any point VECT.  It has 3 modes, controlled by variables*
*  BFLD, RFLD, MAGT, MAGX, MAGZ and MOFF, in common RHSYS, which      *
*  can be specified on input cards:                                   *
*                                                                     *
*  BFLD is the field strength in kGauss.                              *
*  If you do not want to redefine the magnetic field values (strongly *
*  recomended for MAGT=3) BFLD=0.0 should be used.                    *
*  if it is different than 0.0 all field values will be multiplied by *
*  a factor equal:                                                    *
*     BFLD / Bcenter                                                  *
*  where Bcenter is the field strength at (MAGX, 0, MAGZ)             *
*                                                                     *
*  The position of the map can be changed by the MOFF parameter       *
*  the point VECT is THEN offset and  a point defined as:             *
*       VECTOFF = VECT - MOFF  (vector)                               *
*  is used as a position in the map.                                  *
*  THIS IS TRUE for standard PHOBOS map, however it is possible       *
*  for MAGT=2 to define a sub-map positioned absolutely               *
*                                                                     *
*  MAGT = 0:  no magnetic field at all                                *
*                                                                     *
*  MAGT = 1: Diagnostic mode                                          *
*   The field is a simple dipole field: Bx = Bz = 0,                  *
*   By = BFLD if r < RFLD, By = 0 otherwise,                          *
*   where rx = x - MAGX and rz = z - MAGZ.                            *
*                                                                     *
*  MAGT = 2: ASCII 3D-field map mode                                  *
*   At the first call, a map of the magnetic field components is read *
*   in.  First the values of the coordinates at grid points are read  *
*   Then there are the lines with the indexes coresponding to the     *
*   position of the point and the field vector (Bx,By,Bz).            *
*   The field values not exactly at the known point are linearly      *
*   interpolated using FINT for each of components Bx, By, Bz.  The   *
*   coordinates are transformed from the GEANT mother system to the   *
*   fieldmap system using information in COMMON.SYS.                  *
*                                                                     *
*  MAGT=3 The magnetic field is a superposition of two fields from    *
*   magnets. The field from each magnet is calculated in the case of  *
*   cylindrical symetry. Thus the magnetic field map is two dimensio- *
*   nal only. The map is read from an ASCII file with a similar format*
*   as for MAGT=2 (the difference is in the definition of points:     *
*   (x,y,z) for MAGT=2 and (r,v) for MAGT=3                           *
*   The field values from two-dimensional map are interpolated if     *
*   the point lies inside the boundaries. In the case of points       *
*   outside the boundaries the value from the closest point exactly   *
*   at the boundary is taken and multiplied by the factor lineary     *
*   decreasing with the distance from the boundary. It protects from  *
*   dangers of extrapolation of the field and avoids the sharp change *
*   of the field before it is set to zero                             *
*
*  Arguments:
*  *VECT: x = VECT(1)... = point at which B is to be evaluated   (input)
*  B*     B(1)...B(3) = field at point VECT                     (output)
*
*  Called by:
*  Geant tracking subroutines GHELIX or GRKUTA, called by GUSWIM
*
*  Calls:
*  Cern routine FINT to do linear interpolation
*  STOPER - in case of error in the map
*  UFILE  or  SERVER to open map file
*  ADDPMCRINFO
*
*  Implicit inputs, outputs, side effects:
*  The field maps are read on unit Iusrmap, assumed to have been
*    opened in UGINIT; later, open will be done in GUFLD.
*  Initialization on local switch
*
*  History:
*  Date       Name        Description
*  ----       ----        -----------
*  23-Nov-98  K.W.        Add calls to ADDPMCRINFO
*   7-Jul-98  K.W.        Automatic recognition of magnetic field type
*                         (MAGT) from the encountered keywords in the
*                         FLDASC file
*  18-Feb-98  K.W.        Call STOPER if the magnetic field map file FLDASC
*                         can not be opened
*  11-Feb-98  K.W.        FIRSTC moved to a common to enable restart of
*                         simulations in PHAT
*  25-Aug-97  K.W.        Opening of FLDASC file moved from UGINIT to GUFLD
*                         Multiple map files enabled (in MAGT=2 mode)
*   2-Jul-97  K.W.        Unlimited number of sub-maps in MAGT=2 mode
*                         absolute and movable (by MOFF) sub-maps possible
*                         (PHOIND parameter)
*  20-Sep-96  K.W.        Reading of the map file:
*                          MAGV, MAGT - always modyfied
*                          MAGX, MAGZ, RFLD - changed only if it was 0.0
*                         Removed unused FANGL
*  21-Jul-96  R. V.       Removed unused zc, j, yneg
*  12-Jul-96  K.Wozniak   magnetic field center (MAGX,MAGZ) separated from
*                         the field offset (MOFF(3))
*  06-Jul-96  K.Wozniak   Removed ZEBRA map, added output of magnet
*                         information from the map to Ascii OUput file
*                         general reorganization of the code
*  03-Jul-96  A. Trzupek  Prepared map for the CDR magnet (MAGT=2)
*                         which comprises two maps precise one for
*                         spectrometer and less precise but larger
*                         in x and z for TOF. Changed gufld to allow for
*                         reading them.
*  10-Jul-95  K. Wozniak  Redefinition of the MAGT=2 mode
*  02-Mar-93  K. Wozniak  Added facility for cylindrical field map
*  09-Dec-92  C. Conner   Added facility for pre-sorted Zebra field map
*  17-Jun-92  C. Parsons  Extracted from GPH with unspecified changes
*  22-Mar-92  Don Mcleod  Initial version with Ascii field map, in GPH
************************************************************************
*EndDocumentationBlock

***********************************************************************

      IMPLICIT NONE

*KEEP,FLAG.
C  PHOBOS flags, run control, operating modes, misc.   File FLAG.CMN
C   7-Jul-98  K. W.      Added  ITRIG flag (initial numer in TRIG command)
C  27-Nov-97  K. W.      Added IAOUF flag
C  31-Mar-95  R.V.       Change STRA to SECT to avoid GEANT 3.21 conflict
C  22-Nov-92  D. McLeod
C   These are set in input cards
      INTEGER ISASC,ISHBO,ISHDO,ISNTU, ISMEM,ISAOU,IAOUF, ISECT,IDRAW,
     1        IGEN,IANAL,ISCAT,IPIXL,ISEQ,NTRK,IZEB, ITRIG
      REAL WPIXL,WLOSS
      COMMON/RHFLG/ISASC,ISHBO,ISHDO,ISNTU,
     1  ISMEM,ISAOU,IAOUF, ISECT,IDRAW,
     2  IGEN,IANAL,ISCAT,IPIXL,ISEQ,NTRK,IZEB,WPIXL,WLOSS, ITRIG
*KEEP,CYFLDC.
*Include CYBFLD.BLK
*Start_Documentation
***********************************************************************
*                                                                     *
*                     INCLUDE file 'CYBFLD.INC'                       *
*Start_Verbatim                                                       *
*                                                                     *
*  Created:  02-MAR-1993   K. Wozniak                                 *
*                                                                     *
*                 Cylindrical magnetic field map                      *
*                 ------------------------------                      *
*                                                                     *
*  NNRMAX   Parameter - maximal size of the map in radial direction   *
*  NNVMAX   Parameter - maximal size of the map in vertical direction *
*  NNR      actual size of the map in radial direction                *
*  NNV      actual size of the map in vertical direction              *
*  XTAB     radial (XTAB(1)..XTYAB(NNR)  and vertical (XTAB(NNR+1)..  *
*            XTAB(NNR+NNV))  coordiantes of the grid points of the map*
*  BRTAB    table of the radial compoments of the field               *
*  BVTAB    table of the vertical components of the field             *
*           Field value at the point  (XTAB(I), XTAB(NNR+K) is:       *
*             (BRTAB(I+(K-1)*NNR), BVTAB(I+(K-1)*NNR))                *
*                                                                     *
*                                                                     *
*End_Verbatim                                                         *
***********************************************************************
*End_Documentation

      INTEGER NNRMAX, NNR, NNVMAX, NNV

      PARAMETER (NNRMAX = 200, NNVMAX = 150)

      REAL XTAB, BRTAB, BVTAB
      COMMON /CYBFLD/ NNR, NNV, XTAB(NNRMAX+NNVMAX),
     +                BRTAB(NNRMAX*NNVMAX), BVTAB(NNRMAX*NNVMAX)

*KEEP,RHSYS.
C  Common blocks for GPH PHOBOS simulation:  FILE RHSYS.CMN
C  General geometric layout of PHOBOS detector exclusive of detectors
C  These values are on input cards or data statements, except *MAX are
C  calculated to encompass all elements of the system.
C  20-Sep-96  K.W       Removed obsolete BMTBRAD, BERTHIK
C                       Removed FANGL - now local variable in BANG
C  12-Jul-96  K.W       Position of the magnet pole center MAGX, MAGZ
C                       (previously XOFF,ZOFF) separated from the magnet
C                       offset (now MOFF - 3-D vector)
C  14 Dec-92  R.V.      TMAG redefined and changed to integer
C  22-Nov-92 D. McLeod  Renamed from COMMON.SYS; TMAG, ZOFF, XOFF
C                       are added from the original GPH version
      REAL   BFLD         ! Magnetic field strength
     &      ,MAGX,MAGZ    ! Coordinates of field center
     &      ,MOFF(3)      ! offset of the field
     &      ,RFLD         ! Radius of field (perfect dipole only)
     1   ,XMAX,YMAX,ZMAX,ZMIN,XMMAX,ZMMAX
      INTEGER MAGT, TMAG  ! Field type: 0 Zebra bank, 1 = dipole,
                          ! 2 = ASCII map
      COMMON/RHSYS/ BFLD,MAGT,MAGX,MAGZ,MOFF
     1  ,RFLD,XMAX,YMAX,ZMAX,ZMIN,XMMAX,ZMMAX
      EQUIVALENCE (TMAG, MAGT)
*KEEP,IO.
C Common block for GPH logical units etc.: file IO.CMN
c 13-Aug-96  R. V.       Added USRBATCHK for KUIP commands for UGCONS
c 21-Jul-96  R. V.       Remove unused record; note CARDDEF now obsolete
c 22-Feb-93  R. V.       Added USRMCNTU and IMCNTU for MC Ntuple
c 14-Nov-92  R. Verdier  Added UGCIN, IUGCIN, CARDDEF and ICARDDEF
c                        for UGC and EUCARD (obsolete) input units
C    Apr-92  C. Parsons  Replacement for original COMMON.CDT
      INTEGER USRGCIN, CARDDEF, USRIN, USROUT, USRMAP, USRNTU, USRZEB,
     *  USRMCNTU, USRBATKUMAC
      PARAMETER (USRIN = 4, USROUT = 16, USRMAP = 18)
      PARAMETER (USRNTU = 15, USRZEB = 17, USRGCIN = 19, CARDDEF = 20)
      PARAMETER (USRMCNTU = 21, USRBATKUMAC = 22)
      INTEGER ICARDDEF, IMCNTU, IUSRGCIN, IUSRIN, IUSROUT, IUSRMAP,
     *  IUSRNTU, IUSRZEB, IUSRAOU, IUSRBATKUMAC
      COMMON /IO/ IUSRIN, IUSROUT, IUSRMAP, IUSRNTU, IUSRZEB, IUSRGCIN,
     *  ICARDDEF, IMCNTU, IUSRAOU, IUSRBATKUMAC
*KEEP,FIRSTC.
C  PHOBOS "first call" flags,   File FIRSTC.CMN
C  The flags denoting the very first call are put in one common
C  to enable easy restart of simulations in PHAT
C  11-Feb-98  K. W.      Creation
C
      LOGICAL FCGUFLD, FCGUSTEP, FCPHOUT
      COMMON /CFIRSTC/ FCGUFLD, FCGUSTEP, FCPHOUT
*KEND.

C     Input variables

      REAL VECT(*)

C     Output variables

      REAL B(3)

C     Local variables

      REAL X(3)
      REAL VECTOFF(3)
      INTEGER NX,NY,NZ

C     External Functions
C     ------------------

      REAL FINT
      EXTERNAL FINT

C   ---- ASCII MAP DECLARATIONS --------------

      INTEGER I,K, IX,IY,IZ, IMAP, IFILE
      INTEGER NMAPMAX, MAXMAP, MAXGRID
******  maximum value allowed for NX*NY*NZ
      PARAMETER (MAXMAP=520000)
******  maximum value allowed for NX + NY + NZ
      PARAMETER (MAXGRID=400)
******  maximum value allowed for the number of maps
      PARAMETER (NMAPMAX=20)
      REAL A
      REAL BX,BY,BZ
      REAL BMAX2
      INTEGER NMAP, IPHOIND, IMPOS, IGPOS, NA
      COMMON /FLDMAP/ NMAP, IMPOS(NMAPMAX), IGPOS(NMAPMAX),
     1                IPHOIND(NMAPMAX), BMAX2(NMAPMAX),
     2                BX(MAXMAP),BY(MAXMAP),BZ(MAXMAP),
     3                A(MAXGRID), NA(3,NMAPMAX)

      CHARACTER*80 CBUF
      CHARACTER*70 MAPFILE, FILENAME

C  -------- End of ASCII map declarations -------------


      REAL B1(3), B2(3), BR, BV, R1, R2, SINA, COSA, BMAX3
      INTEGER IR, IV
      INTEGER NLINES
      REAL MAGV, MAGNEW
      CHARACTER*120 LINE

*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C

      IF(.NOT. FCGUFLD)  GO TO 200

        FCGUFLD = .FALSE.

      IF(MAGT .EQ. 0 .OR. MAGT .EQ. 1)  THEN
        WRITE(LINE, '(A)')
     +   ' M C Magnetic field set in the control file '
        IF(ISAOU .GT. 0)  WRITE(IUSRAOU, '(A)' )   LINE(1:80)
        IF(ISMEM .GT. 0)  CALL ADDPMCRINFO(LINE)

        IF(MAGT .EQ. 0)  THEN
          WRITE(*,*)  ' GUFLD - NO MAGNETIC FIELD'
          IF(BFLD .NE. 0.0)  WRITE(*,*) ' GUFLD - BFLD = ', BFLD,
     +          '   SET IN THE CONTROL FILE WAS IGNORED'
          WRITE(*,*)  ' GUFLD - BFLD = 0.0'
        ELSE IF(MAGT .EQ. 1)  THEN
          WRITE(*,*) ' GUFLD - MAGNETIC FIELD CONSTANT IN TWO CYLINDERS'
          WRITE(*,*) ' GUFLD - BFLD = ', BFLD
          IF(ABS(BFLD) .LT. 1.0E-10)
     +      WRITE(*,*)  ' GUFLD - USE  MAGT = 0  TO OBTAIN ZERO FIELD!'
        ENDIF
        GO TO 199
      ENDIF

*******************************************************
*   Read field map file in the first call             *
*******************************************************


      IF (MAGT .GE. 2)    THEN  !  read parameters of the map
c                               !  from the magnetic field map file


c Open Ascii file
        MAPFILE = 'FLDASC'
        CALL UFILE (0, MAPFILE, IUSRMAP, FILENAME)
        IF(IUSRMAP .LT. 1)   CALL STOPER('GUFLD',
     +     'FLDASC  CAN NOT BE OPENED', 1)

        IFILE = 1
        WRITE(LINE, '(A)')
     +    ' M C Magnetic field defined by the magnetic field map file'
        IF(ISAOU .GT. 0)  WRITE(IUSRAOU, '(A)' )  LINE(1:80)
        IF(ISMEM .GT. 0)  CALL ADDPMCRINFO(LINE)

        WRITE(LINE, '(A)')  ' M C Map file name (file 1):'
        IF(ISAOU .GT. 0)  WRITE(IUSRAOU, '(A)' )  LINE(1:80)
        IF(ISMEM .GT. 0)  CALL ADDPMCRINFO(LINE)

        WRITE(LINE, '('' M C '', A)')  FILENAME
        IF(ISAOU .GT. 0)  WRITE(IUSRAOU, '(A)' )  LINE(1:80)
        IF(ISMEM .GT. 0)  CALL ADDPMCRINFO(LINE)

        NX = 0   !  initialization for any kind of map
        NY = 0   !  zero denotes undefined field
        NZ = 0
        NNR = 0
        NNV = 0
        NLINES = 0
        NMAP = 0

   10  READ(IUSRMAP, '(A)', END=941)  CBUF
       NLINES = NLINES + 1
        IF(CBUF(1:1) .EQ. '*')  GO TO 10
        IF(CBUF(1:4) .EQ. 'MAGV')   THEN  !  magnetic field version number
           READ(CBUF(6:80), *, ERR=942)  MAGV
           GO TO 10
        ENDIF
        IF(CBUF(1:4) .EQ. 'MAGT')   THEN  !  redefinition of the field type
           READ(CBUF(6:80), *, ERR=942)  MAGT
           IF(MAGT .LT. 0  .OR.  MAGT .GT. 3)  GO TO 946  !  illegal value
           WRITE(*,*) ' GUFLD - MAP TYPE FOUND IN THE FILE = ', MAGT
           GO TO 10
        ENDIF
        IF(CBUF(1:4) .EQ. 'MAGX')   THEN  !  redefinition of magnet center
           READ(CBUF(6:80), *, ERR=942)  MAGNEW
           IF(MAGX .EQ. 0.0)  THEN
              MAGX = MAGNEW   !  MAGX changed only if it was zero
              WRITE(*,*) ' GUFLD - MAGX = ', MAGX
           ELSE
              WRITE(*,*) ' GUFLD - IGNORED MAGX FROM THE FILE:', MAGNEW
              WRITE(*,*) ' GUFLD - USED MAGX: ', MAGX
           ENDIF
           GO TO 10
        ENDIF
        IF(CBUF(1:4) .EQ. 'MAGZ')   THEN  !  redefinition of magnet center
           READ(CBUF(6:80), *, ERR=942)  MAGNEW
           IF(MAGZ .EQ. 0.0)  THEN
              MAGZ = MAGNEW   !  MAGZ changed only if it was zero
              WRITE(*,*) ' GUFLD - MAGZ = ', MAGZ
           ELSE
              WRITE(*,*) ' GUFLD - IGNORED MAGZ FROM THE FILE:', MAGNEW
              WRITE(*,*) ' GUFLD - USED MAGZ: ', MAGZ
           ENDIF
           GO TO 10
        ENDIF
        IF(CBUF(1:4) .EQ. 'RFLD')   THEN  !  redefinition of magnet radius
           READ(CBUF(6:80), *, ERR=942)  MAGNEW
           IF(RFLD .EQ. 0.0)  THEN
              RFLD = MAGNEW   !  RFLD changed only if it was zero
              WRITE(*,*) ' GUFLD - RFLD = ', RFLD
           ELSE
              WRITE(*,*) ' GUFLD - IGNORED RFLD FROM THE FILE:', MAGNEW
              WRITE(*,*) ' GUFLD - USED RFLD: ', RFLD
           ENDIF
           GO TO 10
        ENDIF
        IF(CBUF(1:2) .EQ. 'M ')   THEN  !  information to be copied to AOU file
           WRITE(LINE,'('' M C'',A79)' ) CBUF(2:80)
           IF(ISAOU .GT. 0)  WRITE(IUSRAOU, '(A)' )  LINE(1:83)
           IF(ISMEM .GT. 0)  CALL ADDPMCRINFO(LINE)
           GO TO 10
        ENDIF

****************************************************************
*   End of general information - check the type of the field   *
****************************************************************

        IF(MAGT .EQ. 0)  GO TO 199
        IF(MAGT .EQ. 1)  GO TO 199

        IF(CBUF(1:6) .EQ. 'XAXIS:'  .OR.
     +     CBUF(1:6) .EQ. 'YAXIS:'  .OR.
     +     CBUF(1:6) .EQ. 'ZAXIS:')  THEN !  map type 2
            MAGT = 2
            GO TO 20
        ENDIF

        IF(CBUF(1:7) .EQ. 'RADIAL:'  .OR.
     +     CBUF(1:9) .EQ. 'VERTICAL:')  THEN !  map type 3
            MAGT = 3
            GO TO 111
        ENDIF

        IF(MAGT .EQ. 2)  GO TO 20
        IF(MAGT .EQ. 3)  GO TO 111
       GO TO 945

      ELSE

         GO TO 945   !  wrong type of the magnetic field

      ENDIF   !   MAGT .EQ. 2


*******************************************************
*  ASCII map method: field values given in a box      *
*******************************************************


*********************************************************************
*  Read the three-dimensional ASCII map (in a box)                  *
*  Format:                                                          *
*   - comments (starting with *) - description of the file contents *
*   - 'XAXIS:'  nx      - definition of the number of grids in      *
*                         x - direction                             *
*   - nx     values of X from 0.0 to maximum (sorted) - free format *
*   - comments (if any)                                             *
*   - 'YAXIS:'  ny      - definition of the number of grids in      *
*                         y - direction                             *
*   - ny     values of Y from 0.0 to maximum (sorted) - free format *
*   - comments (if any)                                             *
*   - 'ZAXIS:'  nz      - definition of the number of grids in      *
*                         z - direction                             *
*   - nz     values of Z from 0.0 to maximum (sorted) - free format *
*   - comments (if any)                                             *
*   - 'PHOMAG:' iphoind    - PHOBOS magnet flag:                    *
*      iphoind = 0   - map defined only in the area defined by AXIS *
*                      commands                                     *
*                      MOFF not valid for THIS PART OF the field    *
*      ipohind = 1   - definition for x>0 and y>0  extends to  all  *
*                      signs of x, y                                *
*                      field modified by MOFF                       *
*   - 'FIELD:'  bmax  - declaration of the begining of field values *
*                       bmax - maximal field for y=0                *
*   - field values in the nodes of the grid, one node per line:     *
*     ix  iy  iz   Bx  By  Bz      where:                           *
*     ix, iy, iz   - definition of the node point: (ix),Y(iy),Z(iz) *
*     Bx, By, Bz   - components of the field                        *
*  IT IS ASSUMED, THAT THE FIELD IS DEFFINED FOR:                   *
*        X >= 0,  Y >= 0,   Z in desired range                      *
*********************************************************************

   20 CONTINUE
       NMAP = NMAP + 1
       IF(NMAP .GT. NMAPMAX)  GO TO 957
       IMPOS(NMAP) = 0
       IGPOS(NMAP) = 0
       IF(NMAP .GT. 1)   THEN
          IMPOS(NMAP) = IMPOS(NMAP-1) + NX*NY*NZ
          IGPOS(NMAP) = IGPOS(NMAP-1) + NX + NY + NZ
       ENDIF
       NX = 0
       NY = 0
       NZ = 0
       IPHOIND(NMAP) = 1  !  standard - field from PHOBOS magnet
       GO TO 22

   21 CONTINUE
       READ(IUSRMAP, '(A)', END=941)  CBUF
       NLINES = NLINES + 1

   22  CONTINUE
       IF(CBUF(1:1) .EQ. '*')  GO TO 21
       IF(CBUF(1:2) .EQ. 'M ')   THEN  !  information to be copied to AOU file
           WRITE(LINE,'('' M C'',A79)' ) CBUF(2:80)
           IF(ISAOU .GT. 0)  WRITE(IUSRAOU, '(A)' )  LINE(1:86)
           IF(ISMEM .GT. 0)  CALL ADDPMCRINFO(LINE)
          GO TO 21
       ENDIF
       IF(CBUF(1:6) .EQ. 'XAXIS:')  THEN
          READ(CBUF(7:80), *, ERR=942)  NX
          IF(NX .LT. 2 )    GO TO 956
          IF(NX+NY+NZ+IGPOS(NMAP) .GT. MAXGRID)    GO TO 954
          READ(IUSRMAP, *, END=941, ERR=942)
     +                               (A(IGPOS(NMAP)+I), I=1,NX)
          NLINES = NLINES + 1
          GO TO 21
       ENDIF
       IF(CBUF(1:6) .EQ. 'YAXIS:')  THEN
          IF(NX .LT. 1)  GO TO 951
          READ(CBUF(7:80), *, ERR=942)  NY
          IF(NY .LT. 2)    GO TO 956
          IF(NX+NY+NZ+IGPOS(NMAP) .GT. MAXGRID)    GO TO 954
          READ(IUSRMAP, *, END=941, ERR=942)
     +                            (A(IGPOS(NMAP)+NX+I), I=1,NY)
          NLINES = NLINES + 1
          GO TO 21
       ENDIF
       IF(CBUF(1:6) .EQ. 'ZAXIS:')  THEN
          IF(NX .LT. 1)  GO TO 951
          IF(NY .LT. 1)  GO TO 952
          READ(CBUF(7:80), *, ERR=942)  NZ
          IF(NZ .LT. 2)    GO TO 956
          IF(NX+NY+NZ+IGPOS(NMAP) .GT. MAXGRID)    GO TO 954
          READ(IUSRMAP, *, END=941, ERR=942)
     +                          (A(IGPOS(NMAP)+NX+NY+I),I=1,NZ)
          NLINES = NLINES + 1
          GO TO 21
       ENDIF
       IF(CBUF(1:7) .EQ. 'PHOMAG:')  THEN
          READ(CBUF(8:80), *, ERR=942)  IPHOIND(NMAP)
          IF(IPHOIND(NMAP) .NE. 0  .AND.
     +        IPHOIND(NMAP) .NE. 1)             GO TO 958
          GO TO 21
       ENDIF
       IF(CBUF(1:6) .EQ. 'FIELD:')  THEN
         IF(NX .LT. 1 .OR. NY .LT. 1 .OR. NZ .LT. 1)  GO TO 953
         IF(NX*NY*NZ+IMPOS(NMAP) .GT. MAXMAP)  GO TO 955
         READ(CBUF(7:80), *, ERR=942)  BMAX2(NMAP)  !  mominal maximum at y=0
         WRITE(*,*) ' GUFLD - MAXIMUM ABS(BY) IN Y=0 PLANE:',
     +                                             BMAX2(NMAP)
         IF(BFLD .NE. 0.0 .AND.  IPHOIND(NMAP) .NE. 0)  THEN
            WRITE(*,*) ' GUFLD - RESCALED VALUE = ', BFLD
         ELSE
            WRITE(*,*) ' GUFLD - NO RESCALING'
         ENDIF
         CALL VZERO(BX(IMPOS(NMAP)+1), NX*NY*NZ)
         CALL VZERO(BY(IMPOS(NMAP)+1), NX*NY*NZ)
         CALL VZERO(BZ(IMPOS(NMAP)+1), NX*NY*NZ)
         NA(1,NMAP) = NX
         NA(2,NMAP) = NY
         NA(3,NMAP) = NZ
   25    CONTINUE
          READ(IUSRMAP, '(A)', END=35, ERR=942)  CBUF
          NLINES = NLINES + 1
          READ(CBUF(1:80), *, ERR=31) IX,IY,IZ, B1(1),B1(2),B1(3)
          IF(IX .LT. 1  .OR.  IY .LT. 1  .OR.  IZ .LT. 1  .OR.
     +       IX. GT. NX  .OR.  IY. GT. NY  .OR.  IZ. GT. NZ) GO TO 959
          K = IMPOS(NMAP)  +  IX + NX * (IY-1 + NY*(IZ-1))
          BX(K) = B1(1)
          BY(K) = B1(2)
          BZ(K) = B1(3)
          IF(IX.EQ.NX.AND.IY.EQ.NY.AND.IZ.EQ.NZ) GO TO 30
          GO TO 25
       ENDIF

       GO TO 942  !   ERROR IN THE MAP

   30  CONTINUE
       READ(IUSRMAP, '(A)', END=35)  CBUF
       NLINES = NLINES + 1
   31  CONTINUE
       GO TO 20

   35 CONTINUE
        CALL UFILE (2, MAPFILE, IUSRMAP, FILENAME)
      IF(IUSRMAP .GT. 0)  THEN   !  next map file opened, continue reading
         NLINES = 0
         IFILE = IFILE + 1

         WRITE(LINE, '(A, I2, A)')
     +           ' M C Map file name (file ', IFILE, '):'
         IF(ISAOU .GT. 0)  WRITE(IUSRAOU, '(A)' )  LINE(1:80)
         IF(ISMEM .GT. 0)  CALL ADDPMCRINFO(LINE)

         WRITE(LINE, '('' M C '', A)')  FILENAME
         IF(ISAOU .GT. 0)  WRITE(IUSRAOU, '(A)' )  LINE(1:80)
         IF(ISMEM .GT. 0)  CALL ADDPMCRINFO(LINE)

         GO TO 30
      ENDIF

  99  CONTINUE
      GO TO 199


*********************************************************************
*  Magnetic field is a superposition of two cylinrically symetrical *
*  fields from two magnets                                          *
*  2-dimensional ASCII map (IN cylindrical coordinates).            *
*  Format:                                                          *
*   - comments (starting with *) - description of the file contents *
*   - 'RADIAL:'  nnr   - definition of the number of grids in       *
*                        radial direction                           *
*   - nnr values of R from 0.0 to maximum (sorted) - free format    *
*   - comments (if any)                                             *
*   - 'VERTICAL:' nnv  - definition of the number of grids in       *
*                        vertical direction                         *
*   - nvv values of V from 0.0 to maximum (sorted) - free format    *
*   - comments (if any)                                             *
*   - 'FIELD:'      - declaration of the begining of field values   *
*   - field values in the nodes of the grid, one node per line:     *
*     ir iv  Br  Bv          where:                                 *
*     ir, iv - definition of the node point:  R(ir), V(iv)          *
*     Br  - radial component of the field                           *
*     Bv  - vertical component of the field                         *
*********************************************************************

  110  READ(IUSRMAP, '(A)', END=941)  CBUF
       NLINES = NLINES + 1
  111  IF(CBUF(1:1) .EQ. '*')  GO TO 110
       IF(CBUF(1:7) .EQ. 'RADIAL:')  THEN
          READ(CBUF(8:80), *, ERR=942)  NNR
          IF(NNR .LT. 2  .OR.  NNR .GT. NNRMAX)  GO TO 961
          READ(IUSRMAP, *, END=941, ERR=942) (XTAB(I), I=1,NNR)
          NLINES = NLINES + 1
          GO TO 110
       ENDIF
       IF(CBUF(1:9) .EQ. 'VERTICAL:')  THEN
          READ(CBUF(10:80), *, ERR=942)  NNV
          IF(NNV .LT. 2  .OR.  NNV .GT. NNVMAX)  GO TO 962
          READ(IUSRMAP, *, END=941, ERR=942)  (XTAB(NNR+I), I=1,NNV)
          NLINES = NLINES + 1
          GO TO 110
       ENDIF
       IF(CBUF(1:6) .EQ. 'FIELD:')  THEN
         IF(NNR .LT. 2  .OR.  NNV .LT. 2)   GO TO 963
         CALL VZERO(BRTAB(1), NNR*NNV)
         CALL VZERO(BVTAB(1), NNR*NNV)
  112    CONTINUE
          READ(IUSRMAP, '(A)', END=120, ERR=942)  CBUF
          NLINES = NLINES + 1
          READ(CBUF(1:80), *, ERR=942)  IR, IV, BR, BV
          IF(IR .LT. 1  .OR.  IV .LT. 1  .OR.
     +         IR. GT. NNR  .OR.  IV .GT. NNV)   GO TO 969
          BRTAB(IR+NNR*(IV-1)) = BR
          BVTAB(IR+NNR*(IV-1)) = BV
          GO TO 112
       ENDIF
       GO TO 942

  120  CONTINUE
**************************************************************************
*   Check if the map is completly filled (there should be no nods with   *
*   field equal zero)                                                    *
**************************************************************************
       DO  IR=1,NNR
          DO  IV=1,NNV
             IF(BRTAB(IR+NNR*(IV-1)) .EQ. 0.0  .AND.
     +                    BVTAB(IR+NNR*(IV-1)) .EQ. 0.0)   THEN
               WRITE(*,*) ' GUFLD - ZERO FIELD AT:',
     +                               XTAB(IR), XTAB(NNR+IV)
             ENDIF
          ENDDO
       ENDDO
       CALL  CYFLD(0.0, 0.0, B1(1), B1(2))
       CALL  CYFLD(2.0*MAGX, 0.0, BR, BV)
       B(1) = B1(1) - BR
       B(2) = B1(2) - BV
       BMAX3 = SQRT(B(1)**2 + B(2)**2)
       WRITE(*,*) 'MAGT = 3  BMAX = ', BMAX3
       IF(BFLD .NE. 0.0)  THEN
         WRITE(*,*) ' GUFLD - RESCALED VALUE = ', BFLD
       ELSE
         WRITE(*,*) ' GUFLD - NO RESCALING'
       ENDIF


***********************************************************
*   Output of the field parameters to Ascii OUtput file   *
***********************************************************

  199 CONTINUE
      IF(ISAOU .GT. 0)  THEN
         WRITE(IUSRAOU, '('' M MAGV '', F10.4)' )  MAGV
         WRITE(IUSRAOU, '('' M MAGT '', I10)' )  MAGT
         WRITE(IUSRAOU, '('' M MAGX '', F10.4)' )  MAGX
         WRITE(IUSRAOU, '('' M MAGZ '', F10.4)' )  MAGZ
         WRITE(IUSRAOU, '('' M RFLD '', F10.4)' )  RFLD
         WRITE(IUSRAOU, '('' M MOFF '', 3F10.4)' )  MOFF
         WRITE(IUSRAOU, '('' M BFLD '', F10.4)' )  BFLD
      ENDIF   !  ISAOU .GT. 0
      IF(ISMEM .GT. 0)  THEN
         WRITE(LINE, '('' M MAGV '', F10.4)' )  MAGV
          CALL ADDPMCRINFO(LINE)
         WRITE(LINE, '('' M MAGT '', I10)' )  MAGT
          CALL ADDPMCRINFO(LINE)
         WRITE(LINE, '('' M MAGX '', F10.4)' )  MAGX
          CALL ADDPMCRINFO(LINE)
         WRITE(LINE, '('' M MAGZ '', F10.4)' )  MAGZ
          CALL ADDPMCRINFO(LINE)
         WRITE(LINE, '('' M RFLD '', F10.4)' )  RFLD
          CALL ADDPMCRINFO(LINE)
         WRITE(LINE, '('' M MOFF '', 3F10.4)' )  MOFF
          CALL ADDPMCRINFO(LINE)
         WRITE(LINE, '('' M BFLD '', F10.4)' )  BFLD
          CALL ADDPMCRINFO(LINE)
      ENDIF   !  ISMEM .GT. 0

***********************************************************
*   Calculation of the field values                       *
***********************************************************

  200 CONTINUE
      IF(MAGT .EQ. 0)  GO TO 990   !  zero field everywhere

      VECTOFF(1) = VECT(1) - MOFF(1)
      VECTOFF(2) = VECT(2) - MOFF(2)
      VECTOFF(3) = VECT(3) - MOFF(3)

      If(MAGT .eq. 1) Then      !  perfect dipole
        X(1) = (VECTOFF(3) - MAGZ)
C Bx,Bz always zero
        B(1)=0
        B(3)=0
        If(VECTOFF(1) .gt. 0)  then   !positive side of field
          X(2) = (VECTOFF(1) - MAGX)
          if( SQRT(X(1)**2+X(2)**2) .lt. RFLD )Then
            B(2)=BFLD
          Else
            B(2)=0
          Endif
        Else
          X(2) = (VECTOFF(1) + MAGX)
          if( SQRT(X(1)**2+X(2)**2) .lt. RFLD )Then
            B(2)=-BFLD
          Else
            B(2)=0
          Endif
        endif
        Return
      Endif      !   MAGT .EQ. 1


      IF (MAGT .EQ. 2)   THEN   !   field from the 3-D map of any kind
c                               !   of magnet
        X(1) = ABS(VECTOFF(1))
        X(2) = ABS(VECTOFF(2))
        X(3) = VECTOFF(3)

        DO  220  IMAP = 1,NMAP
           NX = NA(1,IMAP)
           NY = NA(2,IMAP)
           NZ = NA(3,IMAP)
           IF(IPHOIND(IMAP) .EQ. 1)  THEN
              IF(X(1) .GT. A(IGPOS(IMAP)+NX))  GO TO 220
              IF(X(2) .GT. A(IGPOS(IMAP)+NX+NY))  GO TO 220
              IF(X(3) .LT. A(IGPOS(IMAP)+NX+NY+1)  .OR.
     +              X(3) .GT. A(IGPOS(IMAP)+NX+NY+NZ))  GO TO 220
              GO TO 221
           ELSE    ! IPHOIND(IMAP) .EQ. 0
              IF(VECT(1) .LT. A(IGPOS(IMAP)+1)  .OR.
     +              VECT(1) .GT. A(IGPOS(IMAP)+NX))  GO TO 220
              IF(VECT(2) .LT. A(IGPOS(IMAP)+NX+1)  .OR.
     +              VECT(2) .GT. A(IGPOS(IMAP)+NX+NY))  GO TO 220
              IF(VECT(3) .LT. A(IGPOS(IMAP)+NX+NY+1)  .OR.
     +              VECT(3) .GT. A(IGPOS(IMAP)+NX+NY+NZ))  GO TO 220
              GO TO 221
           ENDIF
  220   CONTINUE
        GO TO 990

  221   CONTINUE
        IF(IPHOIND(IMAP) .EQ. 0)   THEN
          B(1) = FINT(3, VECT, NA(1,IMAP), A(IGPOS(IMAP)+1),
     +                                       BX(IMPOS(IMAP)+1) )
          B(2) = FINT(3, VECT, NA(1,IMAP), A(IGPOS(IMAP)+1),
     +                                       BY(IMPOS(IMAP)+1) )
          B(3) = FINT(3, VECT, NA(1,IMAP), A(IGPOS(IMAP)+1),
     +                                       BZ(IMPOS(IMAP)+1) )
        ELSE  !  IPHOIND(IMAP) .EQ. 1
          IF(BFLD .NE. 0.0  .AND.  BMAX2(IMAP) .NE. 0.0)  THEN
            B(1) = FINT(3, X, NA(1,IMAP),
     +                   A(IGPOS(IMAP)+1), BX(IMPOS(IMAP)+1) )
     +                                         * BFLD/BMAX2(IMAP)
            B(2) = FINT(3, X, NA(1,IMAP),
     +                   A(IGPOS(IMAP)+1), BY(IMPOS(IMAP)+1) )
     +                                         * BFLD/BMAX2(IMAP)
            B(3) = FINT(3, X, NA(1,IMAP),
     +                   A(IGPOS(IMAP)+1), BZ(IMPOS(IMAP)+1) )
     +                                         * BFLD/BMAX2(IMAP)
          ELSE
            B(1) = FINT(3, X, NA(1,IMAP), A(IGPOS(IMAP)+1),
     +                                       BX(IMPOS(IMAP)+1) )
            B(2) = FINT(3, X, NA(1,IMAP), A(IGPOS(IMAP)+1),
     +                                       BY(IMPOS(IMAP)+1) )
            B(3) = FINT(3, X, NA(1,IMAP), A(IGPOS(IMAP)+1),
     +                                       BZ(IMPOS(IMAP)+1) )
          ENDIF
C----- transformations in the case the point is not in y>0, x>0 part
          IF(VECTOFF(1) .LT. 0.0)  THEN
            B(2) = -B(2)
            B(3) = -B(3)
          ENDIF
          IF(VECTOFF(2) .LT. 0.0)  THEN
            B(1) = -B(1)
            B(3) = -B(3)
          ENDIF
        ENDIF

        RETURN

  230   CONTINUE   !  outside box the field is equal 0
        go to 990

      ENDIF    !   MAGT .EQ. 2


      IF(MAGT .EQ. 3)   THEN   !   field from the 2-D map of a cylindrically
c                              !   symetric magnet

        R1 = SQRT((VECTOFF(1)-MAGX)**2 + (VECTOFF(3)-MAGZ)**2)
        R2 = SQRT((VECTOFF(1)+MAGX)**2 + (VECTOFF(3)-MAGZ)**2)

        CALL CYFLD(R1, VECTOFF(2), BR, BV)

        IF(R1 .GT. 0.1E-20)  THEN
          SINA = (VECTOFF(1)-MAGX) / R1
          COSA = (VECTOFF(3)-MAGZ) / R1
        ELSE
          SINA = 0.0
          COSA = 0.0
        ENDIF
        B1(1) = SINA * BR
        B1(2) = BV
        B1(3) = COSA * BR

        CALL CYFLD(R2, VECTOFF(2), BR, BV)

        IF(R2 .GT. 0.1E-20)  THEN
          SINA = (VECTOFF(1)+MAGX) / R2
          COSA = (VECTOFF(3)-MAGZ) / R2
        ELSE
          SINA = 0.0
          COSA = 0.0
        ENDIF
        B2(1) = - SINA * BR
        B2(2) = - BV
        B2(3) = - COSA * BR

        B(1) = B1(1) + B2(1)
        B(2) = B1(2) + B2(2)
        B(3) = B1(3) + B2(3)

        IF(BFLD .NE. 0.0)  THEN
          B(1) = B(1) * BFLD/BMAX3
          B(2) = B(2) * BFLD/BMAX3
          B(3) = B(3) * BFLD/BMAX3
        ENDIF

      ENDIF    !    MAGT .EQ. 3

      RETURN


***********************************************************
*   General error messages                                *
***********************************************************

  941 CONTINUE
      WRITE(*,*) ' *** ERROR IN GUFLD - ',
     +           'UNEXPECTED EOF IN THE FIELD MAP FILE'
      WRITE(*,*) ' *** ERROR IN GUFLD - MAP NUMBER:', NMAP
      GO TO 970

  942 CONTINUE
      WRITE(*,*) ' *** ERROR IN GUFLD - ERROR READING FIELD MAP FILE'
      WRITE(*,*) ' *** ERROR IN GUFLD - MAP NUMBER:', NMAP
      GO TO 970

  945 CONTINUE
      WRITE(*,*) ' *** ERROR IN GUFLD - ',
     +           'UNDEFINED TYPE OF THE MAGNETIC FIELD:', MAGT
      GO TO 970

  946 CONTINUE
      WRITE(*,*) ' *** ERROR IN GUFLD -',
     +           'ERROR IN THE MAGNETIC FIELD MAP FILE'
      WRITE(*,*) ' *** ERROR IN GUFLD - ',
     +           'ILLEGAL TYPE OF THE MAGNETIC FIELD:', MAGT
      WRITE(*,*) ' *** ERROR IN GUFLD - ',
     +           'ONLY  MAGT = 2  OR  MAGT = 3  ARE ALLOWED'
      GO TO 970

***********************************************************
*   Error messages for MAGT=2                             *
***********************************************************

  951 CONTINUE
      WRITE(*,*) ' *** ERROR IN GUFLD - X GRID HAS TO BE DEFINED FIRST'
      WRITE(*,*) ' *** ERROR IN GUFLD - MAP NUMBER:', NMAP
      GO TO 970

  952 CONTINUE
      WRITE(*,*) ' *** ERROR IN GUFLD - Y GRID HAS TO BE DEFINED ',
     +               'BEFORE Z GRID'
      WRITE(*,*) ' *** ERROR IN GUFLD - MAP NUMBER:', NMAP
      GO TO 970

  953 CONTINUE
      WRITE(*,*) ' *** ERROR IN GUFLD - ',
     +           'ALL GRIDS HAS TO BE DEFINED BEFORE ',
     +           'FIELD VALUES DEFINITIONS'
      WRITE(*,*) ' *** ERROR IN GUFLD - MAP NUMBER:', NMAP
      GO TO 970

  954 CONTINUE
      WRITE(*,*) ' *** ERROR IN GUFLD - ',
     +           'TOO BIG MAP (MAGT=2) NUMBER: ', NMAP
      WRITE(*,*) ' *** ERROR IN GUFLD - X, Y, Z GRID:', NX, NY, NZ
      WRITE(*,*) ' *** ERROR IN GUFLD - INCREASE PARAMETER  MAXGRID'
      GO TO 970

  955 CONTINUE
      WRITE(*,*) ' *** ERROR IN GUFLD - ',
     +           'TOO BIG MAP (MAGT=2) NUMBER: ', NMAP
      WRITE(*,*) ' *** ERROR IN GUFLD - X, Y, Z GRID:', NX, NY, NZ
      WRITE(*,*) ' *** ERROR IN GUFLD - INCREASE PARAMETER  MAXMAP'
      GO TO 970

  956 CONTINUE
      WRITE(*,*) ' *** ERROR IN GUFLD - ',
     +           'WRONG GRID IN MAP (MAGT=2) NUMBER: ', NMAP
      WRITE(*,*) ' *** ERROR IN GUFLD - X, Y, Z GRID:', NX, NY, NZ
      GO TO 970

  957 CONTINUE
      WRITE(*,*) ' *** ERROR IN GUFLD - ',
     +           'TOO MANY MAPS (MAGT=2) IN THE FILE:', NMAP
      WRITE(*,*) ' *** ERROR IN GUFLD - INCREASE PARAMETER  NMAXMAP'
      GO TO 970

  958 CONTINUE
      WRITE(*,*) ' *** ERROR IN GUFLD - ',
     +           'ILLEGAL PHOMAG PARAMETER VALUE:', IPHOIND(NMAP)
      WRITE(*,*) ' *** ERROR IN GUFLD - MAP NUMBER:', NMAP
      GO TO 970

  959 CONTINUE
      WRITE(*,*) ' *** ERROR IN GUFLD - ',
     +           'ERROR IN INDEX OF THE MAP (MAGT=2) ELEMENT:',
     +                    IX, IY, IZ
      WRITE(*,*) ' *** ERROR IN GUFLD - MAP NUMBER:', NMAP
      GO TO 970

***********************************************************
*   Error messages for MAGT=3                             *
***********************************************************

  961 CONTINUE
      WRITE(*,*) ' *** ERROR IN GUFLD - ',
     +           'TOO BIG SIZE OF THE FIELD MAP (MAGT=3)',
     +           ' OR WRONG VALUE OF R-GRID, R,V=', NNR, NNV
      GO TO 970

  962 CONTINUE
      WRITE(*,*) ' *** ERROR IN GUFLD - ',
     +           'TOO BIG SIZE OF THE FIELD MAP (MAGT=3):',
     +           ' OR WRONG VALUE OF V-GRID, R,V=', NNR, NNV
      GO TO 970

  963 CONTINUE
      WRITE(*,*) ' *** ERROR IN GUFLD - ',
     +           'R,V GRIDS HAS TO BE DEFINED BEFORE ',
     +           'FIELD VALUES DEFINITIONS'
      GO TO 970

  969 CONTINUE
      WRITE(*,*) ' *** ERROR IN GUFLD - ',
     +           'ERROR IN INDEX OF THE MAP (MAGT=3) ELEMENT:',
     +                      IR, IV
      GO TO 970

***********************************************************
*   Final error message                                   *
***********************************************************

  970 CONTINUE
      WRITE(*,*) ' *** ERROR IN GUFLD - ',
     +           'ERROR IN THE MAP FILE, ABOUT LINE: ', NLINES
      WRITE(*,*) ' *** ERROR IN GUFLD - ERROR LINE: ', CBUF
      CALL STOPER('GUFLD', 'FATAL ERROR', 1)


***********************************************************
*   Magnetic field equal 0                                *
***********************************************************

  990 CONTINUE
      B(1) = 0.0
      B(2) = 0.0
      B(3) = 0.0

      RETURN

      END
