#include <math.h>

#include "TPhStraightHough.h"
#include "TPhHit.h"
#include "TPhDetector.h"
#include "TPhLayer.h"
#include "TPhSpectSensor.h"
#include "TPhSubDetector.h"
#include "TPhEvent.h"
#include "TPhVertex.h"
#include "TPhMomentumVector.h"

ClassImp(TPhStraightHough)
////////////////////////////////////////////////////////////////////////
//
// This class implements the Hough transform for the straight tracks in 
// the spectrometer. 
// 
// The normal method of calling, is to Instanciate this class once during
// the run and then call TransformHits() to transform the hits that are in
// the sensors (which could be filled with TPhDetector::FillWithEventHits())
// and then compare any track with CheckVector() to see if it is compatible
// with the Hough transform. 
//
// Author: Gunther Roland
//
// $Id: TPhStraightHough.cxx,v 1.2 1998/09/18 03:15:38 decowski Exp $
//
////////////////////////////////////////////////////////////////////////

TPhStraightHough::TPhStraightHough()
{
  // Normal constructor

  fLimitingLayer = 3;

  fDetector = TPhDetector::GetInstance();;
  if(!fDetector)
    Warning("TPhStraightHough", "No Detector!");

  fVertex = NULL;

  fVerbose = 0;
}

void TPhStraightHough::FillTable(Int_t arm)
{
  // Fill the Hough transformed hits table with the hits in the sensors. 
  // This routine loops over the given arm and fills the Hit table. 
  // The arm can be either kNegArm or kPosArm;

  Int_t   i;
  Int_t   j;
  Int_t   k;
  Int_t   BinXZ;
  Int_t   BinYZ;
  Float_t SlopeXZ;
  Float_t SlopeYZ;

  TPhHit    *hit;
  TPhSensor *sensor;
  TPhLayer  *layer;
    
  // Clear the table.
  for(i = 0; i < kBinsXZ; i++) {
    for(j = 0; j < kBinsYZ; j++) {
      for(k = 0; k < kNLayers; k++) {
	fHits[i][j][k] = 0;
      }
    }
  }
  
  // Get one of the two spectrometer arms
  TPhSubDetector *SpectrometerArm;
  if (arm == kPosArm)
    SpectrometerArm = fDetector->GetSubDetector("SpectP");
  else 
    SpectrometerArm = fDetector->GetSubDetector("SpectN");

  // Loop over all layers and the sensors in the layers. These sensors
  // have the hits in them. The Slopes of the hits with respect to the vertex
  // are calculated and filled in the fHits table. This happens for every 
  // layer in the spectrometer. 

  for(i = 0;i < kNLayers; i++) {
    layer = SpectrometerArm->GetLayer(i);
    TIter NextSensor(layer->GetSensorList());
    while(sensor = (TPhSensor *) NextSensor()) {
      if(fVerbose == 2)
	printf("Now at sensor %s\n",sensor->GetChLabel());
      TIter NextHit(sensor->GetHitList());
      while(hit = (TPhHit *) NextHit()) {
	GetSlopes(hit, &SlopeXZ, &SlopeYZ);

	if(fVerbose==2)
	  printf("Got slope %g %g for hit %g %g %g\n",
		 SlopeXZ,
		 SlopeYZ,
		 hit->GetX(),
		 hit->GetY(),
		 hit->GetZ());

	if(GetBins(&BinXZ, &BinYZ, SlopeXZ, SlopeYZ, arm)) {
	  if(fVerbose == 2)
	    printf("Got bins %d %d\n", BinXZ, BinYZ);
	  fHits[BinXZ][BinYZ][i]++;
	}
      }
    }
  }

  if(fVerbose) {
    for(k = 0;k < kNLayers; k++) {
      for(i = 0;i < kBinsXZ; i++) {
	for(j = 0;j < kBinsYZ; j++) {
	  if(fHits[i][j][k])
	    printf("%d-%d ",i,j);  
	}
      }
      printf("\n");
    }
  }
}

Bool_t TPhStraightHough::CheckVector(TPhMomentumVector *vec)
{
  // Check the momentumvector vec if it is compatible with the hits in the
  // candidate list. The momentum information in the vector is used to 
  // linearly extrapolate in space. The slopes of this point are checked
  // against the candidate table to see if it is compatible or not. If there
  // is a hit in the candidate table in a 3x3 bin region around the 
  // extrapolated slope, then it is considered a valid candidate.
  // 
  // Returns: kTRUE if the Momentum vector is a valid candidate
  //          kFALSE if it is not a valid candidate 

  Float_t SlopeXZ;
  Float_t SlopeYZ;

  Int_t l;
  Int_t m;
  Int_t minXZ;
  Int_t maxXZ;
  Int_t minYZ;
  Int_t maxYZ;
  Int_t BinXZ;
  Int_t BinYZ;
  Int_t Arm;

  TPhHit Hit;

  // Here we use the hit as an arbitrary space point, that is given
  // by the linear extrapolation of the Momentum vector of the track
  // that we are checking. 
  Hit.SetX(0.1 * vec->GetPx()/vec->GetPz());
  Hit.SetY(0.1 * vec->GetPy()/vec->GetPz());
  Hit.SetZ(0.1);

  // Get the slope with respect to the vertex that this extrapolated 
  // hit gives
  GetSlopes(&Hit, &SlopeXZ, &SlopeYZ);

  Arm = ((SlopeXZ > 0) ? kPosArm : kNegArm);

  // Get the bins
  if(GetBins(&BinXZ, &BinYZ, SlopeXZ, SlopeYZ, Arm)) {
    if(fVerbose == 2)
      printf("Check Bins %d %d in Arm %d\n", BinXZ, BinYZ, Arm);

    // Make sure that the bin is not outside the region
    minXZ = BinXZ-1;
    if(minXZ < 0)
      minXZ = 0;
    maxXZ = BinXZ+1;
    if(minXZ > kBinsXZ)
      minXZ = kBinsXZ;

    minYZ = BinYZ-1;
    if(minYZ < 0)
      minYZ = 0;
    maxYZ = BinYZ+1;
    if(minYZ > kBinsYZ)
      minYZ = kBinsYZ;

    // If there was at least one hit in the candidate table return
    // kTRUE
    for(l = minXZ; l < maxXZ; l++) {
      for(m = minYZ; m < maxYZ;m ++) {
	if(fCand[l][m][Arm]) {
	  return kTRUE;
	}
      }
    }
  }
  return kFALSE;
}

void TPhStraightHough::GetSlopes(TPhHit *hit, 
				 Float_t *SlopeXZ, Float_t *SlopeYZ)
{
  // Calculate the slope between the hit and the vertex. The slopes
  // are returned in SlopeXZ and SlopeYZ. 

  Float_t l;
  Float_t dx;
  Float_t dy;
  Float_t dz;

  dx = hit->GetX() - fVertex->GetX();
  dy = hit->GetY() - fVertex->GetY();
  dz = hit->GetZ() - fVertex->GetZ();

  l = sqrt(dx*dx+dz*dz);
  *SlopeXZ = dx/l;
  *SlopeYZ = dy/l;
}

void TPhStraightHough::CalcMinMaxSlopes()
{
  // Loop over all layers that contain sensors until fLimitingLayer
  // and calculate the Min/Max XZ and YZ slopes for both spectrometer arms, 
  // taking the vertex into account. The Min/Max slopes are necessary to
  // calculate the different bins in XZ, YZ space that the hits will be 
  // transformed into.

  Int_t   i;
  Int_t   j;
  Int_t   k;
  Int_t   Arm;

  Float_t SlopeXZ;
  Float_t SlopeYZ;

  TPhSpectSensor *sensor;
  TPhSubDetector *SubDetector;

  TPhLayer  *layer;

  TPhHit Hit;

  for(Arm = 0; Arm < kNArms; Arm++) {
    fMinSlopeXZ[Arm] = 1000000;
    fMaxSlopeXZ[Arm] = -1000000;
    fMinSlopeYZ[Arm] = 1000000;
    fMaxSlopeYZ[Arm] = -1000000;

    if(Arm == kPosArm) 
      SubDetector = fDetector->GetSubDetector("SpectP");
    else 
      SubDetector = fDetector->GetSubDetector("SpectN");

    layer = SubDetector->GetLayer(fLimitingLayer);
    if(layer) {
      TIter NextSensor(layer->GetSensorList());
      while(sensor = (TPhSpectSensor *) NextSensor()) {
	if(fVerbose)
	  printf("Check limits using sensor %s\n",sensor->GetName());
	Hit.SetX(sensor->GetX0());
	Hit.SetY(sensor->GetY0());
	Hit.SetZ(sensor->GetZ0());
	GetSlopes(&Hit, &SlopeXZ, &SlopeYZ);

	if(SlopeXZ > fMaxSlopeXZ[Arm])
	  fMaxSlopeXZ[Arm] = SlopeXZ;
	if(SlopeXZ < fMinSlopeXZ[Arm])
	  fMinSlopeXZ[Arm] = SlopeXZ;

	if(fabs(SlopeYZ) > fMaxSlopeYZ[Arm]) {
	  fMaxSlopeYZ[Arm] = fabs(SlopeYZ);
	  fMinSlopeYZ[Arm] = -fabs(SlopeYZ);
	}

	Hit.SetX(sensor->GetX1());
	Hit.SetY(sensor->GetY1());
	Hit.SetZ(sensor->GetZ1());
	GetSlopes(&Hit, &SlopeXZ, &SlopeYZ);

	if(SlopeXZ > fMaxSlopeXZ[Arm])
	  fMaxSlopeXZ[Arm] = SlopeXZ;
	if(SlopeXZ < fMinSlopeXZ[Arm])
	  fMinSlopeXZ[Arm] = SlopeXZ;

	if(fabs(SlopeYZ) > fMaxSlopeYZ[Arm]) {
	  fMaxSlopeYZ[Arm] = fabs(SlopeYZ);
	  fMinSlopeYZ[Arm] = -fabs(SlopeYZ);
	}
      }
    }
    else {
      Error("CalcMinMaxSlopes", "Can't access layer 0. Exiting\n");
      return;
    }
    if(fVerbose)
      printf("Found slope limits %g %g %g %g\n",
	     fMinSlopeXZ[Arm], fMaxSlopeXZ[Arm], 
	     fMinSlopeYZ[Arm], fMaxSlopeYZ[Arm]);
  }
}


Bool_t TPhStraightHough::GetBins(Int_t *BinXZ, Int_t *BinYZ, 
				 Float_t SlopeXZ, Float_t SlopeYZ, Int_t Arm)
{
  // Returns the XZ and YZ bin for a given slope in XZ, YZ in an arm
  // 
  // Return: kTRUE if all was OK
  //         kFALSE when there was a problem

  Float_t d;

  if(fVerbose > 1)
    printf("Get Bin for %g %g (%g/%g   %g/%g)\n",
	   SlopeXZ, SlopeYZ, fMinSlopeXZ[Arm], fMaxSlopeXZ[Arm],
	   fMinSlopeYZ[Arm], fMaxSlopeYZ[Arm]);

  if(SlopeXZ <= fMinSlopeXZ[Arm] || SlopeXZ >= fMaxSlopeXZ[Arm])
    return kFALSE;
  else if(SlopeYZ <= fMinSlopeYZ[Arm] || SlopeYZ >= fMaxSlopeYZ[Arm])
    return kFALSE;
  else {
    d = (SlopeXZ - fMinSlopeXZ[Arm])/(fMaxSlopeXZ[Arm] - fMinSlopeXZ[Arm]);
    *BinXZ = (Int_t) (d * (Float_t) kBinsXZ);

    d = (SlopeYZ - fMinSlopeYZ[Arm])/(fMaxSlopeYZ[Arm] - fMinSlopeYZ[Arm]);
    *BinYZ = (Int_t) (d * (Float_t) kBinsYZ);
  }
  return kTRUE;
}

void TPhStraightHough::TransformHits()
{
  // Hough the hits in the sensors into a table that can later
  // be checked using CheckVector to find the compatible hits. 
  // 
  // An event should be connected to the detector object. If the event
  // does not contain a vertex, then a nominal vertex is created. 

  TPhEvent *event;

  // See if we have a vertex, create one otherwise.
  if(!fVertex) {
    if(fVerbose)
      printf("Set fVertex object\n");
    event = fDetector->GetEvent();
    if(!event) { 
      Warning("TransformHits", "Event not specified!");
      return;
    }
    fVertex = (TPhVertex *) event->GetObject("MainVertex");

    if(fVertex) {
      if(fVerbose)
	printf("Found vertex at %g,%g,%g\n",
	       fVertex->GetX(),
	       fVertex->GetY(),
	       fVertex->GetZ());
    }
    else {
      fVertex = new TPhVertex("NominalVertex");
      event->AddObject(fVertex);
      fVertex->SetX(0);
      fVertex->SetY(0);
      fVertex->SetZ(0);
      printf("Assume nominal vertex at  %g,%g,%g\n",
	     fVertex->GetX(),
	     fVertex->GetY(),
	     fVertex->GetZ());
    }
  }

  if(fVerbose)
    printf("Get limits on straight line slopes\n");
  CalcMinMaxSlopes();

  FillTable(kPosArm);
  FillCandidateTable(kPosArm);

  FillTable(kNegArm);
  FillCandidateTable(kNegArm);

}

void TPhStraightHough::FillCandidateTable(Int_t Arm)
{
  // Loops over the hittable and fills the candidate table with the number of
  // found hits in a region of 5x5 around a hit in the first plane. 

  Int_t i;
  Int_t j;
  Int_t k;
  Int_t l;
  Int_t m;
  Int_t hit_cnt;
  Int_t minXZ;
  Int_t maxXZ;
  Int_t minYZ;
  Int_t maxYZ;
  Int_t layer;


  layer = 0;

  // loop over all bins
  for(i = 0;i < kBinsXZ; i++) {
    for(j = 0; j < kBinsYZ; j++) {
      fCand[i][j][Arm] = 0;

      // If there is a hit in the first layer, then look in the other
      // layers
      if(fHits[i][j][layer]) {
	hit_cnt = 1;

	minXZ = i-2;
	if(minXZ < 0)
	  minXZ = 0;
	maxXZ = i+2;
	if(minXZ > kBinsXZ)
	  minXZ = kBinsXZ;
	
	minYZ = j-2;
	if(minYZ < 0)
	  minYZ = 0;
	maxYZ = j+2;
	if(minYZ > kBinsYZ)
	  minYZ = kBinsYZ;

	for(k = layer+1 ; k < kNLayers; k++) {
	  for(l = minXZ; l < maxXZ; l++) {
	    for(m = minYZ; m < maxYZ; m++) {
	      if(fHits[l][m][k])
		{
		  hit_cnt++;
		}
	    }
	  }
	}

	//mpd FIXME: hit_cnt == 2 does not count???
	if(hit_cnt > 2) {
	  if(fVerbose)
	    printf("Found track with %d hits at %d %d\n",hit_cnt,i,j);
	  fCand[i][j][Arm] = hit_cnt;
	}
      }
    }
  }
}
