#include "math.h"

#include "TPhTemplateManager.h"
#include "TPhTrackTemplate.h"
#include "TPhStraightHough.h"
#include "TPhDetector.h"
#include "TPhSubDetector.h"
#include "TPhLayer.h"
#include "TPhHit.h"
#include "TPhTrack.h"
#include "TPhRecTrack.h"
#include "TPhEvent.h"
#include "TPhStuff.h"
#include "TRandom.h"

#include "TStopwatch.h"
#include "TFile.h"
#include "TKey.h"

ClassImp(TPhTemplateManager)

TPhTemplateManager::TPhTemplateManager()
{
// Default constructor of TPhTemplateManager class.

    fMagnet         = NULL;
    fDetector       = TPhDetector::GetInstance();
    if(!fDetector)
        fprintf(stderr,"TPhTemplateManager: Couldn't get detector object\n");
    fTemplateList   = NULL; 
    fCandidateList  = new TObjArray();
    fHough          = NULL; 
    fStuff          = new TPhStuff();
    fMaxCompDist    = 3.5;
    fMinCompEnergy  = 0.5;

    fMinHits      = 11;
    fMinProb      = 0.01;
    fVerbose      = 0;
    fTemplateList = new TObjArray();
    fCovarList    = new TObjArray();
}

TPhTemplateManager::~TPhTemplateManager()
{
// Default destructor of TPhTemplateManager class.

  delete fTemplateList;
  delete fCovarList;

  delete fCandidateList;
  delete fStuff;
}

TPhTrackTemplate *TPhTemplateManager::GetTemplate(Float_t p, Float_t theta, Float_t dip)
{
    Float_t dist;
    Float_t min = 10000000;
    
    TPhTrackTemplate *Template     = NULL;
    TPhTrackTemplate *BestTemplate = NULL;

    if(fTemplateList)
    {
        TIter NextTemplate(fTemplateList);
        while(Template = (TPhTrackTemplate *) NextTemplate())
        {
            dist  = fabs(p-Template->GetP());
            if(dist < min)
            {
                min = dist;
                BestTemplate = Template;
            }
        }
        return BestTemplate;
    }
    else
    {
        return NULL;
    }
}

void TPhTemplateManager::Reset()
{

}

Int_t TPhTemplateManager::GetCompatibleHits(TPhTrackTemplate *t)
{
    return  GetCompatibleHits(t, fMaxCompDist);
}

Int_t TPhTemplateManager::GetCompatibleHits(TPhTrackTemplate *t, Double_t MaxDist)
{
    Int_t i, arm;
    Int_t cnt;
    Int_t nhits = 0;

    TObjArray *list;

    TPhSensor *sensor;
    TPhLayer  *layer;  
    TPhHit    *hit;

    t->SetMaxCompDist(MaxDist);

    for(i=0;i<kMaxLayers;i++)
    {
        fHitCnt[i] = 0;
    }

    Int_t n_missed = 0;
    Int_t flag;
    if(!fDetector)
    {
        fDetector = TPhDetector::GetInstance();
        if(!fDetector)
        {
            fprintf(stderr,"TPhTemplateManager::GetCompatibleHits : Can't access Detector\n");
            return 0;
        }
    }
    TPhSubDetector *SpectrometerArm[2];
    SpectrometerArm[0] = fDetector->GetSubDetector("SpectP");
    SpectrometerArm[1] = fDetector->GetSubDetector("SpectN");
    
    for (arm = 0; arm < 1; arm++) {
      TIter NextLayer(SpectrometerArm[arm]->GetLayerList());
      while((layer = (TPhLayer *) NextLayer()) && (n_missed < 3))
	{
	  if(layer->GetLabel() < kMaxLayers)
	    {
	      if(fVerbose > 3)
                printf("Check hits in layer %d (%d sensors)\n",
		       layer->GetLabel(),layer->GetSensorList()->GetEntries());
	      cnt  = fHitCnt[layer->GetLabel()];
	      if(t->GetLayerToRow(layer->GetLabel()) >=0)
		{
		  flag = 1;
		  TIter NextSensor(layer->GetSensorList());
		  while(sensor = (TPhSensor *) NextSensor())
		    {
		      if(fVerbose > 3)
			if(sensor->GetHitList())
			  printf("Check %d hits in sensor %s (%d missed)\n",
				 sensor->GetHitList()->GetEntries(),
				 sensor->GetChLabel(),n_missed);
		      TIter NextHit(sensor->GetHitList());
		      while(hit = (TPhHit *) NextHit())
			{
			  if(t->IsCompatible(hit))
			    {
			      fHitArray[layer->GetLabel()][cnt] = hit;
			      if(cnt < (kMaxHits-1))
                                cnt++;
			      if(fVerbose > 2)
				printf("Add hit in layer %d: %g %g, %g (%g)\n",
				       hit->GetLayer(),
				       hit->GetX(),  
				       hit->GetZ(),
				       t->GetDistXZ(hit),
				       t->GetMeanDistXZ(hit->GetLayer()));
			      nhits++;
			      flag = 0;
			    }
			  else
			    {
			      if(fVerbose > 2 )
				{
				  printf("Reject hit\n");
				}
			    }
			}
		    }
		}
	      fHitCnt[layer->GetLabel()] = cnt;
	      n_missed += flag;
	    }
	}
    }
    return nhits;
}

Bool_t TPhTemplateManager::ReadTemplateList(Char_t *TemplName)
{
    TKey *key;

    TFile f(TemplName);
    if(f.IsOpen())
    {
         TList *l = f.GetListOfKeys();
         TIter NextKey(l);
         if(fVerbose)
             printf("Found a total of %d templs\n",l->GetSize());

         while(key = (TKey *) NextKey())
         {
	   // TKey::Read got renamed to TKey::ReadObj in Root V1.03/09
             TPhTrackTemplate *t = (TPhTrackTemplate *) key->ReadObj();
             t->SetMaxCompDist(fMaxCompDist);
             t->SetMinCompEnergy(fMinCompEnergy);
             t->MakeStamp();
             fTemplateList->Add(t);
         }
    }
    else
    {
        fprintf(stderr,"Can't open file %s\n",TemplName);
        return kFALSE;
    }
    return kTRUE;
}

Bool_t TPhTemplateManager::ReadTemplateList(Char_t *TemplName,Char_t *CovarName)
{
    TKey *key;
// now take the templates and find the matching covariance
// matrices

    Float_t           dpx;
    Float_t           dpz;
    Float_t           dp;
    Float_t           min;
    TPhTrackTemplate *tmpl;
    TPhTrackTemplate *covar;
    TPhTrackTemplate *best; 


    TFile f2(CovarName);
    if(f2.IsOpen())
    {
         TList *l = f2.GetListOfKeys();
         TIter NextKey(l);
         if(fVerbose)
             printf("Found a total of %d covariance matrices\n",l->GetSize());

         while(key = (TKey *) NextKey())
         {
	   // TKey::Read got renamed to TKey::ReadObj in Root V1.03/09
             TPhTrackTemplate *t = (TPhTrackTemplate *) key->ReadObj();
             t->SetMaxCompDist(fMaxCompDist);
             t->SetMinCompEnergy(fMinCompEnergy);
             t->MakeStamp();
             fCovarList->Add(t);
         }
    }
    else
    {
         fprintf(stderr,"Can't open file %s\n",CovarName);
         return kFALSE;
    }

    Int_t dropped = 0;
    TFile f(TemplName);
    if(f.IsOpen())
    {
         TList *l = f.GetListOfKeys();
         TIter NextKey(l);
         if(fVerbose)
             printf("Found a total of %d templs\n",l->GetSize());

         while(key = (TKey *) NextKey())
         {
	   // TKey::Read got renamed to TKey::ReadObj in Root V1.03/09
             TPhTrackTemplate *t = (TPhTrackTemplate *) key->ReadObj();
             t->SetMaxCompDist(fMaxCompDist);
             t->SetMinCompEnergy(fMinCompEnergy);
             t->MakeStamp();
             dropped++;

             min = 1.0;
             TIter NextCovar(GetCovarList());
             while(covar = (TPhTrackTemplate *) NextCovar())
             {
                 dpx = covar->GetPx() - t->GetPx();
                 dpz = covar->GetPz() - t->GetPz();
                 dp = sqrt(dpx*dpx+dpz*dpz);
                 if(dp < min && (t->GetStamp() == covar->GetStamp()))
                 {
                     min  = dp;
                     best = covar;
                 }
             }
             if(fVerbose > 1)
             {
                 printf("fMaxCompDist %g %g\n",fMaxCompDist,t->GetMaxCompDist());
             }
             if(best)
             {
                 if(t->CloneTemplate(best))
                 {
                     fTemplateList->Add(t);
                     dropped--;
                 }
                 
                 if(fVerbose > 1)
                     printf("Best covariance matrix for %g %g is %g %g (%g) %g\n",
                         t->GetPx(),t->GetPz(),
                         best->GetPx(),best->GetPz(),min,t->GetMaxCompDist());
             }
             else
             {
                 printf("Couldn't find matrix for %g %g\n",t->GetPx(),t->GetPz());
             }
         }
         printf("%d valid templates, %d dropped\n",fTemplateList->GetEntries(),dropped);
         return kTRUE;
    }
    else
    {
        fprintf(stderr,"Can't open file %s\n",TemplName);
        return kFALSE;
    }
}
void TPhTemplateManager::CreateTrackCandidates()
{
    Float_t fraction;
    fraction = CreateTrackCandidates(3600.0, fMaxCompDist);
    if(fraction < 1.0)
    {
       printf("CreateTrackCandidates: only a fraction ( %d ) of templates was used\n", fraction);
       printf("                       due to CPU time limit\n");
    }
}

Float_t TPhTemplateManager::CreateTrackCandidates(Float_t CPUlimit)
{
  return CreateTrackCandidates(CPUlimit, fMaxCompDist);
}

Float_t TPhTemplateManager::CreateTrackCandidates(Float_t CPUlimit, Double_t MaxDist)
{
    Int_t   i;
    Int_t   nhits;
    Bool_t  flag;
    Float_t prob, chi2, ndof;

    TPhHit           *hit;
    TPhTrackTemplate *templ;
    TPhRecTrack         *candidate;

    Int_t processedtempl;
    Float_t starttime;
    TStopwatch recontimer;

//  Create candidate list if necessary

    if(!fCandidateList)
    {
        if(fVerbose)
            printf("Create list of candidates\n");
        fCandidateList = new TObjArray();
    }

//  Clear list of old track candidates

    if(fVerbose)
        printf("Clear list of candidates\n");
    fCandidateList->Delete();

// Now loop over all templs and check whether for each
// of them whether it can be matched to the hits managed 
// by the TPhDetector object

    if(fVerbose)
        printf("Loop over %d templates\n",fTemplateList->GetEntries());

    recontimer.Start();
    starttime = recontimer.GetCPUTime();
    processedtempl = 0;
    
    TIter NextTemplate(fTemplateList);
    while((templ = (TPhTrackTemplate *) NextTemplate()))
    {
        if(recontimer.GetCPUTime() - starttime > CPUlimit)
            return( ((Float_t) processedtempl) / ((Float_t) fTemplateList->GetEntries()) );
        processedtempl++;

// Fill the lists managed by fHitLists with all hits for 
// each layer that are close enough to the templ 

        if(fHough)
            flag = fHough->CheckVector(templ);
        else
            flag = kTRUE;
        if(!templ->GetCovar())
            flag = kFALSE;
        if(flag)
        {
            if(fVerbose > 2)
                printf("Check hits for template %g %g\n",templ->GetPx(),templ->GetPz());
            nhits = GetCompatibleHits(templ, MaxDist);
            if(fVerbose > 2)
                printf("Found %d hits\n",nhits);
         }
         else
         { 
             nhits = 0;
         }

// Now use combinations of the compatible hits to create
// track candidates and then check the full chi**2 probability
// of the candidates compared to the templ

        if(nhits >= fMinHits)
        {
            nhits = 0;
            for(i=0;i<kMaxLayers;i++)
            {
               if(fHitCnt[i] > 0)
                  nhits++;
            }
        }
        if(nhits >= fMinHits)
        {

            if(fVerbose > 1)
            {
                for(i=0;i<kMaxLayers;i++)
                    printf("%d ",fHitCnt[i]);
                printf("\n");
            }
            ResetCandidateLoop();
            while(candidate = GetNextCandidate())
            {
                if(fVerbose > 1)
                     printf("Calc chi2 to template\n");
                templ->SetVerbose(0);
                templ->CalcChi2(candidate, &chi2, &ndof);
                candidate->SetChi2(chi2);
                candidate->SetNDof(ndof);
                prob = candidate->CountProb(fStuff);

// if the track candidate passes the chi**2 cut, fill some quantities
// and add to the list of candidates

                if(prob > fMinProb)
                {
                    candidate->SetProb(prob);
                    candidate->SetMass(templ->GetMass());
                    candidate->SetCharge(templ->GetCharge());
                    candidate->SetPx(templ->GetPx());
                    candidate->SetPy(templ->GetPy());
                    candidate->SetPz(templ->GetPz());
                    candidate->SetStatus(1);
                    fCandidateList->Add(candidate);

                    if(fVerbose > 1)
                        candidate->ListHits();
                    if(fVerbose)
                       printf("Keep candidate with %g %g %g prob:%g\n",
                                   candidate->GetPx(),
                                   candidate->GetPy(),
                                   candidate->GetPz(), prob);
                }
                else
                {
                    if(fVerbose > 1)
                        printf("Delete candidate with %g %g %g prob:%g\n",
                                   templ->GetPx(),
                                   templ->GetPy(),
                                   templ->GetPz(), prob);
                    delete candidate;
                }
            }
        }
    }
//mpd Reset the variables that are used in the Straight Hough 
//mpd transform, to make it ready for the next event. 
    if(fHough)
        fHough->Reset();

    if(fVerbose)
        printf("Found %d track candidates\n",fCandidateList->GetEntries());

    return( 1.0 );
}
void TPhTemplateManager::ResetCandidateLoop()
{
    Int_t i;
    for(i=0;i<kMaxLayers;i++)
    {
        fCurrentIndex[i] = 0;
    }
    fCurrentLayer = 0;
}

TPhRecTrack *TPhTemplateManager::GetNextCandidate()
{
    Int_t i;
    Int_t flag;
    TPhHit *hit;

    TPhRecTrack *candidate = new TPhRecTrack();

    if(fCurrentLayer < kMaxLayers)
    {
        if(fVerbose > 2)
        {
            printf("Layer %d:",fCurrentLayer);
            for(i=0;i<kMaxLayers;i++)
               printf("%d ",fCurrentIndex[i]);
            printf("\n");
        }
        for(i=0;i<kMaxLayers;i++)
        {
            if(fCurrentIndex[i] < fHitCnt[i])
            {
                  hit = (TPhHit *) fHitArray[i][fCurrentIndex[i]];
                  if(fVerbose > 2)
                       printf("Add hit %x in layer %d to candidate\n",hit,hit->GetLayer());
                  candidate->AddHit(hit);
            }
        }
    }
    else
    {
        delete candidate;
        return NULL;
    }
    flag = kTRUE;
    fCurrentLayer = 0;
    while(flag && fCurrentLayer < kMaxLayers)
    {
        if(fCurrentIndex[fCurrentLayer] < fHitCnt[fCurrentLayer]-1)
        {
            fCurrentIndex[fCurrentLayer]++;
            flag = kFALSE;
        }
        else
        {
            for(i=0;i<fCurrentLayer;i++)
               fCurrentIndex[i] = 0;
            fCurrentLayer++;
        }
    }    
    if(candidate->GetNHits() >= fMinHits)
    {
        return candidate;
    }
    else 
    {
        delete candidate;
        return NULL;
    }
}
