#include "math.h"

#include "TPhTrackTemplFinder.h"
#include "TPhMCTrackTemplate.h"
#include "TPhStraightHough.h"
#include "TPhTrackSeedFinder.h"
#include "TPhDetector.h"
#include "TPhMerging.h"
#include "TPhHit.h"
#include "TPhRecHit.h"
#include "TPhTrack.h"
#include "TPhRecTrack.h"
#include "TPhEvent.h"
#include "TRandom.h"
#include "TMath.h"

#include "TStopwatch.h"
#include "TFile.h"
#include "TKey.h"

ClassImp(TPhTrackTemplFinder)

TPhTrackTemplFinder::TPhTrackTemplFinder()
{
// Default constructor of TPhTrackTemplFinder class.

    fDetector       = TPhDetector::GetInstance();
    if(!fDetector)
        fprintf(stderr,"TPhTrackTemplFinder: Couldn't get detector object\n");
    fMerging        = NULL;
    fTemplateList   = NULL; 
    fCandidateList  = new TObjArray();
    fHough          = NULL; 
    fSeed           = NULL;
    fMaxCompDist    = 3.5;
    fMinCompEnergy  = 0.5;

    fMinHits      = 11;
    fMinProb      = 0.01;
    fVerbose      = 0;
    fTemplateList = new TObjArray();
    fCovarList    = new TObjArray();
}

TPhTrackTemplFinder::~TPhTrackTemplFinder()
{
// Default destructor of TPhTrackTemplFinder class.

  delete fTemplateList;
  delete fCovarList;

  delete fCandidateList;
}
TPhMCTrackTemplate *TPhTrackTemplFinder::GetTemplate(Float_t ptot, Float_t theta, Float_t dip)
{
   Float_t px, pz, thr;

   thr = TMath::Pi() * theta / 180.0;
   px = ptot * TMath::Sin(thr);
   pz = ptot * TMath::Cos(thr);
   return GetTemplatePxPz(px,pz);
}

TPhMCTrackTemplate *TPhTrackTemplFinder::GetTemplatePxPz(Float_t px, Float_t pz)
{
    Float_t dist;
    Float_t min = 10000000;
    
    TPhMCTrackTemplate *Template     = NULL;
    TPhMCTrackTemplate *BestTemplate = NULL;

    if(fTemplateList)
    {
        TIter NextTemplate(fTemplateList);
        while( (Template = (TPhMCTrackTemplate *) NextTemplate()) )
        {
            dist  = fabs(px-Template->GetPx()) + fabs(pz-Template->GetPz());
            if(dist < min)
            {
                min = dist;
                BestTemplate = Template;
            }
        }
        return BestTemplate;
    }
    else
    {
        return NULL;
    }
}

void TPhTrackTemplFinder::Reset()
{

}

Int_t TPhTrackTemplFinder::GetCompatibleHits(TPhMCTrackTemplate *t)
{
    return  GetCompatibleHits(t, fMaxCompDist, 0.0, -1);
}

Int_t TPhTrackTemplFinder::GetCompatibleHits(TPhMCTrackTemplate *t, Double_t MaxDist)
{
    return  GetCompatibleHits(t, MaxDist, 0.0, -1);
}

Int_t TPhTrackTemplFinder::GetCompatibleHits(TPhMCTrackTemplate *t, Double_t MaxDist, Double_t SlopeY, Int_t SlopeYLayer)
{
    Int_t i, index1, index2;
    Int_t cnt, nh, layerno;
    Int_t nhits = 0;

    TPhRecHit    *hit;

    t->SetMaxCompDist(MaxDist);

    Int_t n_missed = 0;
    Int_t flag;
    Bool_t  comp;

    if(!fMerging)
    {
       fprintf(stderr, "TPhTrackTemplFinder: can not access merged hits\n");
       return 0;
    }

    for(layerno=0; layerno < kMaxLayerNo; layerno++) fHitCnt[layerno] = 0;

    for(layerno=0; layerno < kMaxLayerNo; layerno++)
    {
       cnt  = 0;
       if(t->GetLayerToRow(layerno) >=0)
       {
          flag = 1;
          if((nh = fMerging->GetSpecHitIndexes(layerno, t->GetMinDist45(layerno), t->GetMaxDist45(layerno), &index1, &index2)) > 0)
          {
             if(fVerbose > 3)
                printf("Check hits in layer %d (%d hits)\n", layerno, nh);
             for(i=index1; i<=index2; i++)
             {
                hit = fMerging->GetHit(i);

                if(SlopeYLayer < 0)
                    comp = t->IsCompatible(hit);
                else
                    comp = t->IsCompatible(hit, SlopeY, SlopeYLayer);

                if(comp)
                {
                   fHitArray[layerno][cnt] = hit;
                   if(cnt < (kMaxCompatHits-1))    cnt++;
                   if(fVerbose > 2)
                      printf("Add hit in layer %d: %g %g, %g (%g)\n",
                        hit->GetLayer(),
                        hit->GetX(),  
                        hit->GetZ(),
                        t->GetDistXZ(hit),
                        t->GetMeanDistXZ(hit->GetLayer()));
                   if(flag)  nhits++;
                   flag = 0;
                }
                else
                {
                   if(fVerbose > 2 ) printf("Reject hit\n");
                }   //   if(t->IsCompatible(hit))
             }   //   for(i=index1; i<=index2; i++)
          }   //   if(nhits = fMerging->GetHitIndexes( ...

          fHitCnt[layerno] = cnt;
          n_missed += flag;
          if(n_missed > 2)  return nhits;

       }   //   if(t->GetLayerToRow(layerno) >=0)
    }   //   for(layerno=0; layerno < kMaxLayerNo; layerno++)

    return nhits;
}

Bool_t TPhTrackTemplFinder::ReadTemplateList(Char_t *TemplName)
{
    TKey *key;

    TFile f(TemplName);
    if(f.IsOpen())
    {
         TList *l = f.GetListOfKeys();
         TIter NextKey(l);
         if(fVerbose)
             printf("Found a total of %d templs\n",l->GetSize());

         while( (key = (TKey *) NextKey()) )
         {
	   // TKey::Read got renamed to TKey::ReadObj in Root V1.03/09
             TPhMCTrackTemplate *t = (TPhMCTrackTemplate *) key->ReadObj();
             t->SetMaxCompDist(fMaxCompDist);
             t->SetMinCompEnergy(fMinCompEnergy);
             t->MakeStamp();
             fTemplateList->Add(t);
         }
    }
    else
    {
        fprintf(stderr,"Can't open file %s\n",TemplName);
        return kFALSE;
    }
    return kTRUE;
}

Bool_t TPhTrackTemplFinder::ReadTemplateList(Char_t *TemplName,Char_t *CovarName)
{
    TKey *key;
// now take the templates and find the matching covariance
// matrices

    Float_t           dpx;
    Float_t           dpz;
    Float_t           dp;
    Float_t           min;
    TPhMCTrackTemplate *covar;
    TPhMCTrackTemplate *best; 


    TFile f2(CovarName);
    if(f2.IsOpen())
    {
         TList *l = f2.GetListOfKeys();
         TIter NextKey(l);
         if(fVerbose)
             printf("Found a total of %d covariance matrices\n",l->GetSize());

         while( (key = (TKey *) NextKey()) )
         {
	   // TKey::Read got renamed to TKey::ReadObj in Root V1.03/09
             TPhMCTrackTemplate *t = (TPhMCTrackTemplate *) key->ReadObj();
             t->SetMaxCompDist(fMaxCompDist);
             t->SetMinCompEnergy(fMinCompEnergy);
             t->MakeStamp();
             fCovarList->Add(t);
         }
    }
    else
    {
         fprintf(stderr,"Can't open file %s\n",CovarName);
         return kFALSE;
    }

    Int_t dropped = 0;
    TFile f(TemplName);
    if(f.IsOpen())
    {
         TList *l = f.GetListOfKeys();
         TIter NextKey(l);
         if(fVerbose)
             printf("Found a total of %d templs\n",l->GetSize());

         while( (key = (TKey *) NextKey()) )
         {
	   // TKey::Read got renamed to TKey::ReadObj in Root V1.03/09
             TPhMCTrackTemplate *t = (TPhMCTrackTemplate *) key->ReadObj();
             t->SetMaxCompDist(fMaxCompDist);
             t->SetMinCompEnergy(fMinCompEnergy);
             t->MakeStamp();
             dropped++;

             min = 1.0;
             TIter NextCovar(GetCovarList());
             while( (covar = (TPhMCTrackTemplate *) NextCovar()) )
             {
                 dpx = covar->GetPx() - t->GetPx();
                 dpz = covar->GetPz() - t->GetPz();
                 dp = sqrt(dpx*dpx+dpz*dpz);
                 if(dp < min && (t->GetStamp() == covar->GetStamp()))
                 {
                     min  = dp;
                     best = covar;
                 }
             }
             if(fVerbose > 1)
             {
                 printf("fMaxCompDist %g %g\n",fMaxCompDist,t->GetMaxCompDist());
             }
             if(best)
             {
                 if(t->CloneTemplate(best))
                 {
                     fTemplateList->Add(t);
                     dropped--;
                 }
                 
                 if(fVerbose > 1)
                     printf("Best covariance matrix for %g %g is %g %g (%g) %g\n",
                         t->GetPx(),t->GetPz(),
                         best->GetPx(),best->GetPz(),min,t->GetMaxCompDist());
             }
             else
             {
                 printf("Couldn't find matrix for %g %g\n",t->GetPx(),t->GetPz());
             }
         }
         printf("%d valid templates, %d dropped\n",fTemplateList->GetEntries(),dropped);
         return kTRUE;
    }
    else
    {
        fprintf(stderr,"Can't open file %s\n",TemplName);
        return kFALSE;
    }
}
void TPhTrackTemplFinder::CreateTrackCandidates()
{
    Float_t fraction;
    fraction = CreateTrackCandidates(3600.0, fMaxCompDist);
    if(fraction < 1.0)
    {
       printf("CreateTrackCandidates: only a fraction ( %9.6f ) of templates was used\n", fraction);
       printf("                       due to CPU time limit\n");
    }
}

void TPhTrackTemplFinder::AddTemplate(TPhMCTrackTemplate *t)
{
   //  Add single template to the template list
   if(t == NULL)
   {
      printf("TPhTrackTemplFinder::AddTemplate: NULL template ca not be added\n");
   }
   else
   {
      t->SetMaxCompDist(fMaxCompDist);
      t->SetMinCompEnergy(fMinCompEnergy);
      t->MakeStamp();
      fTemplateList->Add(t);
   }
}


Float_t TPhTrackTemplFinder::CreateTrackCandidates(Float_t CPUlimit)
{
  return CreateTrackCandidates(CPUlimit, fMaxCompDist);
}

Float_t TPhTrackTemplFinder::CreateTrackCandidates(Float_t CPUlimit, Double_t MaxDist)
{
    Int_t   i, is, slayer;
    Int_t   nhits, nseeds;
    Float_t prob, chi2, ndof;
    Double_t px, py, pz, pxz, slopey;

    TPhMCTrackTemplate *templ;
    TObjArray          *seedlist;
    TPhRecTrack        *candidate;
    TPhRecTrack        *seedtrack;

    Int_t processedtempl;
    Float_t starttime;
    TStopwatch recontimer;

//  Create candidate list if necessary

    if(!fCandidateList)
    {
        if(fVerbose)
            printf("Create list of candidates\n");
        fCandidateList = new TObjArray();
    }

//  Clear list of old track candidates

    if(fVerbose)
        printf("Clear list of candidates\n");
    fCandidateList->Delete();

// Now loop over all templs and check whether for each
// of them whether it can be matched to the hits managed 
// by the TPhDetector object

    if(fVerbose)
        printf("Loop over %d templates\n",fTemplateList->GetEntries());

    recontimer.Start();
    starttime = recontimer.GetCPUTime();
    processedtempl = 0;
    
    TIter NextTemplate(fTemplateList);
    while((templ = (TPhMCTrackTemplate *) NextTemplate()))
    {
        if(recontimer.GetCPUTime() - starttime > CPUlimit)
            return( ((Float_t) processedtempl) / ((Float_t) fTemplateList->GetEntries()) );
        processedtempl++;

// Fill the lists managed by fHitLists with all hits for 
// each layer that are close enough to the templ 

        nseeds = 1;
        if(fHough)   
        {
            nseeds = 0;
            if(fHough->CheckVector(templ))   nseeds = 1;
        }
        if(fSeed)       
        {
            seedlist = fSeed->GetCompatibleSeedList(templ);
            nseeds = seedlist->GetEntries();
            slayer = 3;  // temporary - until GetMulguisinLayer is available
            for(slayer = 3; templ->GetRow(slayer) == -1  &&  slayer >= 0; slayer--)
            if(slayer < 0) nseeds = 0;
        }
        if(!templ->GetCovar())   nseeds = 0;

        for(is = 0;  is < nseeds; is++)
        {
            if(fVerbose > 2)
               printf("Check hits for template %g %g\n",templ->GetPx(),templ->GetPz());
            if(fSeed)
            {
               seedtrack = (TPhRecTrack *) seedlist->At(is);
               px = seedtrack->GetPx();
               py = seedtrack->GetPy();
               pz = seedtrack->GetPz();
               pxz = sqrt(px*px + pz*pz);
               if(pxz > 0.001)
                  slopey = py / pxz;
               else
                  slopey = 100.0;
               nhits = GetCompatibleHits(templ, MaxDist, slopey, slayer);
            }
            else
            {
               nhits = GetCompatibleHits(templ, MaxDist);
            }
            if(fVerbose > 2)
                printf("Found %d hits\n",nhits);

// Now use combinations of the compatible hits to create
// track candidates and then check the full chi**2 probability
// of the candidates compared to the templ

           if(nhits >= fMinHits)
           {
               nhits = 0;
               for(i=0;i<kMaxLayerNo;i++)
               {
                  if(fHitCnt[i] > 0)
                     nhits++;
               }
           }
           if(nhits >= fMinHits)
           {

               if(fVerbose > 1)
               {
                   for(i=0;i<kMaxLayerNo;i++)
                       printf("%d ",fHitCnt[i]);
                   printf("\n");
               }
               ResetCandidateLoop();
               while( (candidate = GetNextCandidate()) )
               {
                   if(fVerbose > 1)
                       printf("Calc chi2 to template\n");
                   templ->SetVerbose(fVerbose);
                   templ->CalcChi2(candidate, &chi2, &ndof);
                   templ->SetVerbose(0);
                   candidate->SetNDof(ndof);
                   candidate->SetChi2(chi2);
                   prob = TMath::Prob(candidate->GetChi2(),
				       candidate->GetNDof());

// if the track candidate passes the chi**2 cut, fill some quantities
// and add to the list of candidates

                   if(prob > fMinProb)
                   {
                       candidate->SetProb(prob);
                       candidate->SetMass(templ->GetMass());
                       candidate->SetCharge(templ->GetCharge());
                       candidate->SetPx(templ->GetPx());
                       candidate->SetPy(templ->GetPy());
                       candidate->SetPz(templ->GetPz());
                       candidate->SetStatus(1);
                       fCandidateList->Add(candidate);

                       if(fVerbose > 1)
                           candidate->ListHits();
                       if(fVerbose)
                          printf("Keep candidate with %g %g %g prob:%g\n",
                                   candidate->GetPx(),
                                   candidate->GetPy(),
                                   candidate->GetPz(), prob);
                   }
                   else
                   {
                       if(fVerbose > 1)
                          printf("Delete candidate with %g %g %g prob:%g\n",
                                   templ->GetPx(),
                                   templ->GetPy(),
                                   templ->GetPz(), prob);
                       delete candidate;
                   }   //   if(prob > fMinProb)
               }   //   while( (candidate = ...
           }   //   if(nhits >= fMinHits)
        }   //   for(is = 0;  is < nseeds; is++)
    }   //   while((templ = ...

//mpd Reset the variables that are used in the Straight Hough 
//mpd transform, to make it ready for the next event. 
    if(fHough)
        fHough->Reset();

    if(fVerbose)
        printf("Found %d track candidates\n",fCandidateList->GetEntries());

    return( 1.0 );
}

void TPhTrackTemplFinder::ResetCandidateLoop()
{
    Int_t i;
    for(i=0;i<kMaxLayerNo;i++)
    {
        fCurrentIndex[i] = 0;
    }
    fCurrentLayer = 0;
}

TPhRecTrack *TPhTrackTemplFinder::GetNextCandidate()
{
    Int_t i;
    Int_t flag;
    TPhRecHit *hit;

    TPhRecTrack *candidate = new TPhRecTrack();

    if(fCurrentLayer < kMaxLayerNo)
    {
        if(fVerbose > 2)
        {
            printf("Layer %d:",fCurrentLayer);
            for(i=0;i<kMaxLayerNo;i++)
               printf("%d ",fCurrentIndex[i]);
            printf("\n");
        }
        for(i=0;i<kMaxLayerNo;i++)
        {
            if(fCurrentIndex[i] < fHitCnt[i])
            {
                  hit = (TPhRecHit *) fHitArray[i][fCurrentIndex[i]];
                  if(fVerbose > 2)
                       printf("Add hit in layer %d to candidate\n",hit->GetLayer());
                  candidate->AddHit(hit);
            }
        }
    }
    else
    {
        delete candidate;
        return NULL;
    }
    flag = kTRUE;
    fCurrentLayer = 0;
    while(flag && fCurrentLayer < kMaxLayerNo)
    {
        if(fCurrentIndex[fCurrentLayer] < fHitCnt[fCurrentLayer]-1)
        {
            fCurrentIndex[fCurrentLayer]++;
            flag = kFALSE;
        }
        else
        {
            for(i=0;i<fCurrentLayer;i++)
               fCurrentIndex[i] = 0;
            fCurrentLayer++;
        }
    }    
    if(candidate->GetNHits() >= fMinHits)
    {
        return candidate;
    }
    else 
    {
        delete candidate;
        return NULL;
    }
}
