#ifndef _PhASCIIData_INCLUDED_
#define _PhASCIIData_INCLUDED_

/////////////////////////////////////////////////////////////////////////
//                                                                     //
//    TPhASCIIData                                                     //
//                                                                     //
//    PHOBOS raw data class                                            //
//                                                                     //
//    TPhASCIIData handles raw data I/O in ASCII format                //
//                                                                     //
//    Author  : Gunther Roland (rolandg@vxcern.cern.ch)                //
//    Created : 01/12/95                                               //
//    Version : 1.0                                                    //
//    Changed : 15/09/97 Andrzej Olszewski                             //
//     Added line number tracing when reading AOU header,              //
//     Phobos internal plane and layer number tables, PhobosLayer      //
//     and DetName added, provide relation between sensor name         //
//     and layer number, and allow processing of sensors types         //
//     by number and not by name,                                      //
//     Data are stored in TPhMCTrack, TPhMCHit and TPhHitArray         //
//     objects added to MCTracks, MCHits and MCHitArrays               //
//     TPhObjectContainers objects in the constructed TPhEvent objects,//
//     New constructor!                                                //
//     TPhDetector pointer is required, TPhDetector object is now      //
//     filled with AOU geometry info if version number in the object   //
//     provided (in the (new) contructor call) is equal to zero.       //
//    Changed 15.06.98 A.Olszewski                                     //
//     Updated to the new TPhDetector filling scheme                   //
//     Corrections to PMC conventions in geometry ver. 4.2 introduced  //
//    Changed: 10/12/98 Andrzej Olszewski                              //
//     New header information recognized and put into                  //
//     TPhParKey objects                                               //
//                                                                     //
// $Id: TPhASCIIData.h,v 1.9 1998/12/21 13:36:45 olszewsk Exp $ 
//                                                                     //
/////////////////////////////////////////////////////////////////////////

#include "TObject.h"
#include "THashTable.h"
#include "THashList.h"
#include "TRotMatrix.h"
#include <stdio.h>

class TPhMCHit;  
class TPhMCTrack;  
class TPhHitArray;
class TPhSensor;  
class TPhDetector;
class TPhEDigi;
class TPhEvent;  
class TPhRun;  

const  Int_t maxDEVO = 5000;    //! Maximum number of GEANT detectors
const  Int_t maxDetectorRotMtx = 682;   //! Maximum number of rot. matrices 
                                        //  (the list can be expanded)
class TPhASCIIData : public TObject {

public:
    TPhASCIIData(); 
    TPhASCIIData(Char_t *filename, Int_t runno=0, Int_t firstevent=0);

    virtual ~TPhASCIIData();

    Bool_t       IsOpen();
    Bool_t       Close();

    void         SetVerbose(Int_t Verbose)     { fVerbose = Verbose; }
    Int_t        GetVerbose()                  { return fVerbose;    }

    // force change of the TPhDetector pointer in unusuall cases
    // when an externaly creted TPhDetector object is to be used.
    TPhDetector *SetDetector(TPhDetector *det) { fDetector = det; return det; }
    TPhDetector *GetDetector();

    TPhRun      *GetRun() { return fRun;  }
    TPhEvent    *GetNextEvent();

private:

    Bool_t        Open(Char_t *fname);
    Bool_t        GetrHeader()                 { return frHeader;    }
    Bool_t        ReadASCIIHeader();
    Bool_t        AddASCIIRInfo(Char_t *llread);
    Bool_t        AddASCIIEInfo(Char_t *llread);
    Bool_t        CreateDetector(Char_t *); 
    TPhMCTrack   *GetNextTrack(Char_t *);
    TPhMCHit     *GetNextHit(Char_t *, TPhMCTrack *);
    TPhHitArray  *GetNextHitArray(Char_t *); 

    Int_t  flinenr;                //! Line counter while reading AOU file 
    Bool_t frHeader;               //! Header read mark               
    Int_t  fRunNo;                 //! Run Number for events from this file
    Int_t  fEventCounter;          //! Event counter        

    TPhDetector *fDetector;        //! Pointer to a TPhDetector object
    TPhEDigi    *fEDigi;           //! Pointer to a TPhEDigi object
    TPhEvent    *fEvent;           //! Current event object
    TPhRun      *fRun;             //! Current run object
    Char_t      fASCIIEGName[120]; //! MC event generator name 
    Int_t       fVerbose;          //! Verbosity
    Char_t fFileName[256];         //  ASCII file name containing raw hits
    FILE  *fFilePointer;           //! File pointer to ASCII file


public:
    ClassDef(TPhASCIIData,1) // Interface to Wozniak ASCII format

};

#endif



