#include <stdlib.h>
#include <math.h>
#include <unistd.h>

#include "TMath.h"

#include "TPhASCIIData.h"
#include "TPhObjectContainer.h"
#include "TPhHashContainer.h"

#include "TPhMCHit.h"
#include "TPhMCTrack.h"
#include "TPhEvent.h"
#include "TPhRun.h"

#include "TPhHitArray.h"
#include "TPhEDigi.h"
#include "TPhDetector.h"
#include "TPhSensor.h"

#include "TPhParManager.h"

#undef DEBUG
//#define DEBUG 1

ClassImp(TPhASCIIData)
////////////////////////////////////////////////////////////////
//                                                            //
// TPhASCIIData provides an interface to PMC output in ASCII  //
// format. This class has been prepared by A. Olszewski.      //
//                                                            //
// $Id: TPhASCIIData.cxx,v 1.12 1999/03/12 07:16:00 decowski Exp $
//
////////////////////////////////////////////////////////////////

TPhASCIIData::TPhASCIIData()
{
//  Constructor. Set fFilePointer and fEventCounter to NULL
    fFilePointer  = NULL;
    fEventCounter = 0;
    frHeader = kFALSE;
    fDetector = NULL;
    fEDigi = NULL;
    fEvent = NULL;
    fRun = NULL;
    fASCIIEGName[0] = '\0';
    flinenr = 0;
    fVerbose = 0;
}

TPhASCIIData::TPhASCIIData(Char_t *filename, Int_t runno, Int_t firstevent)
{
  // Constructor. Set fFilePointer and fEventCounter to NULL
  // and open the file pointed to by *filename, set the TPhDetector pointer.
  // The previous detector object is deleted and a new one is created 
  // from AOU info!

    fFilePointer  = NULL;
    fRunNo = runno;
    fEventCounter = firstevent;
    frHeader = kFALSE;
    fDetector   = TPhDetector::CreateInstance();
    fEDigi = NULL;
    fEvent = NULL;
    fRun = new TPhRun(runno);
    fASCIIEGName[0] = '\0';
    flinenr = 0;
    fVerbose = 0;
    Open(filename);
}

TPhASCIIData::~TPhASCIIData() 
{ 
  // Destructor does not remove TPhEvent related objects 
  // although they are created inside this class!
  // This should be probably changed, by calling GetNextEvent with a pointer
  // to an empty TPhEvent object, created outside of this class.
 
    if(fFilePointer) fclose(fFilePointer); 
    if(fEDigi)    delete fEDigi;
}

Bool_t TPhASCIIData::Open(Char_t *fname)
{
// Open the ASCII data file *fname. Returns kFALSE
// if the file can't be opened.
// Otherwise the fFileName data member is set to the 
// corresponding filepointer

    if(!fname)
    {
        Warning("Open","Can't open file NULL");
        return kFALSE;
    }

    sprintf(fFileName,"%s",fname);
    fFilePointer = fopen(fFileName,"r");
    if(!fFilePointer)
    {
        Warning("Open","Can't open File %s",fFileName);
        return kFALSE;
    }
    else
    {
        return kTRUE;
    }
}

Bool_t TPhASCIIData::IsOpen() 
{
  // Returns kTRUE in case file is open and kFALSE if file is not open.

  return fFilePointer ? kTRUE : kFALSE;
}

Bool_t TPhASCIIData::Close()
{
// Close the file pointed to by fFilePointer. Returns 
// kFALSE if no file was open.

    if(fFilePointer)
    {
        fclose(fFilePointer);
        fFilePointer = NULL;
        return kTRUE;
    }
    else
    {
        return kFALSE;
    }
}

TPhDetector *TPhASCIIData::GetDetector()
{
  // Reads header from the ASCII data file. If appropriate section is found 
  // creates detector object and return pointer to this object.

    if( !GetrHeader() )
    {
#ifdef DEBUG
       printf("Reading AOU file header\n");
#endif
       if ( !ReadASCIIHeader() ) return NULL;
    }
    return fDetector;
}

TPhEvent *TPhASCIIData::GetNextEvent()
{
  // Reads one event from the ASCII data file. A TPhEvent 
  // object is created and the MCHits, MCTracs and MCHitArrays data containers
  // are filled with data read from the file and connected to the list 
  // in the TPhEvent object. 
  // 
  // Releasing the memory for this event is the responsibility
  // of the function/macro calling GetNextEvent(). The memory
  // is completely released by a call to the TPhEvent destructor
  // (e.g. delete event;)

    Int_t      i;
    Int_t      num;

    Char_t     name[64];
    Char_t     llread[120];
 
    TPhMCTrack     *track;
    TPhMCHit       *hit; 
    TPhHitArray    *hitArray; 
    TObjArray      *HitList;

    if( !GetrHeader() )
    {
#ifdef DEBUG
       printf("Reading AOU file header\n");
#endif
       if ( !ReadASCIIHeader() ) return NULL;
    }
    flinenr++;
    if( fgets (llread, 120, fFilePointer) == NULL ) return NULL;

#ifdef DEBUG
    printf("Create new event\n");
#endif

    fEvent= new TPhEvent(fRunNo,fEventCounter);
    // The tracks and hits are stored in an object container and 
    // the hitarrays are stored in a hash container (since they are named 
    // with SetName). 
    TPhObjectContainer *TrackBuffer = new TPhObjectContainer("MCTracks");
    TPhObjectContainer *HitBuffer   = new TPhObjectContainer("MCHits");
    TPhHashContainer *HitArrays   = new TPhHashContainer("MCHitArrays");

    fEvent->AddObject(TrackBuffer);
    fEvent->AddObject(HitBuffer);
    fEvent->AddObject(HitArrays);

    // Loop over the rest of run headers until 
    // event or particle line are found
    while( AddASCIIRInfo(llread) ) {
      flinenr++;
      if( fgets(llread, 120, fFilePointer) == NULL ) return NULL;
    }

    // Loop over event headers until found particle line
    while( AddASCIIEInfo(llread) ) {
      flinenr++;
      if( fgets(llread, 120, fFilePointer) == NULL ) return NULL;
    }

//  Loop over tracks until EOE (end of event) line is found.
    while((track = GetNextTrack(llread)) != NULL)
    {
#ifdef DEBUG
      //gSystem->Sleep(1000);
#endif
         TrackBuffer->Add(track);
         num = 0;
         if( (HitList = track->GetHitList()) != NULL ) 
	   num = HitList->GetEntries();
#ifdef DEBUG
         printf("%d hits on this track\n",num);
#endif
         for(i=0;i<num;i++)
         {
             hit = (TPhMCHit *) HitList->At(i);
             HitBuffer->Add(hit);
         }
         flinenr++;
         if( fgets(llread, 120, fFilePointer) == NULL ) break;
         if( (strncmp (llread, " EOE", 4) == 0) ) break; 
    }

//   Check for the DIGI section marked by "D" at the start of the line
//   Loop over detectors until EOE (end of event) line is found.
    TPhHitArray *tempArray;
    if( strncmp (llread, " D", 2) == 0 ) {
       while((hitArray = GetNextHitArray(llread)) != NULL)
       {
#ifdef DEBUG
	 //        gSystem->Sleep(1000);
#endif
           HitArrays->Add(hitArray);
           flinenr++;
           tempArray = hitArray;
           if( fgets(llread, 120, fFilePointer) == NULL ) break;
           if( (strncmp (llread, " EOE", 4) == 0) ) break; 
       }
    }

//  Set file pointer to the beginning of the next event,
//  if something wrong went with the present one
    while( strncmp (llread, " EOE", 4) != 0) {
         if( fgets(llread, 120, fFilePointer) == NULL ) break;
    } 

    sprintf(name,"event_%d_%d",fEvent->GetNRun(),fEvent->GetNEvent());
    fEvent->SetName(name);
    sprintf(name,"Phobos Event %d %d",fEvent->GetNRun(),fEvent->GetNEvent());
    fEvent->SetTitle(name);
#ifdef DEBUG
    printf("Event %d has %d hits, %d tracks \n",
              fEvent->GetNEvent(),
              HitBuffer->GetEntries(),
              TrackBuffer->GetEntries());
#endif
    fEventCounter++;
    fRun->AddEvent(fEvent);
    return fEvent;    
}

Bool_t TPhASCIIData::AddASCIIRInfo(Char_t *runinfo)
{
#ifdef DEBUG
   printf("AddASCIIRInfo was called with:\n Run parameter:%s\n", runinfo);
#endif
   Char_t temp[120];
   Char_t ind[2], keyname[20], keyvalue[120];
   TPhParManager *parman = NULL;
   TPhParKey *parkey = NULL;
   Char_t sourceMC[]="MC", sourcePMC[]="PMC";

   // Real generator name is given as a parameter
   // Only after this is known the proper source name is set
   // for other generator parameters 
   if(*fASCIIEGName=='\0') {
     if(!fRun) fRun = new TPhRun(fRunNo);
     parman = fRun->GetParManager();
     if(parman) 
       if(parman->GetParTextValue(temp, "GENERATOR", sourceMC)) 
	 sscanf(temp,"%119s",fASCIIEGName);
   }

   sscanf(runinfo,"%s %s %[ -~]", ind, keyname, keyvalue);
   // S,M,G are assumed to come from PMC parameters
   if(!strcmp(ind,"S") || !strcmp(ind,"M") || !strcmp(ind,"G")) {
     parkey = new TPhParKey(sourcePMC,keyname,keyvalue);
   }
   // R indicates event generator run parameters
   else if(!strcmp(ind,"R")) {
     if(*fASCIIEGName!='\0') 
          parkey = new TPhParKey(fASCIIEGName,keyname,keyvalue);
     else parkey = new TPhParKey(sourceMC,keyname,keyvalue);
   }
   // others are ignored
   else return kFALSE;

   fRun->AddParameter(parkey);
   return kTRUE;
}

Bool_t TPhASCIIData::AddASCIIEInfo(Char_t *evinfo)
{
#ifdef DEBUG
   printf("AddASCIIEInfo was called with:\n Event parameter:%s\n", evinfo);
#endif
   Char_t temp[120];
   Char_t ind[2], keyname[20], keyvalue[120];
   TPhParManager *parman = NULL;
   TPhParKey *parkey = NULL;
   Char_t sourceMC[]="EventGenerator";

   // Real generator name is given as a parameter
   // Only after this is known the proper source name is set
   // for other generator parameters
   if(!fRun) fRun = new TPhRun(fRunNo);
   if(*fASCIIEGName!='\0') {
     parman = fRun->GetParManager();
     if(parman) 
       if(parman->GetParTextValue(temp, "GENERATOR", sourceMC)) 
	 sscanf(temp,"%119s",fASCIIEGName);
   }

   sscanf(evinfo,"%s %s %[!-~]", ind, keyname, keyvalue);
   // only H is expected to be seen here
   if(!strcmp(ind,"H")) 
     if(*fASCIIEGName!='\0') 
          parkey = new TPhParKey(fASCIIEGName,keyname,keyvalue);
     else parkey = new TPhParKey(sourceMC,keyname,keyvalue);
   // others are ignored
   else return kFALSE;

   fEvent->AddParameter(parkey);
   return kTRUE;
}


Bool_t TPhASCIIData::ReadASCIIHeader()
{
// Read AOU header information.  
// When returning succesfully, last line read into Char_t llread[] string 
// contains first line of information on particles in the event.

#ifdef DEBUG
    printf("Read header info\n");
#endif

    Char_t  stmp1[5];
    Char_t  stmp2[5];                 // Temporary for header label
    Float_t ftmp[10];                 // Temporary for header values
    Char_t  laread[120];

    if( GetrHeader() ) 
    { 
      Warning("ReadASCIIHeader","Repeated attempt to read header info.\nTPhASCIIData object has registered %d lines read.", flinenr);
      return kFALSE;
    }

//  Loop over lines, exiting when the first track is found,
    while ( fgets (laread, 120, fFilePointer) != NULL ) // Warning: end of file
    {
        flinenr++;
        // Look for geometry information in G sector. 
        // In case it is found a TPhDetector object is created.
        // Other header info is stored in the run object
        if ( sscanf(laread, " G %s %s %f", 
             stmp1, stmp2, &ftmp[0]) >= 2)
        {
	  if ( !CreateDetector(laread) ) {
                   Warning("ReadASCIIHeader","Failed to create TPhDetector object.\n" \
                   "TPhASCIIData object has registered %d lines read.", flinenr);
                   return kFALSE;
          }
        }

	// CreateDetector returns last line read that comes 
	// from section other than G
        // This line is immediately processed

        //  First track line found.
        if ( sscanf(laread, " EO%s", stmp1) > 0) 
        {
           if (strcmp( stmp1, "H"  ) == 0) 
           {
             frHeader = kTRUE;
             return kTRUE;
	   }
	}

        // If this is not geometry info (" G") 
	// not a comment (" *") 
	// and not end of header line,
	// put it as a run parameter into a run object
        if ( laread[1] != '*' )
        {
            // Call routine storing run info
            AddASCIIRInfo(laread); 
        }

    }  
    Warning("ReadASCIIHeader","File unreadable or unexpected end of file!\n" \
            "TPhASCIIData object has registered %d lines read.", flinenr);
    return kFALSE;  
}

Bool_t TPhASCIIData::CreateDetector(Char_t *llread)
{
// Read AOU header geometry info and create TPhDetector singleton object.
// This function reads the AOU header and calls TPhDetector::AddSensor() function
// to fill the following subdetectors:
//         "RingMult" Ring Multiplicity detectors
//         "Octagon" Octagon Mult detector detector
//         "Vertex" Vertex detector
//         "SpectP" Positive Spectrometer arm
//         "SpectN" Negative Spectrometer arm
//         "TOFP" Positive Time of flight detector
//         "TOFN" Positive Time of flight detector
//         "Trigger" Trigger detector
// with the sensors and layers. 

  Char_t   stmp1[5];
  Char_t   stmp2[5];                 // Temporary for header label
  Float_t  ftmp[10];                 // Temporary for header values

  Float_t  fGEOV;                    // Geometry version as read from AOU file
  Float_t  rotm[maxDetectorRotMtx][3][3]; 
           // 9 rotation parameter matrices for TPhDetector creation
  Float_t  rota[maxDetectorRotMtx][6];    
           // 6 rotation angle matrices for TPhDetector creation

  Int_t    devn, rnumber;
  Int_t    nrpadf,nrrots,nspars,nrrot[10];
  Float_t  pad[11], sp[40];
  Char_t   SensorName[5];
  Char_t   VersionString[20];

  TRotMatrix *rotmat = NULL;

  fEDigi = new TPhEDigi();

  flinenr--; // small correction
  do {
    flinenr++;
    sscanf(llread, "%2s", stmp1);
    // continue when comment found
    if ( stmp1[0] == '*' ) continue;

    // We finish here contruction of the TPhDetector object
    // hoping that it is complete when the next AOU header section starts
    if ( (sscanf(llread, " G %s %s %f", stmp1, stmp2, &ftmp[0]) < 2) ) {
    return kTRUE;
    }

    if (strcmp( stmp1, "GEOV" ) == 0) {
        sscanf( stmp2, "%f", &fGEOV );
        // We include the PMC version of the geometry in the detector name 
        sprintf(VersionString, "PMC_V%1.2f", fGEOV);
        fDetector->SetName(VersionString);
        fDetector->SetVersion(fGEOV);
        Warning("CreateDetector","Geometry version %6.3f found in AOU file.", 
                 fGEOV);
 	if(fGEOV < 4.0) {
 	  Warning("CreateDetector", "UGCData versions prior to V4.0 will "\
 		  "not have any Ring Sensors in this version of PhAT!\n");
 	}
        continue;
    }
    if (strcmp( stmp1, "DEVO" ) == 0) {
      sscanf(llread, " G %*s %*s %*f %f %f %f %f %f %f", 
	     &ftmp[1], &ftmp[2], &ftmp[3], &ftmp[4], &ftmp[5], &ftmp[6]);
      devn = Int_t(ftmp[0])-1;
      strncpy(SensorName, stmp2, 5);
//  Now reading pad parameters
      flinenr++; 
      if( fgets (llread, 120, fFilePointer) == NULL ) {
	Warning("CreateDetector",
          "Unexpected End of File reading DEVO lines in the Header data\n%s",
         llread);
	Warning("CreateDetector",
          "TPhASCIIData object has registered %d lines read.", flinenr);
	return kFALSE;
      } else {
	nrpadf = sscanf(llread, " %f %f %f %f %f %f %f %f %f %f %f", 
			&pad[0], &pad[1], &pad[2], &pad[3], &pad[4], &pad[5], 
                        &pad[6], &pad[7], &pad[8], &pad[9], &pad[10]);
	if( nrpadf != 11 ) {
	  Warning("CreateDetector",
           "Shortage in pad params reading DEVO lines in the Header data\n%s",
            llread);
	  Warning("CreateDetector",
           "TPhASCIIData object has registered %d lines read.", flinenr);
	  return kFALSE;
	}
      }
//  Now reading shape parameters 
      if( (Int_t)ftmp[2]>40 ) {
	Warning("CreateDetector",
         "Too many shape params in DEVO lines in the Header data\n%s",llread);
	Warning("CreateDetector",
         "TPhASCIIData object has registered %d lines read.", flinenr);
	return kFALSE;
      }
       for( Int_t k=0; k<=(Int_t)ftmp[2]/10; k++) {
         flinenr++;
         if( fgets (llread, 120, fFilePointer) == NULL ) {
           Warning("CreateDetector",
             "Unexpected End of File reading DEVO lines in the Header data\n%s",
              llread);
           Warning("CreateDetector",
                   "TPhASCIIData object has registered %d lines read.", flinenr);
           return kFALSE;
         } else {
           nspars = sscanf(llread, " %f %f %f %f %f %f %f %f %f %f", 
               &sp[0+k*10], &sp[1+k*10], &sp[2+k*10], &sp[3+k*10], &sp[4+k*10],
               &sp[5+k*10], &sp[6+k*10], &sp[7+k*10], &sp[8+k*10], &sp[9+k*10]);
           if( (nspars != 10) && ((nspars+k*10) != (Int_t)ftmp[2]) ) {
             Warning("CreateDetector",
               "Shortage in shape params in DEVO lines in the Header data\n%s",
                llread);
              Warning("CreateDetector",
                      "TPhASCIIData object has registered %d lines read.", 
                       flinenr);
              return kFALSE;
           }
         }
       }
       // Now reading rotation matrix numbers  
       if( (Int_t)ftmp[3]>10 ) {
         Warning("CreateDetector",
                 "Too many rotation matrices in DEVO lines in the Header data.");
         Warning("CreateDetector",
                 "TPhASCIIData object has registered %d lines read.", flinenr);
           return kFALSE;
       }
       // After changes to PMC in September 1997 only one, 
       // final rotation matrix is written to the AOU file. 
       // Although more then one matrix can be read, matrix multiplication 
       // is not implemented.
       if( (Int_t)ftmp[3]>1 ) {
           Warning("CreateDetector",
             "%d rotation matrices in DEVO lines in the Header data\n  - only the first one is actually used to calculate detector geometry parameters",
 (Int_t)ftmp[3]);
           Warning("CreateDetector",
                   "TPhASCIIData object has registered %d lines read.", flinenr);
       }
       flinenr++;
       if( fgets (llread, 120, fFilePointer) == NULL ) {
         Warning("CreateDetector",
            "Unexpected End of File reading DEVO lines in the Header data.\n%s",
             llread);
         Warning("CreateDetector",
                 "TPhASCIIData object has registered %d lines read.", flinenr);
         return kFALSE;
       } else {
           nrrots = sscanf(llread, " %d %d %d %d %d %d %d %d %d %d", 
                    &nrrot[0], &nrrot[1], &nrrot[2], &nrrot[3], &nrrot[4],
                    &nrrot[5], &nrrot[6], &nrrot[7], &nrrot[8], &nrrot[9]);
       if( nrrots != (Int_t)ftmp[3] ) {
         Warning("CreateDetector",
             "Too few rotation nos reading DEVO lines in the Header data.\n%s",
              llread);
         Warning("CreateDetector",
                 "TPhASCIIData object has registered %d lines read.", flinenr);
             return kFALSE;
           }
       }

       Int_t shapeno = (Int_t)ftmp[1]; 
       Int_t nvolpar = (Int_t)ftmp[2]; 
       Int_t padtype = (Int_t)pad[0];
       Float_t *center = ftmp+4;

       // Rotation Matrix is prepared outside the TPhDetector scope and then set 
       // for each sensor object.
       // Correction to rotation matrices for some sensors is applied if
       // geometry is constructed from PMC geometry version 4.2.
       // This is to make some octagon sensors MOAA,MOAB,MOAC,MOAD and
       // the vertex VB.. sensors directed along the increasing phi (XY) angle.
       rotmat = NULL;
       Float_t rotal[6];

       // only one [0] rotation matrix is considered
       rnumber = nrrot[0];
       rotal[0] = rota[rnumber-1][0]; rotal[1] = rota[rnumber-1][1];
       rotal[2] = rota[rnumber-1][2]; rotal[3] = rota[rnumber-1][3];
       rotal[4] = rota[rnumber-1][4]; rotal[5] = rota[rnumber-1][5];

       if( TMath::Nint(10*fDetector->GetVersion()) == 42 ) {      //version 4.2 
          if( strncmp(SensorName, "MOA", 3)==0 ||
              strncmp(SensorName, "MOB", 3)==0 ||
              strncmp(SensorName, "MOC", 3)==0 ||
              strncmp(SensorName, "MOD", 3)==0 ||
              strncmp(SensorName, "VB",  2)==0   )
          {
              rotal[3] = rotal[3] + 180.0;
              rotal[5] = rotal[5] + 180.0;
#ifdef DEBUG
              printf("Change rotation matrix for sensor %s\n", SensorName);
#endif
	  }
       }
       Char_t rotname[40]; 
       sprintf(rotname, "Rot_%d", rnumber-1);
       rotmat = new TRotMatrix(rotname,rotname,
                   rotal[0], rotal[1], rotal[2], rotal[3], rotal[4], rotal[5]);

       // Sensors and layers are created by TPhDetector functions.
       fDetector->AddSensor(devn,SensorName,shapeno,center,nvolpar,sp,
                            rotmat,padtype,pad+1);
       continue;
    }
    if (strcmp( stmp1, "ROTM" ) == 0) 
    {
      sscanf(stmp2, "%d", &rnumber);
      if((rnumber <= 0) || (rnumber>maxDetectorRotMtx)) {
         Warning("CreateDetector",
                 "Error in ROTM data, %d number out of range",rnumber);
         return kFALSE;
      }
      if( sscanf(llread, " G %*s %*f %f %f %f %f %f %f %f %f %f", 
          &rotm[rnumber-1][0][0],&rotm[rnumber-1][1][0],&rotm[rnumber-1][2][0],
          &rotm[rnumber-1][0][1],&rotm[rnumber-1][1][1],&rotm[rnumber-1][2][1],
	  &rotm[rnumber-1][0][2],&rotm[rnumber-1][1][2],&rotm[rnumber-1][2][2]) 
          < 9) {
        Warning("CreateDetector",
                "Error in ROTM data line while reading Header info.\n%s",llread);
        Warning("CreateDetector",
                "TPhASCIIData object has registered %d lines read.", flinenr);
        return kFALSE;
      }
      continue;
    }
    if (strcmp( stmp1, "ROTA" ) == 0) {
       sscanf(stmp2, "%d", &rnumber);
      if((rnumber <= 0) || (rnumber>maxDetectorRotMtx)) {
         Warning("CreateDetector",
                 "Error in ROTA data, %d number out of range",rnumber);
         return kFALSE;
      }
       if( sscanf(llread, " G %*s %*f %f %f %f %f %f %f", 
                  &rota[rnumber-1][0],&rota[rnumber-1][1],&rota[rnumber-1][2],
 	          &rota[rnumber-1][3],&rota[rnumber-1][4],&rota[rnumber-1][5]) 
                  < 6) {
         Warning("CreateDetector",
            "Error in ROTA data line while reading Header info.\n%s",llread);
         Warning("CreateDetector",
            "TPhASCIIData object has registered %d lines read.", flinenr);
         return kFALSE;
       }
       continue;
    }
    Warning("CreateDetector",
            "unknown label in geometry section of the AOU header!\n%s",llread);
    flinenr++;
  } while ( fgets (llread, 120, fFilePointer) != NULL ); // Warning: end of file
  Warning("CreateDetector","File unreadable or unexpected end of file!");
  Warning("CreateDetector","TPhASCIIData object has registered %d lines read.", 
           flinenr);
  return kFALSE;  
}

TPhMCTrack *TPhASCIIData::GetNextTrack(Char_t *llread)
{
// Read information for one track from the data file 
// and create and fill one TPhMCTrack object.

   Int_t i;
   Char_t    lOrigin[10];  // Name of the track's origin volume
   Int_t     lTrackNo;  // Particle number from GEANT
   Int_t     lParentNo;   // Number of the parent track (0 if primary)
   Int_t     lNumHits;  // Number of hits
   Int_t     lParTyp;   // GEANT particle type
   Float_t   lPx;       // Momentum at vertex, x component
   Float_t   lPy;       // Momentum at vertex, y component
   Float_t   lPz;       // Momentum at vertex, z component
   Float_t   lVx;       // Vertex x position
   Float_t   lVy;       // Vertex y position
   Float_t   lVz;       // Vertex z position
   Int_t     lIStop;    // Indicates stopped particle

   Int_t gcharge[50]={ 0, 1,-1, 0, 1,-1, 0, 1,-1, 0,
                       1,-1, 0, 1,-1, 0, 0, 0, 1, 0,
                      -1, 0,-1,-1, 0, 0,-1, 0, 1, 0,
                       1, 1, 0, 0, 0, 0, 0, 0, 0, 0,
                       0, 0, 0, 0, 1, 1, 2, 0, 2, 0};
   Float_t gmass[50]={ 0, 0.00051099906, 0.00051099906, 0, 0.105658389, 
                       0.105658389, 0.1349764, 0.1395700, 0.1395700, 0.497672,
                       0.493677, 0.493677, 0.93956563, 0.83827231,  0.83827231, 
                       0.497672, 0.54745, 1.115684, 1.18937, 1.19255,
                       1.197436, 1.3149, 1.32132, 1.67245, 0.93956563, 
                       1.115684, 1.18937, 1.19255, 1.197436, 1.3149,
                       1.32132, 1.67245, 0, 0, 0, 
                       0, 0, 0, 0, 0,
                       0, 0, 0, 0, 1.875613, 
                       2.80925, 3.727417, 0, 2.80923, 0};

   TPhMCHit *hit;

   lTrackNo = 0;
   lParentNo = 0;
   lNumHits = 0;
   lParTyp = 0;
   lPx = 0;
   lPy = 0;
   lPz = 0;
   lVx = 0;
   lVy = 0;
   lVz = 0;
   lIStop = 0;

#ifdef DEBUG
   printf("Read track info\n");
#endif

   if ( sscanf(llread, " P %s %d %d %d %d",
        lOrigin, &lTrackNo, &lParentNo, &lNumHits, &lParTyp) <= 3)
     {
#ifdef DEBUG
          printf("NULL track pointer returned, error in line format\n");
#endif
          return NULL;           // Error in line format 
     }   
     else
     {
          sscanf(llread, " P %s %d %d %d %d %f %f %f %f %f %f %d",
                 lOrigin, &lTrackNo, &lParentNo, &lNumHits, &lParTyp,
                 &lPx, &lPy, &lPz,&lVx, &lVy, &lVz, &lIStop);
#ifdef DEBUG
          printf("Found track %d hits %d type %d - momentum %g %g %g - stopped %3d\n",
                  lTrackNo, lNumHits, lParTyp, lPx, lPy, lPz, lIStop);
#endif
          TPhMCTrack *track = new TPhMCTrack();
	  track->MakeHitList();

          track->SetOriginName(lOrigin);      
          track->SetTrackNo(lTrackNo);      
          track->SetParentNo(lParentNo);        
          track->SetPx(lPx);
          track->SetPy(lPy);
          track->SetPz(lPz);
          track->SetXProd(lVx);
          track->SetYProd(lVy);
          track->SetZProd(lVz);
          track->SetGPid(lParTyp);
          track->SetIStop(lIStop);
          track->SetCharge( gcharge[lParTyp-1] );
          track->SetMass( gmass[lParTyp-1] );

          for(i=0; i<lNumHits; i++)
          {
             flinenr++;
             if( fgets (llread, 120, fFilePointer) == NULL) break;
	     if( (hit = GetNextHit(llread,track)) != NULL)
	       {
                 track->AddHit(hit);
               }
	  }
	  return track;
     }
}


TPhMCHit *TPhASCIIData::GetNextHit(Char_t *llread, TPhMCTrack *track)
{
// Read information for one hit from the data file 
// and create and fill a TPhMCHit object.

  TPhSensor   *sensor;
    Int_t     NSensor;  // Detector number in the AOU header table of detectors 
    Float_t   X;        // x position of hit, unpixeled
    Float_t   Y;        // y position of hit, unpixeled
    Float_t   Z;        // z position of hit, unpixeled
    Float_t   De;       // Energy loss in the detector
    Float_t   Dt;       // Time of arrival of a particle
    Float_t   Dx;       // x component of vector from entrance to exit
    Float_t   Dy;       // y component of vector from entrance to exit
    Float_t   Dz;       // z component of vector from entrance to exit
    Int_t   PadX;       // Pixel number in "horizontal" direction
    Int_t   PadY;       // Pixel number in "vertical" direction
    Float_t Length;     // Length of the track from the vertex up to the hit
    Int_t   nrow;       // Number of rows in the sensor
    Int_t   ncol;       // Number of columns in the sensor
    Char_t  SensorName[5]; // Name of the sensor where the hit was

    NSensor = 0;
    X = 0;
    Y = 0;
    Z = 0;
    De = 0;
    Dt = 0;
    Dx = 0;
    Dy = 0;
    Dz = 0;
    PadX = 0;
    PadY = 0;
    Length = 0;
    if ( sscanf(llread," %d %f %f %f %f %f %f %f %f %d %d %f",&NSensor,
                         &X, &Y, &Z, &De, &Dt, &Dx, &Dy, &Dz, &PadX, &PadY, &Length)  < 9)
      {
        Warning("GetNextHit",
                "NULL returned while reading hits, error in line format.");
        return NULL;
      }
    else
      {
        TPhMCHit *hit = new TPhMCHit();
        hit->SetX(X);
        hit->SetY(Y);
        hit->SetZ(Z);

	sensor = fDetector->GetSensor(NSensor-1);
        if( !sensor ) {
	  if(fVerbose) {
	    Warning("GetNextHit","Hit in GEANT detector %d, unknown sensor!"\
		    "Hit deleted!", NSensor);
	  }
	  delete hit;
	  return NULL;
        }
	strncpy(SensorName, sensor->GetName(), 5);
#ifdef DEBUG
             printf("Hit in %s sensor (no: %dth in Phobos geometry) read\n",
                            SensorName, NSensor);
#endif
        hit->SetLayer(sensor->GetLayer());
        hit->SetSensorLabel(sensor->GetLabel());
        hit->SetTrackNo(track->GetTrackNo());
        hit->SetDe(De);
        hit->SetDt(Dt);
        hit->SetDx(Dx);
        hit->SetDy(Dy);
        hit->SetDz(Dz);
        hit->SetLength(Length);

	nrow = sensor->GetRows();
        ncol = sensor->GetColumns();

        // Changing PMC convention of pixel numbering used in PMC geometry v.4.2
        // to make it in agreement with the convention used 
	// in the electronic maps
        if( TMath::Nint(10*fDetector->GetVersion()) == 42 ) {  //version 4.2 
          if( strncmp(SensorName, "MOA", 3)==0 ||
              strncmp(SensorName, "MOB", 3)==0 ||
              strncmp(SensorName, "MOC", 3)==0 ||
              strncmp(SensorName, "MOD", 3)==0 ||
              strncmp(SensorName, "VB",  2)==0   )
          {
#ifdef DEBUG
              printf(
              "GetNextHit: Change in sensor %s: Old Ypix = %d, New Ypix = %d\n",
              SensorName,nrow,PadY,nrow-1-PadY);
#endif
              PadY = nrow - 1 - PadY;
	    }
	}

	// I will check on boundaries for pixel numbers
	// PMC sometimes gives values off max/min by 1
        if(ncol) {            // only if sane (>0) value
          if( PadX > ncol-1 ) {
	    if( fVerbose>0 )
             printf(
        "GetNextHit: Change in sensor %s PadX = %d (>ncol(%d)-1): set to %d\n",
        SensorName,PadX,ncol,ncol-1);
     	    PadX = ncol - 1;
          }
          if( PadX<0 ) {
	    if( fVerbose>0 )
             printf("GetNextHit: Change in sensor %s PadX = %d (<0): set to 0\n",
             SensorName,PadY);
            PadX = 0;
          }
	}
        if(nrow) {            // only if sane (>0) value
          if( PadY > nrow-1 ) {
	    if( fVerbose>0 )
             printf(
          "GetNextHit: Change in sensor %s PadY = %d (>nrow(%d)-1): set to %d\n",
          SensorName,PadY,nrow,nrow-1);
     	    PadY = nrow - 1;
          }
          if( PadY<0 ) {
	    if( fVerbose>0 )
             printf("GetNextHit: Change in sensor %s PadY = %d (<0): set to 0\n",
             SensorName,PadY);
            PadY = 0;
          }
	}
        hit->SetPadX(PadX);
        hit->SetPadY(PadY);

        return hit;
    }
}


TPhHitArray *TPhASCIIData::GetNextHitArray(Char_t *llread)
{
  // Read information for one array of digitized pixel energy deposits from 
  // the data file, convert them to ADC values using TPhEDigi object 
  // functions, create and fill TPhHitArray object for one sensor.

   TPhSensor   *sensor;
   TPhHitArray *hitArray;

   Int_t     i;
   Int_t     nrow, ncol;
   Char_t    Plane[5]; // Name of the track's origin volume
   Int_t     Dnumber;  // Detector number from GEANT list
   Int_t     NumSigs;  // Number of signals

   Float_t   X, Y, Z, E, Dt; // X,Y,Z position of the hit, digitized energy,
                                 // time of arrival, (skipped)        
   Int_t     De, PadX, PadY, NPart; // ADC value, X,Y pixel number, 
                                        // number of particles hitting the pixel

#ifdef DEBUG
   printf("Read DIGI header info\n");
#endif

   if( !fDetector ) {
       Warning("GetNextHitArray",
         "Unable to use pixelized data -no TPhDetector- skipping!"); 

       do {
         flinenr++;
         if( fgets(llread, 120, fFilePointer)==NULL ) {
             Warning("GetNextHitArray",
               "Unexpected end of DIGI signal information lines");
             return NULL;
         } 
       } while (strncmp(llread," EOE",4) != 0); 
       return NULL;
   } 

   if (sscanf(llread, " D %s %d %d", Plane, &Dnumber, &NumSigs) < 3) return NULL;
#ifdef DEBUG
   printf("Found DIGI section of %s with %d signals\n", Plane, NumSigs);
#endif
   sensor = NULL;
   if( (Dnumber<maxDEVO) && (sensor = fDetector->GetSensor(Dnumber-1))) {
      hitArray = new TPhHitArray( sensor,Plane,Plane );
      for(i=0; i<NumSigs; i++) {
         flinenr++;
         if( (fgets(llread, 120, fFilePointer)==NULL) || 
             (strncmp(llread," EOE",4)==0) ) {
             Warning("GetNextHitArray",
                     "Unexpected end of DIGI signal information lines in %s", 
                      Plane);
            return NULL;
         } 
         if ( sscanf(llread," %f %f %f %f %f %d %d %d",
                     &X, &Y, &Z, &E, &Dt, &PadX, &PadY, &NPart) < 8) {
            Warning("GetNextHit",
                    "NULL returned while reading DIGI arrays of hit signals.");
            return NULL;
         } 
         if (fEDigi) {
           De = fEDigi->EtoADC( E ); // Energy to ADC conversion
         } else {
           Warning("GetNextHit","No EDigi object found, ADC set to 0!");
           De = 0;
         }  

	nrow = sensor->GetRows();
        ncol = sensor->GetColumns();

        // Changing PMC convention of pixel numbering used in PMC geometry v.4.2
        // to make it in agreement with the convention used 
	// in the electronic maps

        if( TMath::Nint(10*fDetector->GetVersion()) == 42 ) {  //version 4.2 
          if( strncmp(Plane, "MOA", 3)==0 ||
              strncmp(Plane, "MOB", 3)==0 ||
              strncmp(Plane, "MOC", 3)==0 ||
              strncmp(Plane, "MOD", 3)==0 ||
              strncmp(Plane, "VB",  2)==0   )
          {
              PadY = nrow - 1 - PadY;
	  }
	}
	// I will check on boundaries for pixel numbers
	// PMC sometimes gives values off max/min by 1
        if(ncol) {            // only if sane (>0) value
          if( PadX > ncol-1 ) {
  	    if( fVerbose>0 )
              printf(
               "GetNextHitArray: Change in sensor %s PadX = %d (<0): set to 0\n",
                       Plane,PadY);
       	      PadX = ncol-1;
          }
          if( PadX<0 ) {
	    if( fVerbose>0 )
             printf(
               "GetNextHitArray: Change in sensor %s PadX = %d (<0): set to 0\n",
                    Plane,PadY);
            PadX = 0;
          }
	}
        if(nrow) {            // only if sane (>0) value
          if( PadY > nrow-1 ) {
	    if( fVerbose>0 )
             printf(
    "GetNextHitArray: Change in sensor %s PadY = %d (>nrow(%d)-1): set to %d\n",
                    Plane,PadY,nrow,nrow-1);
     	    PadY = nrow - 1;
          }
          if( PadY<0 ) {
	    if( fVerbose>0 )
             printf(
             "GetNextHitArray: Change in sensor %s PadY = %d (<0): set to 0\n",
                    Plane,PadY);
            PadY = 0;
          }
	}
        hitArray->SetADC( PadY,PadX,De ); 
      }
      hitArray->SetNHits(NumSigs);
      return hitArray;
    }  else {
     if(fVerbose) {
       Warning("GetNextHitArray","Unable to use pixelized data "\
	       "-unknown sensor no %d - skipping hitarray for this"\
	       " sensor!", Dnumber); 
     }
     do {
       flinenr++;
       if( fgets(llread, 120, fFilePointer)==NULL ) {
	 Warning("GetNextHitArray","Unexpected end of DIGI signal "\
		 "information lines in %s", Plane);
	 return NULL;
       } 
     } while (strncmp(llread," EOE",4) != 0); 
     return NULL;
   }
}

